Package: collinear
Title: Automated Multicollinearity Management
Version: 3.0.2
Authors@R: 
    person(given = "Blas M.", 
          family = "Benito", , 
          email = "blasbenito@gmail.com", 
          role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-5105-7232"))
URL: https://blasbenito.github.io/collinear/
BugReports: https://github.com/blasbenito/collinear/issues
Description: Provides a comprehensive and automated workflow for managing multicollinearity in data frames with numeric and/or categorical variables. The package integrates five robust methods into a single function: (1) target encoding of categorical variables based on response values (Micci-Barreca, 2001 (Micci-Barreca, D. 2001 <doi:10.1145/507533.507538>); (2) automated feature prioritization to preserve key predictors during filtering; (3 and 4) pairwise correlation and VIF filtering across all variable types (numeric–numeric, numeric–categorical, and categorical–categorical); (5) adaptive correlation and VIF thresholds. Together, these methods enable a reliable multicollinearity management in most use cases while maintaining model integrity. The package also supports parallel processing and progress tracking via the packages 'future' and 'progressr', and provides seamless integration with the 'tidymodels' ecosystem through a dedicated recipe step.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: progressr, future.apply, mgcv, ranger, recipes (>= 1.0.9),
        rlang, spatialData
Suggests: future, testthat (>= 3.0.0), spelling
Config/testthat/edition: 3
Depends: R (>= 4.1.0)
LazyData: true
LazyDataCompression: xz
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-04 06:32:49 UTC; blas
Author: Blas M. Benito [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5105-7232>)
Maintainer: Blas M. Benito <blasbenito@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-04 06:50:02 UTC
Built: R 4.6.0; ; 2026-06-04 09:29:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(bake,step_collinear)
S3method(prep,step_collinear)
S3method(print,collinear_output)
S3method(print,collinear_selection)
S3method(summary,collinear_output)
S3method(summary,collinear_selection)
export(case_weights)
export(collinear)
export(collinear_select)
export(collinear_stats)
export(cor_clusters)
export(cor_cramer)
export(cor_df)
export(cor_matrix)
export(cor_select)
export(cor_stats)
export(drop_geometry_column)
export(f_auto)
export(f_auto_rules)
export(f_binomial_gam)
export(f_binomial_glm)
export(f_binomial_rf)
export(f_categorical_rf)
export(f_count_gam)
export(f_count_glm)
export(f_count_rf)
export(f_functions)
export(f_numeric_gam)
export(f_numeric_glm)
export(f_numeric_rf)
export(identify_categorical_variables)
export(identify_logical_variables)
export(identify_numeric_variables)
export(identify_response_type)
export(identify_valid_variables)
export(identify_zero_variance_variables)
export(model_formula)
export(preference_order)
export(score_auc)
export(score_cramer)
export(score_r2)
export(step_collinear)
export(target_encoding_lab)
export(target_encoding_loo)
export(target_encoding_mean)
export(target_encoding_rank)
export(validate_arg_df)
export(validate_arg_df_not_null)
export(validate_arg_encoding_method)
export(validate_arg_f)
export(validate_arg_function_name)
export(validate_arg_max_cor)
export(validate_arg_max_vif)
export(validate_arg_predictors)
export(validate_arg_preference_order)
export(validate_arg_quiet)
export(validate_arg_responses)
export(vif)
export(vif_df)
export(vif_select)
export(vif_stats)
importFrom(recipes,bake)
importFrom(recipes,prep)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# `collinear` Automated Multicollinearity Management <a href="https://github.com/BlasBenito/collinear"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- Development badges 
&#10;[![R-CMD-check](https://github.com/BlasBenito/collinear/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/BlasBenito/collinear/actions/workflows/R-CMD-check.yaml)
[![Devel-version](https://img.shields.io/badge/devel%20version-1.0.1-blue.svg)](https://github.com/blasbenito/collinear)
&#10;<!-- badges: start -->

[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.10039396.svg)](https://doi.org/10.5281/zenodo.10039396)
[![CRAN
status](https://www.r-pkg.org/badges/version/collinear)](https://cran.r-project.org/package=collinear)
[![CRAN_Download_Badge](http://cranlogs.r-pkg.org/badges/grand-total/collinear)](https://CRAN.R-project.org/package=collinear)
[![R-CMD-check](https://github.com/BlasBenito/collinear/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/BlasBenito/collinear/actions/workflows/R-CMD-check.yaml)

<!-- badges: end -->

## Summary

The R package `collinear` provides a comprehensive toolkit for smart
multicollinearity management in datasets with mixed variable types. The
main function,
[`collinear()`](https://blasbenito.github.io/collinear/reference/collinear.html),
integrates five core components:

- [**Target
  Encoding**](https://blasbenito.github.io/collinear/articles/target_encoding.html)
  (function
  [target_encoding_lab()](https://blasbenito.github.io/collinear/reference/target_encoding_lab.html)):
  Transparently converts categorical predictors to numeric when
  required, enabling VIF and correlation analysis across mixed data
  types.

- [**Intelligent Predictor
  Ranking**](https://blasbenito.github.io/collinear/articles/intelligent_predictor_ranking.html)
  (function
  [preference_order()](https://blasbenito.github.io/collinear/reference/preference_order.html)):
  Prioritizes predictors by their univariate association with the
  response to ensure that the most relevant ones are retained during
  filtering.

- [**Unified Correlation
  Framework**](https://blasbenito.github.io/collinear/articles/unified_correlation_framework.html)
  (function
  [cor_df()](https://blasbenito.github.io/collinear/reference/cor_df.html)):
  Computes pairwise correlations between any variable types using
  Pearson correlation (numeric-numeric), target encoding
  (numeric-categorical), and Cramer’s V (categorical-categorical) within
  a single, consistent workflow.

- [**Adaptive Filtering
  Thresholds**](https://blasbenito.github.io/collinear/articles/adaptive_filtering_thresholds.html)
  (function
  [collinear()](https://blasbenito.github.io/collinear/reference/collinear.html)):
  Automatically configures correlation and VIF thresholds based on each
  dataset’s correlation structure, eliminating guesswork while allowing
  manual override.

- [**Dual Filtering
  Strategy**](https://blasbenito.github.io/collinear/articles/dual_filtering_strategy.html)
  (function
  [collinear_select()](https://blasbenito.github.io/collinear/reference/collinear_select.html)):
  Combines pairwise correlation and Variance Inflation Factor filtering
  while considering predictor rankings to manage multicollinearity while
  maximizing the predictive power of the resulting selection of
  predictors.

These methods, except target encoding are also fully integrated into the
`tidymodels` implementation
[step_collinear()](https://blasbenito.github.io/collinear/reference/step_collinear.html).

The package also provides diagnostic functions
([cor_df()](https://blasbenito.github.io/collinear/reference/cor_df.html),
[vif_df()](https://blasbenito.github.io/collinear/reference/vif_df.html),
[collinear_stats()](https://blasbenito.github.io/collinear/reference/collinear_stats.html))
to help users assess multicollinearity in their data before and after
filtering.

## Install

The package `collinear` can be installed from CRAN.

``` r
install.packages("collinear")
```

The development version can be installed from GitHub.

``` r
remotes::install_github(
  repo = "blasbenito/collinear", 
  ref = "development"
  )
```

Previous versions are in the `archive_xxx` branches of the GitHub
repository.

``` r
remotes::install_github(
  repo = "blasbenito/collinear", 
  ref = "archive_v2.0.0"
  )
```

## Getting Started

### Setup

Most functions in the `collinear` package support parallelization and
progress bars via [`future`](https://future.futureverse.org/) and
[`progressr`](https://progressr.futureverse.org/). Parallelization,
however, is only advantageous for large datasets including categorical
predictors.

``` r
library(collinear)
library(future)
library(progressr)

future::plan(
  future::multisession,
  workers = future::availableCores() - 1
  )

#does not work in Rmarkdown
#progressr::handlers(global = TRUE)
```

### Example Data

The example dataframe
[vi_smol](https://blasbenito.github.io/collinear/reference/vi_smol.html)
has several response variables and a large set of predictors. Here we
focus on the numeric response `vi_numeric`, and the complete set of
numeric and categorical predictors, stored in the vector
[vi_predictors](https://blasbenito.github.io/collinear/reference/vi_predictors.html).

``` r
data(vi_smol, vi_predictors, package = "spatialData")
nrow(vi_smol)
#> [1] 580
length(vi_predictors)
#> [1] 58
```

### Multicollinearity Analysis

The package provides several functions to assess multicollinearity.

The functions
[cor_df()](https://blasbenito.github.io/collinear/reference/cor_df.html)
and
[vif_df()](https://blasbenito.github.io/collinear/reference/vif_df.html)
generate dataframes with the pairwise correlations and VIF scores of the
predictors, while the functions
[cor_stats()](https://blasbenito.github.io/collinear/reference/cor_stats.html),
[vif_stats()](https://blasbenito.github.io/collinear/reference/vif_stats.html)
(used below) and
[collinear_stats()](https://blasbenito.github.io/collinear/reference/collinear_stats.html)
generate descriptive multicollinearity statistics. Notice that the
function takes a while to execute because computing correlations for the
categorical variables in `vi_predictors` is computationally expensive.

``` r
collinear::vif_stats(
  df = vi_smol,
  predictors = vi_predictors,
  quiet = TRUE
)
#> Warning: 
#> collinear::vif_stats()
#> └── collinear::vif_df()
#>     └── collinear::cor_matrix()
#>         └── collinear::cor_df(): 11 categorical predictors have cardinality > 2 and may bias the multicollinearity analysis. Applying target encoding to convert them to numeric will solve this issue.
#>   method     statistic    value
#> 1    vif             n  58.0000
#> 2    vif       minimum  -6.0304
#> 3    vif quantile_0.05  -1.0855
#> 4    vif quantile_0.25   4.5649
#> 5    vif          mean 105.5328
#> 6    vif        median  50.9527
#> 7    vif quantile_0.75 124.4136
#> 8    vif quantile_0.95 409.8273
#> 9    vif       maximum 835.4799
```

The quantile 0.75 returned by `vif_stats()` indicates that 25% of the
predictors have a VIF score higher than 6.5. This suggests substantial
redundancy in the set of predictors.

### Multicollinearity Filtering

To reduce multicollinearity in `vi_smol` we apply
[collinear()](https://blasbenito.github.io/collinear/reference/collinear.html)
with a minimal setup.

``` r
x <- collinear::collinear(
  df = vi_smol,
  response = "vi_numeric",
  predictors = vi_predictors
)
#> 
#> collinear::collinear()
#> └── collinear::validate_arg_df(): converted the following character columns to factor:
#>  - koppen_zone
#>  - koppen_group
#>  - koppen_description
#>  - biogeo_ecoregion
#>  - biogeo_biome
#>  - biogeo_realm
#>  - country_name
#>  - continent
#>  - region
#>  - subregion
#> 
#> collinear::collinear()
#> └── collinear::cor_df(): 11 categorical predictors have cardinality > 2 and may bias the multicollinearity analysis. Applying target encoding to convert them to numeric will solve this issue.
#> 
#> collinear::collinear(): setting 'max_cor' to 0.6846.
#> 
#> collinear::collinear(): setting 'max_vif' to 6.5354.
#> 
#> collinear::collinear()
#> └── collinear::preference_order()
#>     └── collinear::f_auto(): selected function 'f_numeric_rf()' to compute preference order.
#> 
#> collinear::collinear()
#> └── collinear::collinear_select()
#>     └── collinear::vif(): some VIF values exceeded 1M and were set to Inf.
#> 
#> collinear::collinear(): selected predictors: 
#>  - rainfall_mean
#>  - swi_mean
#>  - evapotranspiration_max
#>  - evapotranspiration_range
#>  - swi_min
#>  - soil_soc
#>  - humidity_range
#>  - topo_elevation
#>  - cloud_cover_range
#>  - continent
#>  - soil_sand
#>  - soil_clay
#>  - topo_diversity
#>  - topo_slope
```

The function returns an object of class `collinear_output`, that has its
own `print()` and `summary()` methods.

``` r
x
#> Result
#> ===================
#>  - response: vi_numeric
#>    --------------------
#> 
#>  + df:
#>    - rows: 580
#>    - cols: 15
#> 
#>  + preference order:
#>    + df:
#>      - rows: 58
#>      - cols: 6
#>    + preference:
#>      - rainfall_mean
#>      - growing_season_rainfall
#>      - growing_season_length
#>      - swi_mean
#>      - aridity_index
#>    + f: f_numeric_rf
#> 
#>  + selection:
#>    - rainfall_mean
#>    - swi_mean
#>    - evapotranspiration_max
#>    - evapotranspiration_range
#>    - swi_min
#>    - ... (9 ommited)
#> 
#>  + formulas:
#>    - linear: vi_numeric ~ rainfall_mean + swi_mean + evapotranspiration_max + evapotranspiration_range + swi_min + ... (9 terms omitted) 
#>    - smooth: vi_numeric ~ s(rainfall_mean) + s(swi_mean) + s(evapotranspiration_max) + s(evapotranspiration_range) + s(swi_min) + ... (9 terms omitted)
```

The object `df` contains the response and the selected predictors.

``` r
colnames(x$vi_numeric$df)
#>  [1] "vi_numeric"               "rainfall_mean"           
#>  [3] "swi_mean"                 "evapotranspiration_max"  
#>  [5] "evapotranspiration_range" "swi_min"                 
#>  [7] "soil_soc"                 "humidity_range"          
#>  [9] "topo_elevation"           "cloud_cover_range"       
#> [11] "continent"                "soil_sand"               
#> [13] "soil_clay"                "topo_diversity"          
#> [15] "topo_slope"
```

The object `preference_order` contains the ranking of predictors. It is
computed by the function
[preference_order()](https://blasbenito.github.io/collinear/reference/preference_order.html)
by assessing the association between the response and the predictors. In
this case, it fits univariate models between the response and each
predictor using
[f_numeric_rf](https://blasbenito.github.io/collinear/reference/f_numeric_rf.html),
and returns the R-squared of the observations vs the model predictions.
This ranking ensures that the most important predictors are protected
during multicollinearity filtering.

``` r
head(x$vi_numeric$preference_order)
#>     response               predictor            f    metric  score rank
#> 1 vi_numeric           rainfall_mean f_numeric_rf R-squared 0.8830    1
#> 2 vi_numeric growing_season_rainfall f_numeric_rf R-squared 0.8716    2
#> 3 vi_numeric   growing_season_length f_numeric_rf R-squared 0.8660    3
#> 4 vi_numeric                swi_mean f_numeric_rf R-squared 0.8330    4
#> 5 vi_numeric           aridity_index f_numeric_rf R-squared 0.8276    5
#> 6 vi_numeric             koppen_zone f_numeric_rf R-squared 0.8160    6
```

The object `selection` contains non-collinear predictors chosen by
taking into account their pairwise correlation and VIF against all other
predictors, and their position in the ranking above.

``` r
x$vi_numeric$selection
#>  [1] "rainfall_mean"            "swi_mean"                
#>  [3] "evapotranspiration_max"   "evapotranspiration_range"
#>  [5] "swi_min"                  "soil_soc"                
#>  [7] "humidity_range"           "topo_elevation"          
#>  [9] "cloud_cover_range"        "continent"               
#> [11] "soil_sand"                "soil_clay"               
#> [13] "topo_diversity"           "topo_slope"              
#> attr(,"validated")
#> [1] TRUE
```

We can check that this selection of predictors shows low
multicollinearity by running the function
[vif_df()](https://blasbenito.github.io/collinear/reference/vif_df.html)
on them.

``` r
collinear::vif_df(
  df = x$vi_numeric$df,
  predictors = x$vi_numeric$selection
)
#>        vif                predictor
#> 1   4.9323        cloud_cover_range
#> 2   3.4887                continent
#> 3   3.3868   evapotranspiration_max
#> 4   3.3125 evapotranspiration_range
#> 5   3.1740           humidity_range
#> 6   2.1249            rainfall_mean
#> 7   2.0856                soil_clay
#> 8   2.0259                soil_sand
#> 9   1.8021                 soil_soc
#> 10  1.6395                 swi_mean
#> 11  1.5953                  swi_min
#> 12  1.5152           topo_diversity
#> 13  1.4237           topo_elevation
#> 14 -0.1557               topo_slope
```

All VIF scores are below 2.5, `collinear()` did a good job here!

Finally, `collinear()` also returns modeling formulas to help kick start
exploratory modelling.

``` r
x$vi_numeric$formulas
#> $linear
#> vi_numeric ~ rainfall_mean + swi_mean + evapotranspiration_max + 
#>     evapotranspiration_range + swi_min + soil_soc + humidity_range + 
#>     topo_elevation + cloud_cover_range + continent + soil_sand + 
#>     soil_clay + topo_diversity + topo_slope
#> <environment: 0x5ca0746a67c0>
#> 
#> $smooth
#> vi_numeric ~ s(rainfall_mean) + s(swi_mean) + s(evapotranspiration_max) + 
#>     s(evapotranspiration_range) + s(swi_min) + s(soil_soc) + 
#>     s(humidity_range) + s(topo_elevation) + s(cloud_cover_range) + 
#>     continent + s(soil_sand) + s(soil_clay) + s(topo_diversity) + 
#>     s(topo_slope)
#> <environment: 0x5ca0746a67c0>
```

The function returns linear formulas for numeric outcomes, and
classification formulas for categorical outcomes.

### Model Fitting

The output of `collinear()` can be used right away to fit exploratory
models, as shown below.

``` r
m <- lm(
  formula = x$vi_numeric$formulas$linear, 
  data = x$vi_numeric$df,
  na.action = na.omit
  )

summary(m)
#> 
#> Call:
#> lm(formula = x$vi_numeric$formulas$linear, data = x$vi_numeric$df, 
#>     na.action = na.omit)
#> 
#> Residuals:
#>      Min       1Q   Median       3Q      Max 
#> -0.42086 -0.04433  0.00069  0.04656  0.23475 
#> 
#> Coefficients:
#>                            Estimate Std. Error t value Pr(>|t|)    
#> (Intercept)               2.444e-01  6.263e-02   3.902 0.000107 ***
#> rainfall_mean             3.900e-05  7.425e-06   5.252 2.14e-07 ***
#> swi_mean                  6.849e-03  4.593e-04  14.910  < 2e-16 ***
#> evapotranspiration_max   -1.165e-03  1.759e-04  -6.624 8.24e-11 ***
#> evapotranspiration_range -1.928e-04  1.449e-04  -1.330 0.183968    
#> swi_min                  -2.014e-03  6.539e-04  -3.080 0.002168 ** 
#> soil_soc                 -4.742e-04  1.989e-04  -2.384 0.017453 *  
#> humidity_range           -1.145e-03  6.473e-04  -1.769 0.077474 .  
#> topo_elevation           -4.054e-05  7.512e-06  -5.397 1.00e-07 ***
#> cloud_cover_range         6.015e-04  3.665e-04   1.641 0.101330    
#> continentAsia            -2.660e-02  1.280e-02  -2.078 0.038199 *  
#> continentEurope           4.071e-02  1.995e-02   2.041 0.041765 *  
#> continentNorth America    6.344e-02  1.595e-02   3.977 7.90e-05 ***
#> continentOceania          3.910e-02  1.579e-02   2.476 0.013573 *  
#> continentSouth America    5.631e-02  1.296e-02   4.345 1.65e-05 ***
#> soil_sand                 9.846e-04  4.855e-04   2.028 0.043043 *  
#> soil_clay                 1.127e-03  8.111e-04   1.390 0.165089    
#> topo_diversity            3.321e-03  9.168e-04   3.622 0.000319 ***
#> topo_slope                4.702e-03  1.457e-03   3.227 0.001324 ** 
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> Residual standard error: 0.08245 on 558 degrees of freedom
#>   (3 observations deleted due to missingness)
#> Multiple R-squared:  0.852,  Adjusted R-squared:  0.8472 
#> F-statistic: 178.4 on 18 and 558 DF,  p-value: < 2.2e-16
```

### Integration with `tidymodels`

The function
[step_collinear()](https://blasbenito.github.io/collinear/reference/step_collinear.html)
wraps
[collinear()](https://blasbenito.github.io/collinear/reference/collinear.html)
to facilitate its usage in `tidymodels` recipes. Please notice that
`step_collinear()` does not perform target encoding, as combining this
functionality with multicollinearity filtering does not fit well with
how `recipes` works.

``` r
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
library(recipes)
#> 
#> Attaching package: 'recipes'
#> The following object is masked from 'package:stats':
#> 
#>     step
library(parsnip)
library(workflows)

# model formula
vi_formula <- collinear::model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = vi_predictors
)

# recipe
vi_recipe <- recipes::recipe(
  formula = vi_formula,
  data = vi_smol
  ) |>
  #multicollinearity filtering
  collinear::step_collinear(
    recipes::all_predictors()
  )

#linear model
vi_model <- parsnip::rand_forest() |>
  parsnip::set_engine("ranger", importance = "permutation") |>
  parsnip::set_mode("regression")

#create and fit workflow
vi_model <- parsnip::linear_reg() |>
  parsnip::set_engine("lm")

# create and fit workflow
vi_workflow <- workflows::workflow() |>
  workflows::add_recipe(vi_recipe) |>
  workflows::add_model(vi_model) |>
  workflows::fit(data = vi_smol)
#> Warning: 
#> collinear::collinear()
#> └── collinear::cor_df(): 11 categorical predictors have cardinality > 2 and may bias the multicollinearity analysis. Applying target encoding to convert them to numeric will solve this issue.

vi_workflow
#> ══ Workflow [trained] ══════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: linear_reg()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────
#> 1 Recipe Step
#> 
#> • step_collinear()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────
#> 
#> Call:
#> stats::lm(formula = ..y ~ ., data = data)
#> 
#> Coefficients:
#>              (Intercept)             rainfall_mean                  swi_mean  
#>                2.444e-01                 3.900e-05                 6.849e-03  
#>   evapotranspiration_max  evapotranspiration_range                   swi_min  
#>               -1.165e-03                -1.928e-04                -2.014e-03  
#>                 soil_soc            humidity_range            topo_elevation  
#>               -4.742e-04                -1.145e-03                -4.054e-05  
#>        cloud_cover_range             continentAsia           continentEurope  
#>                6.015e-04                -2.660e-02                 4.071e-02  
#>   continentNorth America          continentOceania    continentSouth America  
#>                6.344e-02                 3.910e-02                 5.631e-02  
#>                soil_sand                 soil_clay            topo_diversity  
#>                9.846e-04                 1.127e-03                 3.321e-03  
#>               topo_slope  
#>                4.702e-03
```

## Citation

If you find this package useful, please cite it as:

*Blas M. Benito (2025). collinear: R Package for Automated
Multicollinearity Management. Version 3.0.0. doi:
10.5281/zenodo.10039396*

## Getting help

If you encounter bugs or issues with the documentation, please [file a
issue on GitHub](https://github.com/BlasBenito/collinear/issues).

# Experiment: Validating Adaptive Multicollinearity Thresholds

# This experiment validates the sigmoid-based adaptive threshold system in
# collinear() by running 10,000 iterations on random subsets of a synthetic
# dataset with realistic correlation structures.
#
# The goal is to verify that:
# 1. Output VIF stays bounded between ~2.5 and ~7.5 across most input conditions
# 2. The system adapts appropriately to different correlation structures
# 3. Predictor retention scales reasonably with input size

#dependencies ----
library(collinear)
library(future)
library(future.apply)
library(progressr)
library(dplyr)
library(ggplot2)
library(patchwork)
library(distantia)

# parallelization ----
future::plan(
  strategy = future::multisession,
  workers = future::availableCores() - 1
)

# simulation parameters ----

#simulated data size
input_columns <- 500
input_rows <- 10000

#total number of repetitions
n_iterations <- 10000

#minimum and maximum number of predictors allowed during subsetting
min_predictors <- 10
max_predictors <- 100

#random seed
seed <- 1

#synthetic data ----
df <- distantia::zoo_simulate(
  name = "sim",
  cols = input_columns,
  rows = input_rows,
  seed = seed,
  seasons = 0,
  independent = FALSE
) |>
  as.data.frame()

#iteration parameters ----
#rows = predictors * 30 to ensure stable VIF estimates
set.seed(seed)
iterations_df <- data.frame(
  n_cols = sample(
    x = min_predictors:max_predictors,
    size = n_iterations,
    replace = TRUE
  )
) |>
  dplyr::rowwise() |>
  dplyr::mutate(
    n_rows = n_cols * sample(
      x = 30:100,
      size = 1
    )
  ) |>
  dplyr::ungroup() |>
  as.data.frame()

# functions ----

#' Extract multicollinearity stats
#' @param stats output from collinear_stats()
#' @param m method: "correlation" or "vif"
#' @param s statistic name: "quantile_0.75", "maximum", etc.
#' @return Numeric value
get_stat <- function(stats, m, s) {
  stats$value[stats$method == m & stats$statistic == s]
}

#' Run one experiment iteration
#' @param df full dataframe with predictors
#' @param n_cols number of columns to subset
#' @param n_rows number of rows to subset
#' @return dataframe with multicollinearity stats
run_iteration <- function(df, n_cols, n_rows) {

  #random subset of rows and predictors
  df_subset <- df[
    sample(nrow(df), n_rows),
    sample(ncol(df), n_cols),
    drop = FALSE
  ]

  #validate (to avoid double validation)
  df_subset <- collinear::validate_arg_df(
    df = df_subset,
    predictors = colnames(df_subset),
    quiet = TRUE
  )

  #input stats
  input_stats <- collinear::collinear_stats(
    df = df_subset,
    quiet = TRUE
  )

  #multicollinearity filtering
  result <- collinear::collinear(
    df = df_subset,
    predictors = colnames(df_subset),
    quiet = TRUE
  )

  #output statistics
  output_stats <- collinear::collinear_stats(
    df = result$result$df,
    predictors = result$result$selection,
    quiet = TRUE
  )

  #results dataframe
  data.frame(
    input_rows = n_rows,
    input_predictors = n_cols,
    output_predictors = length(result$result$selection),
    input_cor_q75 = get_stat(
      input_stats,
      "correlation",
      "quantile_0.75"
    ),
    output_cor_q75 = get_stat(
      output_stats,
      "correlation",
      "quantile_0.75"
    ),
    input_cor_max = get_stat(
      input_stats,
      "correlation",
      "maximum"
    ),
    output_cor_max = get_stat(
      output_stats,
      "correlation",
      "maximum"
    ),
    input_vif_max = get_stat(
      input_stats,
      "vif",
      "maximum"
    ),
    output_vif_max = get_stat(
      output_stats,
      "vif",
      "maximum"
    )
  )

}



#progress bar ----
progressr::handlers(global = TRUE)
progressr::handlers("txtprogressbar")
progressr::with_progress({

  p <- progressr::progressor(steps = n_iterations)

#run iterations in parallel ----
  results_list <- future.apply::future_lapply(
    X = seq_len(n_iterations),
    FUN = function(i) {
      p()
      run_iteration(
        df = df,
        n_cols = iterations_df$n_cols[i],
        n_rows = iterations_df$n_rows[i]
      )
    },
    future.seed = TRUE
  )

})

#combine results ----
experiment_adaptive_thresholds <- results_list |>
  dplyr::bind_rows() |>
  dplyr::arrange(output_vif_max)

#visualization ----
#left panel: shows output VIF across correlation structures
p1 <- ggplot2::ggplot(experiment_adaptive_thresholds) +
  ggplot2::aes(
    x = input_cor_q75,
    y = output_vif_max,
    color = input_predictors
  ) +
  ggplot2::geom_point(alpha = 0.5) +
  ggplot2::geom_hline(yintercept = 2.5, linetype = "dashed", color = "gray30") +
  ggplot2::geom_hline(yintercept = 7.5, linetype = "dashed", color = "gray30") +
  ggplot2::annotate(
    "text",
    x = min(experiment_adaptive_thresholds$input_cor_q75),
    y = 7.65,
    label = "VIF = 7.5",
    hjust = 0,
    size = 3.5,
    color = "gray30"
  ) +
  ggplot2::annotate(
    "text",
    x = min(experiment_adaptive_thresholds$input_cor_q75),
    y = 2.65,
    label = "VIF = 2.5",
    hjust = 0,
    size = 3.5,
    color = "gray30"
  ) +
  ggplot2::scale_color_viridis_c(option = "turbo") +
  ggplot2::scale_y_continuous(breaks = c(1, 2, 3, 4, 5, 6, 7, 8), limits = c(1, 8)) +
  ggplot2::labs(
    title = "Adaptive Threshold Effectiveness",
    x = "Input Correlation (quantile 0.75)",
    y = "Output Maximum VIF",
    color = "Input\nPredictors"
  ) +
  ggplot2::theme_bw()

#predictor retention: shows filtering scales reasonably with input size
p2 <- ggplot2::ggplot(experiment_adaptive_thresholds) +
  ggplot2::aes(
    x = input_predictors,
    y = output_predictors,
    color = output_vif_max
  ) +
  ggplot2::geom_point(alpha = 0.5) +
  ggplot2::geom_smooth(
    method = "gam",
    formula = y ~ s(x, k = 3),
    color = "gray30",
    se = TRUE
  ) +
  ggplot2::scale_color_viridis_c(option = "turbo") +
  ggplot2::scale_y_continuous(breaks = seq(0, 20, 5)) +
  ggplot2::labs(
    title = "Predictor Retention",
    x = "Input Predictors",
    y = "Selected Predictors",
    color = "Output\nmax VIF"
  ) +
  ggplot2::theme_bw()

#combine panels
p1 | p2


#save results ----
usethis::use_data(experiment_adaptive_thresholds, overwrite = TRUE)

#reset parallelization ----
future::plan(sequential)

# Experiment: Pearson Correlation vs Cramer's V Comparison

# This experiment explores the relationship between Pearson correlation and
# Cramer's V across different cardinalities (number of categories) by running
# 1,000 iterations on synthetic paired integer vectors.
#
# The goal is to understand:
# 1. How Cramer's V relates to Pearson correlation for discretized numeric data
# 2. How cardinality affects this relationship
# 3. Whether Cramer's V systematically over- or under-estimates association

# dependencies ----
library(collinear)
library(distantia)
library(ggplot2)

# simulation parameters ----

# total number of repetitions
n_iterations <- 1000

# range of sample sizes
min_rows <- 30
max_rows <- 1000

# cardinality levels to test (number of unique categories)
cardinality_candidates <- seq(
  from = 2,
  to = 10,
  by = 2
)

# random seed
seed <- 1

# iteration parameters ----
set.seed(seed)

iterations_df <- data.frame(
  n_rows = sample(
    x = min_rows:max_rows,
    size = n_iterations,
    replace = TRUE
  ),
  cardinality = sample(
    x = cardinality_candidates,
    size = n_iterations,
    replace = TRUE
  ),
  out_cramer_v = NA_real_,
  out_pearson_cor = NA_real_
)

# functions ----

#' Run one iteration of correlation vs Cramer's V comparison
#' @param n_rows number of rows to simulate
#' @param cardinality number of unique integer values
#' @return named list with Pearson correlation and Cramer's V
run_iteration <- function(n_rows, cardinality) {
  # simulate paired integer vectors
  df_sim <- distantia::zoo_simulate(
    rows = n_rows,
    time_range = c(1, n_rows),
    cols = 2,
    data_range = c(0, cardinality),
    irregular = FALSE
  ) |>
    as.matrix()

  x <- as.integer(df_sim[, 1])
  y <- as.integer(df_sim[, 2])

  # compute metrics
  pearson_cor <- stats::cor(
    x = x,
    y = y
  ) |>
    abs()

  cramer_v <- collinear::cor_cramer(
    x = x,
    y = y
  )

  list(
    out_pearson_cor = pearson_cor,
    out_cramer_v = cramer_v
  )
}

# run iterations ----
for (i in seq_len(nrow(iterations_df))) {
  result_i <- run_iteration(
    n_rows = iterations_df[i, "n_rows"],
    cardinality = iterations_df[i, "cardinality"]
  )

  iterations_df[i, "out_pearson_cor"] <- result_i$out_pearson_cor
  iterations_df[i, "out_cramer_v"] <- result_i$out_cramer_v
}

# combine results ----
experiment_cor_vs_cramer <- iterations_df |>
  dplyr::arrange(cardinality)

# visualization ----
ggplot2::ggplot(experiment_cor_vs_cramer) +
  ggplot2::aes(
    x = out_pearson_cor,
    y = out_cramer_v,
    group = factor(cardinality),
    color = factor(cardinality)
  ) +
  ggplot2::geom_point(alpha = 0.5) +
  ggplot2::geom_smooth(
    method = "gam",
    formula = y ~ s(x, k = 3),
    se = TRUE
  ) +
  ggplot2::geom_abline(
    intercept = 0,
    slope = 1,
    linetype = "dashed",
    color = "gray30"
  ) +
  ggplot2::scale_color_viridis_d(
    option = "turbo",
    name = "Cardinality"
  ) +
  ggplot2::labs(
    title = "Pearson Correlation vs Cramer's V",
    subtitle = "Across different cardinality levels",
    x = "Pearson Correlation (absolute)",
    y = "Cramer's V"
  ) +
  ggplot2::theme_bw() +
  ggplot2::theme(
    legend.position = "right"
  )

# Experiment: Correlation vs VIF Equivalence Mapping

# This experiment establishes the relationship between max_cor and max_vif thresholds by running 10,000 iterations comparing cor_select() and vif_select() outputs.
#
# The goal is to find the max_vif threshold that produces the most similar variable selection as a given max_cor threshold, measured by Jaccard similarity.
#
# Results are used to:
# 1. Understand the relationship between `max_cor` and `max_vif`
# 2. Create the gam_cor_to_vif model for automatic threshold conversion
# 3. Generate the lookup table prediction_cor_vif

#dependencies ----
library(collinear)
library(future)
library(future.apply)
library(progressr)
library(dplyr)
library(ggplot2)
library(distantia)
library(mgcv)

#parallelization ----
future::plan(
  strategy = future::multisession,
  workers = future::availableCores() - 1
)

#simulation parameters ----

##simulated data size
input_columns <- 500
input_rows <- 10000

##total number of repetitions
n_iterations <- 10000

#minimum and maximum number of predictors allowed during subsetting
min_predictors <- 10
max_predictors <- 50

#threshold candidates
max_vif_candidates <- seq(
  from = 1,
  to = 10,
  by = 0.1
  )

max_cor_candidates <- seq(
  from = 0.1,
  to = 0.99,
  by = 0.01
  )

#random seed
seed <- 1

#synthetic data ----
df <- distantia::zoo_simulate(
  name = "sim",
  cols = input_columns,
  rows = input_rows,
  seed = seed,
  seasons = 4,
  independent = FALSE
) |>
  as.data.frame()

#iteration parameters ----
#rows = predictors * 30 to ensure stable VIF estimates
set.seed(seed)
iterations_df <- data.frame(
  max_cor = sample(
    x = max_cor_candidates,
    size = n_iterations,
    replace = TRUE
  ),
  n_cols = sample(
    x = seq(
      from = min_predictors,
      to = max_predictors,
      by = 1
      ),
    size = n_iterations,
    replace = TRUE
  )
) |>
  dplyr::rowwise() |>
  dplyr::mutate(
    n_rows = n_cols * sample(
      x = 30:100,
      size = 1
      )
  ) |>
  dplyr::ungroup() |>
  as.data.frame()

#functions ----

#' Run one iteration of cor vs vif comparison
#' @param i iteration index
#' @param params dataframe with iteration parameters
#' @param data full dataset to subsample from
#' @return dataframe row with results, or NULL if filtering fails
run_iteration <- function(df, n_cols, n_rows, max_cor) {

  #random subset of rows and predictors
  df_subset <- df[
    sample(nrow(df), n_rows),
    sample(ncol(df), n_cols),
    drop = FALSE
  ]


  #validate (to avoid double validation)
  df_subset <- collinear::validate_arg_df(
    df = df_subset,
    predictors = colnames(df_subset),
    quiet = TRUE
  )

  #correlation-based filtering
  selection_cor <- collinear::cor_select(
    df = df_subset,
    predictors = colnames(df_subset),
    max_cor = max_cor,
    quiet = TRUE
  )

  #find max_vif that produces most similar selection to cor_select by maximizing Jaccard similarity
  best_jaccard <- 0
  best_vif_index <- 1
  best_selection_vif <- character(0)

  for (j in seq_along(max_vif_candidates)) {

    selection_vif <- collinear::vif_select(
      df = df_subset,
      predictors = colnames(df_subset),
      max_vif = max_vif_candidates[j],
      quiet = TRUE
    )

    #jaccard similarity
    jaccard <- length(intersect(selection_vif, selection_cor)) /
      length(union(selection_vif, selection_cor))

    ## Update if this is better
    if (jaccard > best_jaccard) {
      best_jaccard <- jaccard
      best_vif_index <- j
      best_selection_vif <- selection_vif
    }
  }

  #return results using best match
  data.frame(
    input_rows = n_rows,
    input_predictors = n_cols,
    output_predictors = length(best_selection_vif),
    max_cor = max_cor,
    max_vif = max_vif_candidates[best_vif_index],
    out_selection_jaccard = best_jaccard
  )

}


#progress bar ----
progressr::handlers(global = TRUE)
progressr::handlers("txtprogressbar")
progressr::with_progress({

  p <- progressr::progressor(steps = n_iterations)

#run iterations in parallel ----
  results_list <- future.apply::future_lapply(
    X = seq_len(n_iterations),
    FUN = function(i) {
      p()
      run_iteration(
        df = df,
        n_cols =  iterations_df[i, "n_cols"],
        n_rows = iterations_df[i, "n_rows"],
        max_cor = iterations_df[i, "max_cor"]
      )
    },
    future.seed = TRUE
  )
})

#combine results ----
experiment_cor_vs_vif <- dplyr::bind_rows(results_list) |>
  dplyr::arrange(out_selection_jaccard)


#model gam_cor_to_vif ----

#check source for weights
hist(experiment_cor_vs_vif$out_selection_jaccard, breaks = 50)

#optimization
#find suitable k and weight exponent to fit the final model
optimization_df <- expand.grid(
  k = 3:9,
  weight_exp = 2:6
)

#iterate over combinations of k and weight_exp
for(i in seq_len(nrow(optimization_df))){

  m.i <- mgcv::gam(
    formula = max_vif ~ s(max_cor, k = optimization_df[i, "k"]),
    weights = experiment_cor_vs_vif$out_selection_jaccard^optimization_df[i, "weight_exp"],
    data = experiment_cor_vs_vif
  )

  #r-squared
  optimization_df[i, "r_squared"] <- summary(m.i)$r.sq

  #degrees of freedom
  optimization_df[i, "edf"] <- sum(m.i$edf)

}

#select top 90% r_squared and bottom 10% edf
#arrange by r_squared
#select top 1
optimization_df <- optimization_df |>
  dplyr::filter(
    r_squared >= stats::quantile(x = r_squared, probs = 0.90)
  ) |>
  dplyr::filter(
    edf <= stats::quantile(x = edf, probs = 0.10)
  ) |>
  dplyr::arrange(
    dplyr::desc(r_squared)
  ) |>
  dplyr::slice_head(n = 1)

optimization_df


#fit model
gam_cor_to_vif <- mgcv::gam(
  formula = max_vif ~ s(max_cor, k = optimization_df$k),
  weights = experiment_cor_vs_vif$out_selection_jaccard^optimization_df$weight_exp,
  data = experiment_cor_vs_vif
)

summary(gam_cor_to_vif)


#visualization
ggplot2::ggplot(experiment_cor_vs_vif) +
  ggplot2::aes(
    x = max_cor,
    y = max_vif,
    color = out_selection_jaccard
  ) +
  ggplot2::geom_point(alpha = 0.5) +
  ggplot2::geom_smooth(
    method = "gam",
    formula = as.formula(paste0("y ~ s(x, k = ", optimization_df$k, ")")),
    mapping = ggplot2::aes(weight = out_selection_jaccard^optimization_df[i, "weight_exp"]),
    color = "black",
    se = TRUE
  ) +
  ggplot2::scale_color_viridis_c(
    option = "turbo",
    direction = 1,
    name = "Jaccard\nSimilarity"
  ) +
  ggplot2::labs(
    title = "Correlation vs VIF Threshold Equivalence",
    x = "Input max_cor Threshold",
    y = "Equivalent max_vif Threshold"
  ) +
  ggplot2::theme_bw() +
  ggplot2::theme(
    legend.position = "right"
  )

#dataframe prediction_cor_to_vif ----
prediction_cor_to_vif <- data.frame(
  max_cor = seq(0.10, 1.00, by = 0.001)
)

prediction_cor_to_vif$max_vif <- mgcv::predict.gam(
  object = gam_cor_to_vif,
  newdata = prediction_cor_to_vif
) |>
  round(digits = 3)


#save results ----
usethis::use_data(experiment_cor_vs_vif, overwrite = TRUE)

usethis::use_data(gam_cor_to_vif, overwrite = TRUE)

usethis::use_data(prediction_cor_to_vif, overwrite = TRUE)


#reset parallelization ----
future::plan(sequential)

#' Convert categorical predictors to numeric via target encoding
#'
#' @description
#'
#' Target encoding maps the values of categorical variables (of class \code{character} or \code{factor}) to numeric using another numeric variable as reference. The encoding methods implemented here are:
#'
#' \itemize{
#'   \item "mean" (implemented in [target_encoding_mean()]): Maps each category to the average of reference numeric variable across the category cases. Variables encoded with this method are identified with the suffix "__encoded_mean". It has a method to control overfitting implemented via the argument \code{smoothing}. The integer value of this argument indicates a threshold in number of rows. Categories sized above this threshold are encoded with the group mean, while groups below it are encoded with a weighted mean of the group's mean and the global mean. This method is named "mean smoothing" in the relevant literature.
#'   \item "rank" (implemented in [target_encoding_rank()]): Returns the rank of the group as a integer, being 1 he group with the lower mean of the reference variable. Variables encoded with this method are identified with the suffix "__encoded_rank".
#'   \item "loo" (implemented in [target_encoding_loo()]): Known as the "leave-one-out method" in the literature, it encodes each categorical value with the mean of the response variable across all other group cases. This method controls overfitting better than "mean". Variables encoded with this method are identified with the suffix "__encoded_loo".
#' }
#'
#' Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).
#'
#' @inheritParams collinear
#'
#' @param response (optional, character string) Name of a numeric response variable in \code{df}. Default: NULL.
#'
#' @param encoding_method (optional; character vector or NULL). Name of the target encoding methods. One or several of: "mean", "rank", "loo". If NULL, target encoding is ignored, and \code{df} is returned with no modification. Default: "loo"
#'
#' @param smoothing (optional; integer vector) Argument of the method "mean". Groups smaller than this number have their means pulled towards the mean of the response across all cases. Default: 0
#'
#' @param overwrite (optional; logical) If TRUE, the original predictors in \code{df} are overwritten with their encoded versions, but only one encoding method, smoothing, white noise, and seed are allowed. Otherwise, encoded predictors with their descriptive names are added to \code{df}. Default: FALSE
#'
#' @inheritParams collinear
#'
#' @return dataframe
#' @examples
#'
#' data(vi_smol, package = "spatialData")
#'
#' #applying all methods for a continuous response
#' df <- target_encoding_lab(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = "koppen_zone",
#'   encoding_method = c(
#'     "mean",
#'     "loo",
#'     "rank"
#'   )
#' )
#'
#' #identify encoded predictors
#' predictors.encoded <- grep(
#'   pattern = "*__encoded*",
#'   x = colnames(df),
#'   value = TRUE
#' )
#'
#' head(df[, predictors.encoded])
#'
#'
#' @autoglobal
#' @family target_encoding
#' @author Blas M. Benito, PhD
#' @references
#' \itemize{
#'  \item Micci-Barreca, D. (2001) A Preprocessing Scheme for High-Cardinality Categorical Attributes in Classification and Prediction Problems. SIGKDD Explor. Newsl. 3, 1, 27-32. doi: 10.1145/507533.507538
#' }
#' @export
target_encoding_lab <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  encoding_method = "loo",
  smoothing = 0,
  overwrite = FALSE,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::target_encoding_lab()",
    ... = ...
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  # early stops ----

  ## df is NULL ----
  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  # encoding method
  encoding_method <- validate_arg_encoding_method(
    encoding_method = encoding_method,
    overwrite = overwrite,
    quiet = quiet,
    function_name = function_name
  )

  if (is.null(encoding_method)) {
    return(df)
  }

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  if (is.null(response)) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'response' is NULL, skipping target encoding."
      )
    }

    return(df)
  }

  predictors <- validate_arg_predictors(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #check that response is numeric
  response_type <- identify_response_type(
    df = df,
    response = response,
    quiet = quiet,
    function_name = function_name
  )

  if (response_type %in% c("categorical", "unknown")) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": 'response' column '",
        response,
        "' is not numeric, skipping target encoding."
      )
    }

    return(df)
  }

  ## predictors ----
  predictors <- identify_categorical_variables(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )$valid

  if (length(predictors) == 0) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": no categorical predictors in argument 'predictors', skipping target encoding."
      )
    }

    return(df)
  }

  # overwrite ----
  if (is.logical(overwrite) == FALSE) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'overwrite' must be logical, resetting it to 'FALSE'."
      )
    }

    overwrite <- FALSE
  }

  ## smoothing ----
  if ("mean" %in% encoding_method) {
    smoothing <- as.integer(smoothing)

    smoothing <- smoothing[smoothing >= 0 & smoothing <= nrow(df)]

    if (length(smoothing) == 0) {
      smoothing <- 0

      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": invalid values in argument 'smoothing', resetting it to '",
          smoothing,
          "'."
        )
      }
    }

    #depends on overwrite
    if (
      overwrite == TRUE &&
        length(smoothing) > 1
    ) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": only one 'smoothing' value allowed when 'overwrite = TRUE', resetting it to '",
          smoothing[1],
          "'."
        )
      }

      smoothing <- smoothing[1]
    }
  }

  #message to start encoding
  if (quiet == FALSE) {
    message(
      "\n",
      function_name,
      ": using response '",
      response,
      "' to encode these categorical predictors:\n - ",
      paste0(
        predictors,
        collapse = "\n - "
      )
    )
  }

  # target-encoding ----
  combinations.df <- expand.grid(
    predictor = predictors,
    smoothing = smoothing,
    encoding_method = encoding_method,
    stringsAsFactors = FALSE
  )

  #progress bar
  p <- progressr::progressor(
    steps = nrow(combinations.df)
  )

  #original dataframe names
  df_names <- colnames(df)

  #parallel encoding
  encoded_list <- future.apply::future_apply(
    X = combinations.df,
    MARGIN = 1,
    FUN = function(x) {
      p()

      #testing vector
      # x <- c("biogeo_ecoregion", "0", "1", "mean")
      # names(x) <- c("predictor", "smoothing", "seed", "method")

      #args
      smoothing.i <- as.integer(x["smoothing"])
      encoding_method.i <- x["encoding_method"]
      predictor.i <- x["predictor"]

      #subset df
      df.i <- df[, c(response, predictor.i)]

      #to character
      df.i[[predictor.i]] <- as.character(df.i[[predictor.i]])

      #NA to "NA"
      df.i[[predictor.i]] <- replace(
        x = df.i[[predictor.i]],
        list = is.na(df.i[[predictor.i]]),
        values = "NA"
      )

      #new predictor name
      #name of smoothing method
      #only for "mean" and smoothing != 0
      name.smoothing <- ifelse(
        test = encoding_method.i == "mean" && smoothing.i != 0,
        yes = paste0("__smoothing_", smoothing.i),
        no = ""
      )

      #predictor name
      predictor_encoded_name.i <- paste0(
        predictor.i,
        "__encoded_",
        encoding_method.i,
        name.smoothing
      )

      #get encoding function
      f.i <- switch(
        encoding_method.i,
        loo = target_encoding_loo,
        mean = target_encoding_mean,
        rank = target_encoding_rank
      )

      #apply encoding function
      df.i <- f.i(
        df = df.i,
        response = response,
        predictor = predictor.i,
        encoded_name = predictor_encoded_name.i,
        smoothing = smoothing.i,
        function_name = function_name
      )

      #replacing original variable with encoded version
      if (overwrite == TRUE) {
        #remove original column
        df.i[[predictor.i]] <- NULL

        #rename encoded column
        colnames(df.i)[
          colnames(df.i) == predictor_encoded_name.i
        ] <- predictor.i
      } else {
        predictor.i <- predictor_encoded_name.i
      }

      df.i[, predictor.i, drop = FALSE]
    }, #end of lambda function
    future.seed = TRUE
  ) #end of loop

  # encoded predictors ----
  encoded_df <- do.call(
    what = "cbind",
    args = encoded_list
  )

  #remove original predictors
  if (overwrite == TRUE) {
    df <- df[, !(colnames(df) %in% predictors), drop = FALSE]
  }

  #merge encoded data
  df <- cbind(
    df,
    encoded_df
  )

  #add validated tag
  attr(
    x = df,
    which = "validated"
  ) <- TRUE

  df
}

testthat::test_that("`target_encoding_lab()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #stops

  #null response
  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = NULL
    ),
    regexp = "argument 'response' is NULL, skipping target encoding"
  )

  #categorical response
  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_categorical"
    ),
    regexp = "'response' column 'vi_categorical' is not numeric, skipping target encoding"
  ) |>
    suppressMessages()

  #no categorical predictors
  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric
    ),
    regexp = "no categorical predictors in argument 'predictors', skipping target encoding"
  ) |>
    suppressMessages()

  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_categorical,
      overwrite = "hola"
    ),
    regexp = "argument 'overwrite' must be logical, resetting it to 'FALSE'"
  ) |>
    suppressMessages()

  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_categorical,
      overwrite = FALSE,
      encoding_method = NULL
    ),
    regexp = "argument 'encoding_method' is NULL, skipping target encoding"
  ) |>
    suppressMessages()

  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_categorical,
      overwrite = FALSE,
      encoding_method = "hola"
    ),
    regexp = "argument 'encoding_method' is not valid, resetting it to 'loo'"
  ) |>
    suppressMessages()

  #smoothing reset to zero
  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_categorical,
      overwrite = FALSE,
      encoding_method = "mean",
      smoothing = nrow(vi_smol) + 1
    ),
    regexp = "invalid values in argument 'smoothing', resetting it to '0'"
  ) |>
    suppressMessages()

  #several smoothing and overwrite = TRUE
  testthat::expect_message(
    df <- target_encoding_lab(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_categorical,
      overwrite = TRUE,
      encoding_method = "mean",
      smoothing = c(10, 20),
    ),
    regexp = "only one 'smoothing' value allowed when 'overwrite = TRUE', resetting it to '10'"
  ) |>
    suppressMessages()

  #proper call
  df <- target_encoding_lab(
    df = vi_smol,
    response = "vi_numeric",
    predictors = "koppen_zone",
    overwrite = FALSE,
    encoding_method = c("mean", "rank", "loo"),
    smoothing = c(0, 30),
    quiet = TRUE
  )

  # Check if the result is a data frame
  testthat::expect_true(
    is.data.frame(df)
  )

  # Check if the encoded variables have been added
  testthat::expect_true(
    "koppen_zone__encoded_mean" %in% colnames(df)
  )

  testthat::expect_true(
    "koppen_zone__encoded_mean__smoothing_30" %in% colnames(df)
  )

  testthat::expect_true(
    "koppen_zone__encoded_rank" %in% colnames(df)
  )

  testthat::expect_true(
    "koppen_zone__encoded_loo" %in% colnames(df)
  )

  # Check if encoding methods have been applied
  testthat::expect_true(
    is.numeric(df$koppen_zone__encoded_mean)
  )

  testthat::expect_true(
    is.numeric(df$koppen_zone__encoded_mean__smoothing_30)
  )

  testthat::expect_true(
    is.numeric(df$koppen_zone__encoded_rank)
  )

  testthat::expect_true(
    is.numeric(df$koppen_zone__encoded_loo)
  )
})

#' @title Dual multicollinearity filtering algorithm
#'
#' @description
#' Automatizes multicollinearity filtering via pairwise correlation and/or variance inflation factors in dataframes with numeric and categorical predictors.
#'
#' The argument \code{max_cor} determines the maximum pairwise correlation allowed in the resulting selection of predictors, while \code{max_vif} does the same for variance inflation factors.
#'
#' The argument \code{preference_order} accepts a character vector of predictor names ranked from first to last index, or a dataframe resulting from [preference_order()]. When two predictors in this vector or dataframe are highly collinear, the one with a lower ranking is removed. This option helps protect predictors of interest. If not provided, predictors are ranked from lower to higher multicollinearity.
#' Please check the sections **Variance Inflation Factors**, **VIF-based Filtering**, and **Pairwise Correlation Filtering** at the end of this help file for further details.
#'
#'
#' @inheritSection collinear Pairwise Correlation Filtering
#' @inheritSection collinear Variance Inflation Factors
#' @inheritSection collinear VIF-based Filtering
#'
#' @inheritParams collinear
#' @param response (optional; character or NULL) Name of one response variable in \code{df}. Used to exclude columns when \code{predictors} is NULL, and to filter \code{preference_order} when it is a dataframe and contains several responses. Default: NULL.
#' @param max_cor (optional; numeric or NULL) Maximum correlation allowed between pairs of \code{predictors}. Valid values are between 0.01 and 0.99, and recommended values are between 0.5 (strict) and 0.9 (permissive). Default: 0.7
#'
#' @param max_vif (optional, numeric or NULL) Maximum Variance Inflation Factor allowed for \code{predictors} during multicollinearity filtering. Recommended values are between 2.5 (strict) and 10 (permissive). Default: 5
#' @return character vector: names of selected predictors
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' x <- collinear_select(
#'   df = vi_smol,
#'   predictors = c(
#'   "koppen_zone",           #character
#'   "soil_type",             #factor
#'   "topo_elevation",        #numeric
#'   "soil_temperature_mean"  #numeric
#'     ),
#'   max_cor = 0.7,
#'   max_vif = 5
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_filtering
#' @author Blas M. Benito, PhD
#' @export
collinear_select <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  preference_order = NULL,
  max_cor = 0.61,
  max_vif = 5,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::collinear_select()",
    function_name = dots$function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    function_name = function_name,
    quiet = quiet
  )

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  predictors <- validate_arg_predictors(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df.ncol <- ncol(df)

  df <- validate_arg_df(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #revalidate predictors if any columns were removed
  if (ncol(df) < df.ncol) {
    attributes(response)$validated <- NULL
    attributes(predictors)$validated <- NULL

    response <- validate_arg_responses(
      df = df,
      responses = response,
      max_responses = 1,
      quiet = quiet,
      function_name = function_name
    )

    predictors <- validate_arg_predictors(
      df = df,
      responses = response,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #univariate case
  if (length(predictors) == 1) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": only one valid predictor in 'predictors', skipping multicollinearity filtering."
      )
    }

    return(predictors)
  }

  if (is.null(max_cor) && is.null(max_vif)) {
    stop(
      "\n",
      function_name,
      ": arguments 'max_cor' and 'max_vif' cannot be NULL at once.",
      call. = FALSE
    )
  }

  m <- cor_matrix(
    df = df,
    predictors = predictors,
    function_name = function_name,
    quiet = quiet,
    m = dots$m
  )

  max_cor <- validate_arg_max_cor(
    max_cor = max_cor,
    function_name = function_name,
    quiet = quiet
  )

  max_vif <- validate_arg_max_vif(
    max_vif = max_vif,
    function_name = function_name,
    quiet = quiet
  )

  preference.order <- validate_arg_preference_order(
    df = df,
    response = response,
    predictors = predictors,
    preference_order = preference_order,
    quiet = quiet,
    function_name = function_name,
    m = m
  )

  preference.order <- preference.order$predictor

  #organize the correlation matrix according to preference_order
  m.class <- class(m)
  m <- m[
    preference.order,
    preference.order
  ]
  class(m) <- m.class

  #check if filtering is needed
  skip_cor <- FALSE
  skip_vif <- FALSE

  #check cor threshold
  if (!is.null(max_cor)) {
    if (max(abs(m[upper.tri(m)])) <= max_cor) {
      skip_cor <- TRUE

      if (quiet == FALSE && is.null(max_vif)) {
        message(
          "\n",
          function_name,
          ": maximum pairwise correlation is <= ",
          max_cor,
          ", multicollinearity filtering is not required."
        )

        return(predictors)
      }
    }
  }

  if (!is.null(max_vif)) {
    current_max_vif <- max(
      vif(
        m = m,
        quiet = quiet,
        function_name = function_name
      )
    )

    #check vif threshold
    if (current_max_vif <= max_vif) {
      skip_vif <- TRUE

      if (quiet == FALSE && is.null(max_cor)) {
        message(
          "\n",
          function_name,
          ": maximum VIF is <= ",
          max_vif,
          ", multicollinearity filtering is not required."
        )

        return(predictors)
      }
    }
  }

  #both are below thresholds
  if (skip_cor && skip_vif) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": multicollinearity is below 'max_cor' and 'max_vif', filtering is not required."
      )
    }

    return(predictors)
  }

  #filtering loop
  selected <- preference.order[1]
  candidates <- preference.order[-1]

  #iterate over candidates
  for (candidate in candidates) {
    #correlation criterion (if applicable)
    cor_ok <- TRUE

    if (!is.null(max_cor) && !skip_cor) {
      cor_ok <- max(abs(m[selected, candidate])) <= max_cor
    }

    #VIF criterion, only checked if cor passed
    vif_ok <- TRUE

    if (!is.null(max_vif) && !skip_vif && cor_ok) {
      selected_columns <- c(
        selected,
        candidate
      )

      current_vif <- vif(
        m = m[
          selected_columns,
          selected_columns
        ],
        quiet = quiet,
        function_name = function_name
      ) |>
        max()

      vif_ok <- current_vif <= max_vif
    }

    #add candidate if all criteria is TRUE
    if (cor_ok && vif_ok) {
      selected <- c(selected, candidate)
    }
  }

  attr(
    x = selected,
    which = "validated"
  ) <- TRUE

  selected
}

#' Check and constrain argument \code{max_vif}
#'
#' @inheritParams collinear_select
#' @inheritParams validate_arg_quiet
#'
#' @return numeric or NULL
#' @autoglobal
#' @export
#' @family argument_validation
#' @examples
#' max_vif <- validate_arg_max_vif(
#'   max_vif = 11, #wrong value
#'   quiet = FALSE
#' )
#'
#' max_vif
#' attributes(max_vif)$validated
validate_arg_max_vif <- function(
  max_vif = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_max_vif()",
    function_name = function_name
  )

  max_vif_default <- 5

  if (isTRUE(attr(x = max_vif, which = "validated"))) {
    return(max_vif)
  }

  if (is.null(max_vif)) {
    return(NULL)
  }

  if (is.numeric(max_vif)) {
    if (length(max_vif) > 1) {
      max_vif <- max_vif[1]

      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": argument 'max_vif' must be of length one, using value '",
          max_vif,
          "'."
        )
      }
    }
  } else {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'max_vif' is non-numeric, resetting it to to '",
        max_vif_default,
        "'."
      )
    }

    max_vif <- max_vif_default
  }

  if (max_vif > 10 || max_vif < 1) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'max_vif' is outside its valid range (>=1 to <=10), resetting it to '",
        max_vif_default,
        "'."
      )
    }

    max_vif <- max_vif_default
  }

  attr(
    x = max_vif,
    which = "validated"
  ) <- TRUE

  max_vif
}

testthat::test_that("`validate_arg_max_vif()` works", {
  testthat::skip_on_cran()

  x <- validate_arg_max_vif(
    max_vif = NULL,
    quiet = FALSE
  )

  testthat::expect_null(x)

  testthat::expect_message(
    x <- validate_arg_max_vif(
      max_vif = c(5, 10),
      quiet = FALSE
    ),
    regexp = "argument 'max_vif' must be of length one, using value '5'"
  )

  testthat::expect_equal(as.numeric(x), 5, tolerance = 1e-10)

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_no_message(
    x <- validate_arg_max_vif(
      max_vif = x,
      quiet = FALSE
    )
  )

  testthat::expect_message(
    x <- validate_arg_max_vif(
      max_vif = "hola",
      quiet = FALSE
    ),
    regexp = "argument 'max_vif' is non-numeric, resetting it to to '5'"
  )

  testthat::expect_equal(as.numeric(x), 5, tolerance = 1e-10)

  testthat::expect_message(
    x <- validate_arg_max_vif(
      max_vif = 12,
      quiet = FALSE
    ),
    regexp = "is outside its valid range"
  )

  testthat::expect_equal(as.numeric(x), 5, tolerance = 1e-10)

  testthat::expect_true(
    attributes(x)$validated
  )
})

#' Build hierarchical function names for messages
#'
#' @description
#' Concatenates parent and child function names for a better message, warning, and error tracing.
#'
#'
#' @param default_name (optional, character) Name of the calling function. Default: NULL
#' @param function_name (optional, character) Name of the parent function. Default: NULL
#' @param ... (optional) Used to pass \code{function_name} within these functions that don't have this argument.
#'
#' @returns character
#' @export
#' @family argument_validation
#' @examples
#' x <- validate_arg_function_name(
#'   default_name = "child_function",
#'   function_name = "parent_function"
#' )
#'
#' message(x)
#' @autoglobal
validate_arg_function_name <- function(
  default_name = NULL,
  function_name = NULL,
  ...
) {
  if (all(is.null(c(default_name, function_name)))) {
    return(NULL)
  }

  if (is.null(function_name) && !is.null(default_name)) {
    return(default_name)
  }

  hirarchy_symbol <- "\u2514\u2500\u2500 "
  spaces <- "    "

  spaces_multiplier <- length(
    regmatches(
      x = function_name,
      m = gregexpr(
        pattern = hirarchy_symbol,
        text = function_name
      )
    )[[1]]
  )

  spaces <- paste0(
    rep(x = "    ", times = spaces_multiplier),
    collapse = ""
  )

  function_name <- paste0(
    function_name,
    "\n",
    spaces,
    hirarchy_symbol,
    default_name
  )

  function_name
}

testthat::test_that("`validate_arg_function_name()` works", {
  testthat::skip_on_cran()

  x <- validate_arg_function_name(
    default_name = NULL,
    function_name = NULL
  )

  testthat::expect_null(x)

  x <- validate_arg_function_name(
    default_name = "hola",
    function_name = NULL
  )

  testthat::expect_true(
    x == "hola"
  )

  x <- validate_arg_function_name(
    default_name = "hola",
    function_name = "adios"
  )

  testthat::expect_true(
    grepl(pattern = "hola", x = x)
  )

  testthat::expect_true(
    grepl(pattern = "adios", x = x)
  )
})

#to ensure there are no check NOTES about spatialData

#' Declare spatialData
#' @returns dataset spatialData::vi_smol
#' @autoglobal
#' @keywords internal
declare_spatialData <- function() {
  spatialData::vi_smol
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage("collinear ", utils::packageVersion("collinear"))
}

#' Detect response variable type for model selection
#'
#' @description
#' Used by [f_auto()] to identify the type of a response variable and select a proper modelling method to compute preference order.
#' Supported types are:
#' \itemize{
#'   \item "continuous-binary": decimal numbers and two unique values; results in a warning, as this type is difficult to model.
#'   \item "continuous-low": decimal numbers and 3 to 5 unique values; results in a message, as this type is difficult to model.
#'   \item "continuous-high": decimal numbers and more than 5 unique values.
#'   \item "integer-binomial": integer with 0s and 1s, suitable for binomial models.
#'   \item "integer-binary": integer with 2 unique values other than 0 and 1; returns a warning, as this type is difficult to model.
#'   \item "integer-low": integer with 3 to 5 unique values or meets specified thresholds.
#'   \item "integer-high": integer with more than 5 unique values suitable for count modelling.
#'   \item "categorical": character or factor with 2 or more levels.
#'   \item "unknown": when the response type cannot be determined.
#' }
#'
#' @inheritParams collinear
#' @inheritParams f_auto
#'
#' @return character string: response type
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' identify_response_type(
#'   df = vi_smol,
#'   response = "vi_numeric"
#' )
#'
#' identify_response_type(
#'   df = vi_smol,
#'   response = "vi_counts"
#' )
#'
#' identify_response_type(
#'   df = vi_smol,
#'   response = "vi_binomial"
#' )
#'
#' identify_response_type(
#'   df = vi_smol,
#'   response = "vi_categorical"
#' )
#'
#' identify_response_type(
#'   df = vi_smol,
#'   response = "vi_factor"
#' )
#'
#' @export
#' @family data_types
#' @autoglobal
identify_response_type <- function(
  df = NULL,
  response = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_response_type()",
    ... = ...
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (is.null(response)) {
    stop(
      "\n",
      function_name,
      ": argument 'response' must not be NULL.",
      call. = FALSE
    )
  }

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df(
    df = df,
    responses = response,
    predictors = NULL,
    quiet = quiet,
    function_name = function_name
  )

  # extract response
  x <- sort(unique(df[[response]]))
  x_length <- length(x)

  #if constant, error
  if (x_length == 1) {
    stop(
      "\n",
      function_name,
      ": Argument 'response' with value '",
      response,
      "' names a column with constant values. Please select a different response column.",
      call. = FALSE
    )
  }

  #numeric types
  if (is.numeric(x)) {
    # continuous values
    if (!all(x %% 1 == 0)) {
      if (x_length == 2) {
        if (quiet == FALSE) {
          message(
            "\n",
            function_name,
            ": argument 'response' names a numeric non-integer column with two unique values. Please consider recoding it as categorical, or select a different response column."
          )
        }

        return("continuous-binary")
      } else if (x_length <= 5) {
        if (quiet == FALSE) {
          message(
            "\n",
            function_name,
            ": argument 'response' names a numeric non-integer column with 5 or fewer values. Please consider recoding it as integer or categorical, or select a different response column."
          )
        }

        return("continuous-low")
      } else {
        return("continuous-high")
      }
    } else {
      #integer values

      #integer counts
      if (all(x == as.integer(x))) {
        if (x_length == 2) {
          # binomial: 0s and 1s
          if (all(x %in% c(0, 1))) {
            return("integer-binomial")
          } else {
            return("integer-binary")
          }
        } else if (
          x_length <= 5 ||
            max(x) <= 15 ||
            mean(x) <= 5
        ) {
          return("integer-low")
        } else {
          return("integer-high")
        }
      }
    }
  }

  # categorical
  if (is.factor(x) || is.character(x)) {
    return("categorical")
  }

  return("unknown")
}

testthat::test_that("`identify_response_type()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  testthat::expect_error(
    x <- identify_response_type(
      df = NULL,
      response = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  testthat::expect_error(
    x <- identify_response_type(
      df = vi_smol,
      response = NULL
    ),
    regexp = "argument 'response' must not be NULL"
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "vi_numeric",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "continuous-high"
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "vi_counts",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "integer-high"
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "vi_binomial",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "integer-binomial"
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "vi_categorical",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "categorical"
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "vi_factor",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "categorical"
  )

  vi_smol$response_constant <- 1

  testthat::expect_error(
    x <- identify_response_type(
      df = vi_smol,
      response = "response_constant"
    ),
    regexp = "argument 'df' has no valid columns"
  ) |>
    suppressMessages()

  vi_smol$response_binary <- c(1.1, 2.1)

  testthat::expect_message(
    x <- identify_response_type(
      df = vi_smol,
      response = "response_binary"
    ),
    regexp = "argument 'response' names a numeric non-integer column with two unique values"
  ) |>
    suppressMessages()

  vi_smol$response_binary <- c(1, 2)

  x <- identify_response_type(
    df = vi_smol,
    response = "response_binary",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "integer-binary"
  )

  vi_smol$response_trinary <- sample(
    c(1.1, 2.1, 3.1),
    size = nrow(vi_smol),
    replace = TRUE
  )

  testthat::expect_message(
    x <- identify_response_type(
      df = vi_smol,
      response = "response_trinary"
    ),
    regexp = "argument 'response' names a numeric non-integer column with 5 or fewer values"
  ) |>
    suppressMessages()

  vi_smol$response_trinary <- sample(
    c(1, 2, 3),
    size = nrow(vi_smol),
    replace = TRUE
  )

  x <- identify_response_type(
    df = vi_smol,
    response = "response_trinary",
    quiet = TRUE
  )

  testthat::expect_true(
    x == "integer-low"
  )
})

#' Group predictors by hierarchical correlation clustering
#'
#' @description
#'
#' Hierarchical clustering of predictors from their correlation matrix. Computes the correlation matrix with [cor_df()] and [cor_matrix()], transforms it to a distance matrix using \code{stats::dist()}, computes a clustering solution with \code{stats::hclust()}, and applies \code{stats::cutree()} to separate groups based on the value of the argument \code{max_cor}.
#'
#' Returns a dataframe with predictor names and their clusters, and optionally, prints a dendrogram of the clustering solution.
#'
#' Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).
#'
#' @inheritParams cor_matrix
#' @param max_cor (optional; numeric or NULL) Correlation value used to separate clustering groups. Valid values are between 0.01 and 0.99. Default: 0.7
#' @param method (optional, character string) Argument of [stats::hclust()] defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC) or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well. Default: "complete".
#'
#' @return list:
#' \itemize{
#'   \item df: dataframe with predictor names and their cluster IDs.
#'   \item hclust: clustering object
#' }
#'
#' @examples
#'data(vi_smol, package = "spatialData")
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' #group predictors using max_cor as clustering threshold
#' clusters <- cor_clusters(
#'   df = vi_smol,
#'   predictors = c(
#'     "koppen_zone", #character
#'     "soil_type", #factor
#'     "topo_elevation", #numeric
#'     "soil_temperature_mean" #numeric
#'   ),
#'   max_cor = 0.75
#' )
#'
#' #clusters dataframe
#' clusters$df
#'
#' ##plot hclust object
#' # graphics::plot(clusters$hclust)
#'
#' ##plot max_cor threshold
#' # graphics::abline(
#' #   h = 1 - 0.75,
#' #   col = "red4",
#' #   lty = 3,
#' #   lwd = 2
#' # )
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @export
#' @family multicollinearity_assessment
#' @autoglobal
cor_clusters <- function(
  df = NULL,
  predictors = NULL,
  max_cor = 0.7,
  method = "complete",
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_clusters()",
    ... = ...
  )

  m <- cor_matrix(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  ) |>
    abs()

  m <- stats::as.dist(1 - m)

  hc <- tryCatch(
    stats::hclust(d = m, method = method),
    error = function(e) {
      stop(
        function_name <- validate_arg_function_name(
          default_name = "stats::hclust()",
          function_name = function_name
        ),
        ": clustering failed: ",
        e$message,
        call. = FALSE
      )
    }
  )

  hc_groups <- stats::cutree(
    tree = hc,
    h = 1 - max_cor,
  )

  df_clusters <- data.frame(
    predictor = names(hc_groups),
    cluster = hc_groups
  )

  df_clusters <- df_clusters[order(df_clusters$cluster), ]

  rownames(df_clusters) <- NULL

  out <- list(
    df = df_clusters,
    hclust = hc
  )

  out
}

testthat::test_that("`cor_clusters()` works", {
  testthat::skip_on_cran()
  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  out <- cor_clusters(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(out$df)
  )

  testthat::expect_true(
    all(c("predictor", "cluster") %in% colnames(out$df))
  )

  testthat::expect_true(
    class(out$df$predictor) == "character"
  )

  testthat::expect_true(
    class(out$df$cluster) == "integer"
  )

  testthat::expect_true(
    max(out$df$cluster) < nrow(vi_smol)
  )

  testthat::expect_true(
    inherits(x = out$hclust, what = "hclust")
  )
})

#' Find logical variables in a dataframe
#'
#' @description
#' Identifies logical predictors and excludes those with constant values.
#'
#' @inheritParams identify_valid_variables
#' @return list:
#' \itemize{
#'   \item \code{valid}: character vector with valid logical predictor names.
#'   \item \code{invalid}: character vector with invalid logical predictor names.
#' }
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' #invalid logical
#' vi_smol$logical_invalid <- TRUE
#'
#' #valid logical
#' vi_smol$logical_valid <- sample(
#'   x = c(TRUE, FALSE),
#'   size = nrow(vi_smol),
#'   replace = TRUE
#' )
#'
#' x <- identify_logical_variables(
#'   df = vi_smol,
#'   predictors = c(
#'     vi_predictors,
#'     "logical_invalid",
#'     "logical_valid"
#'   )
#' )
#'
#' x$valid
#' x$invalid
#'
#' @autoglobal
#' @family data_types
#' @author Blas M. Benito, PhD
#' @export
identify_logical_variables <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_logical_variables()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  if (!is.null(responses)) {
    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  vars_string <- if (!is.null(predictors) && !is.null(responses)) {
    "variables"
  } else if (!is.null(predictors)) {
    "predictors"
  } else if (!is.null(responses)) {
    "responses"
  } else {
    "variables"
  }

  predictors <- c(responses, predictors)

  if (is.null(predictors) || length(predictors) == 0) {
    stop(
      "\n",
      function_name,
      ": there are no ",
      vars_string,
      " to identify.",
      call. = FALSE
    )
  }

  out_list <- list(
    valid = NULL,
    invalid = NULL
  )

  # Get logical predictors
  predictors_logical_all <- predictors[
    vapply(
      X = df[, predictors, drop = FALSE],
      FUN = is.logical,
      FUN.VALUE = logical(1)
    )
  ] |>
    stats::na.omit()

  if (length(predictors_logical_all) == 0) {
    return(out_list)
  }

  #keep predictors with unique length different than one or nrow(df)
  length_unique <- vapply(
    X = df[, predictors_logical_all, drop = FALSE],
    FUN = function(x) {
      length(unique(x))
    },
    FUN.VALUE = integer(1)
  )

  predictors_logical_valid <- predictors_logical_all[length_unique > 1]

  predictors_logical_invalid <- setdiff(
    x = predictors_logical_all,
    y = predictors_logical_valid
  )

  if (
    quiet == FALSE &&
      length(predictors_logical_invalid) > 0
  ) {
    message(
      "\n",
      function_name,
      ": invalid logical ",
      vars_string,
      " due to constant values:\n - ",
      paste(
        predictors_logical_invalid,
        collapse = "\n - "
      )
    )
  }

  if (length(predictors_logical_valid) > 0) {
    out_list$valid <- predictors_logical_valid
  }

  if (length(predictors_logical_invalid) > 0) {
    out_list$invalid <- predictors_logical_invalid
  }

  out_list
}

testthat::test_that("`identify_logical_variables()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #invalid logical
  df <- vi_smol
  df$logical_invalid <- TRUE

  #valid logical
  df$logical_valid <- sample(
    x = c(TRUE, FALSE),
    size = nrow(vi_smol),
    replace = TRUE
  )

  testthat::expect_message(
    x <- identify_logical_variables(
      df = df,
      predictors = c(
        vi_predictors,
        "logical_invalid",
        "logical_valid"
      )
    ),
    regexp = "invalid logical predictors"
  )

  testthat::expect_true(
    "logical_invalid" == x$invalid
  )

  testthat::expect_true(
    "logical_valid" == x$valid
  )

  x <- identify_logical_variables(
    df = vi_smol,
    predictors = vi_predictors_numeric
  )

  testthat::expect_true(
    all(is.null(x$valid), is.null(x$invalid))
  )

  ##
  testthat::expect_message(
    x <- identify_logical_variables(
      df = df,
      responses <- c(
        "logical_invalid",
        "logical_valid"
      ),
      predictors = vi_predictors
    ),
    regexp = "invalid logical variables"
  )

  testthat::expect_true(
    "logical_invalid" == x$invalid
  )

  testthat::expect_true(
    "logical_valid" == x$valid
  )

  ##
  testthat::expect_message(
    x <- identify_logical_variables(
      df = df,
      responses <- c(
        "logical_invalid",
        "logical_valid"
      )
    ),
    regexp = "invalid logical responses"
  )

  testthat::expect_true(
    "logical_invalid" == x$invalid
  )

  testthat::expect_true(
    "logical_valid" == x$valid
  )

  ##
  testthat::expect_error(
    x <- identify_logical_variables(
      df = vi_smol
    ),
    regexp = "there are no variables to identify"
  )

  ##
  x <- identify_logical_variables(
    df = vi_smol,
    predictors = vi_predictors_numeric
  )

  testthat::expect_true(
    all(is.null(x$valid), is.null(x$invalid))
  )
})

#' Decision rules for `f_auto()`
#'
#' @description
#' Dataframe with rules used by [f_auto()] to select the function \code{f} in [f_functions()] to compute preference order in [preference_order()]. In most cases, random forest is selected as base model to provide homogeneous results across case types.
#'
#'
#' @return dataframe
#' @family preference_order_tools
#' @export
#' @autoglobal
#' @examples
#' f_auto_rules()
f_auto_rules <- function() {
  f_list <- list(
    c("f_numeric_glm", "continuous-binary", "numeric"),
    c("f_numeric_rf", "continuous-binary", "categorical"),
    c("f_numeric_rf", "continuous-binary", "mixed"),
    c("f_numeric_glm", "continuous-low", "numeric"),
    c("f_numeric_rf", "continuous-low", "categorical"),
    c("f_numeric_rf", "continuous-low", "mixed"),
    c("f_numeric_glm", "continuous-high", "numeric"),
    c("f_numeric_rf", "continuous-high", "categorical"),
    c("f_numeric_rf", "continuous-high", "mixed"),
    c("f_binomial_glm", "integer-binomial", "numeric"),
    c("f_binomial_rf", "integer-binomial", "categorical"),
    c("f_binomial_rf", "integer-binomial", "mixed"),
    c("f_count_glm", "integer-binary", "numeric"),
    c("f_count_rf", "integer-binary", "categorical"),
    c("f_count_rf", "integer-binary", "mixed"),
    c("f_count_glm", "integer-low", "numeric"),
    c("f_count_rf", "integer-low", "categorical"),
    c("f_count_rf", "integer-low", "mixed"),
    c("f_count_glm", "integer-high", "numeric"),
    c("f_count_rf", "integer-high", "categorical"),
    c("f_count_rf", "integer-high", "mixed"),
    c("f_categorical_rf", "categorical", "numeric"),
    c("f_categorical_rf", "categorical", "mixed"),
    c("f_categorical_rf", "categorical", "categorical")
  )

  f_df <- f_list |>
    as.data.frame() |>
    t() |>
    as.data.frame()

  rownames(f_df) <- NULL

  colnames(f_df) <- c(
    "name",
    "response_type",
    "predictors_type"
  )

  f_df
}

testthat::test_that("`f_auto_rules()` works", {
  testthat::skip_on_cran()

  df <- f_auto_rules()

  testthat::expect_true(
    is.data.frame(df)
  )

  testthat::expect_true(
    all(df$name %in% ls(getNamespace("collinear"), all.names = TRUE))
  )
})

#' Check and validate argument \code{f}
#'
#' @inheritParams preference_order
#' @inheritParams validate_arg_df
#' @param f_name (optional, string) Name of the function \code{f}, as returned by \code{deparse(substitute(f))}. Default: NULL
#'
#' @return function
#' @autoglobal
#' @family argument_validation
#' @export
#' @examples
#' x <- validate_arg_f(f = f_auto)
validate_arg_f <- function(
  f = NULL,
  f_name = NULL,
  function_name = NULL
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_f()",
    function_name = function_name
  )

  #NULL
  if (is.null(f)) {
    return(NULL)
  }

  #valid
  if (isTRUE(attr(x = f, which = "validated"))) {
    return(f)
  }

  #is function
  if (!is.function(f)) {
    stop(
      "\n",
      function_name,
      ": argument 'f' must be a uquoted function name without parentheses.",
      call. = FALSE
    )
  }

  #has 'df' argument
  f_args <- names(formals(f))
  if (!"df" %in% f_args) {
    stop(
      "\n",
      function_name,
      ": the function 'f' must have the argument 'df'.",
      call. = FALSE
    )
  }

  #attributes
  if (!is.null(f_name)) {
    attr(x = f, which = "name") <- f_name
  }

  attr(x = f, which = "validated") <- TRUE

  f
}

testthat::test_that("`validate_arg_quiet()` works", {
  testthat::skip_on_cran()

  x <- validate_arg_f(
    f = f_auto,
    f_name = "f_auto"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    attributes(x)$name == "f_auto"
  )

  #pass validated
  x <- validate_arg_f(
    f = x
  )

  #pass NULL
  x <- validate_arg_f(
    f = NULL
  )

  testthat::expect_true(
    is.null(x)
  )

  #pass non-function
  testthat::expect_error(
    x <- validate_arg_f(
      f = "hola"
    ),
    regexp = "argument 'f' must be a uquoted function name without parentheses"
  )

  x <- validate_arg_f(f = f_count_rf)

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_error(
    x <- validate_arg_f(f = stats::lm),
    regexp = "the function 'f' must have the argument 'df'"
  )
})

#' Compute area under the ROC curve between binomial observations and probabilistic predictions
#'
#' @description
#' Internal function to compute the AUC of binomial models within [preference_order()]. Used within [f_binomial_glm()], [f_binomial_gam()], and [f_binomial_rf()]. This function is build for speed and it does not check the inputs.
#'
#' @param o (required, numeric vector) Binomial observations (values 0 and 1). Default: NULL
#' @param p (required, numeric vector) Prediction of binomial model in the range 0-1. Default: NULL
#' @inheritParams collinear
#'
#' @return numeric: Area Under the ROC Curve
#' @export
#' @family modelling_tools
#' @autoglobal
#' @examples
#'   score_auc(
#'     o = c(1, 1, 1, 1, 0, 0, 0),
#'     p = c(1, 0.8, 0.7, 0.6, 0.5, 0.6, 0.7)
#'   )
#'
score_auc <- function(
  o = NULL,
  p = NULL,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::score_auc()",
    ... = ...
  )

  if (is.null(o)) {
    stop(
      "\n",
      function_name,
      ": argument 'o' cannot be NULL.",
      call. = FALSE
    )
  }

  if (is.null(p)) {
    stop(
      "\n",
      function_name,
      ": argument 'p' cannot be NULL.",
      call. = FALSE
    )
  }

  # remove incomplete cases
  complete <- !is.na(o) & !is.na(p)
  o <- o[complete]
  p <- p[complete]

  #predicted values of the ones and the zeroes
  ones <- p[o == 1]
  zeros <- p[o == 0]

  if (sum(ones) == 0) {
    stop(
      "\n",
      function_name,
      ": argument 'o' must be a binomial vector of 0s and 1s.",
      call. = FALSE
    )
  }

  #lengths of each vector
  ones.n <- length(ones)
  zeros.n <- length(zeros)

  #curve computation
  curve <- sum(
    rank(c(ones, zeros))[1:ones.n]
  ) -
    (ones.n * (ones.n + 1) / 2)

  #area under the curve
  out <- curve / (zeros.n * ones.n)

  out
}

testthat::test_that("`score_auc()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  #perfect prediction
  x <- score_auc(
    o = vi_smol$vi_binomial,
    p = vi_smol$vi_numeric
  )

  testthat::expect_equal(x, 1, tolerance = 1e-10)

  #random prediction
  x <- score_auc(
    o = vi_smol$vi_binomial,
    p = runif(n = nrow(vi_smol))
  )

  testthat::expect_true(
    x < 1
  )

  #NA in input vectors
  o <- vi_smol$vi_binomial
  p <- vi_smol$vi_numeric
  o[1:5] <- NA
  p[10:15] <- NA

  x <- score_auc(
    o = o,
    p = p
  )

  #error
  testthat::expect_error(
    x <- score_auc(
      o = runif(n = nrow(vi_smol)),
      p = vi_smol$vi_binomial
    )
  )

  testthat::expect_error(
    x <- score_auc(
      o = NULL,
      p = vi_smol$vi_binomial
    )
  )

  testthat::expect_error(
    x <- score_auc(
      o = vi_smol$vi_binomial,
      p = NULL
    )
  )

  # NA handling (issue #19)
  o_with_na <- vi_smol$vi_binomial
  o_with_na[1] <- NA

  x <- score_auc(
    o = o_with_na,
    p = vi_smol$vi_numeric
  )
  testthat::expect_true(is.numeric(x))

  p_with_na <- vi_smol$vi_numeric
  p_with_na[1] <- NA

  x <- score_auc(
    o = vi_smol$vi_binomial,
    p = p_with_na
  )
  testthat::expect_true(is.numeric(x))

  o_with_na[2] <- NA
  p_with_na[2] <- NA

  x <- score_auc(
    o = o_with_na,
    p = p_with_na
  )
  testthat::expect_true(is.numeric(x))
})

#' Summarize single response selection results of `collinear`
#'
#' @param object (sub-list in output of [collinear()] ) Object to summarize. Default: NULL
#' @param ... Ignored, kept for consistency with generic.
#' @return list: response name and character vector of selected predictors.
#' @method summary collinear_selection
#' @family S3_methods
#' @autoglobal
#' @export
summary.collinear_selection <- function(
  object = NULL,
  ...
) {
  out <- object$selection

  # response ----
  if (!is.null(object$response)) {
    msg <- paste0(
      "response: ",
      object$response
    )

    msg_length <- nchar(msg)

    underline <- paste0(rep(x = "-", times = nchar(msg)), collapse = "")

    cat(
      msg,
      fill = TRUE
    )

    cat(underline)

    cat("\n")
  }

  # selection ----
  if (length(out) > 1) {
    cat(
      "+ selection:\n  -",
      paste(
        out,
        collapse = "\n  - "
      )
    )
  } else {
    cat(
      "- selection:\n  -",
      out
    )
  }

  cat("\n")

  out
}

testthat::test_that("`summary.collinear_selection()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  x <- collinear(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric,
    preference_order = NULL,
    quiet = TRUE
  )

  testthat::expect_output(
    print(x),
    regexp = "vi_numeric"
  )

  x <- collinear(
    df = vi_smol,
    response = c("vi_numeric", "vi_categorical"),
    predictors = vi_predictors_numeric,
    preference_order = NULL,
    quiet = TRUE
  )

  testthat::expect_output(
    print(x),
    regexp = "vi_numeric"
  )

  testthat::expect_output(
    print(x),
    regexp = "vi_categorical"
  )

  testthat::expect_output(
    print(x),
    regexp = "selection"
  )

  testthat::expect_output(
    print(x),
    regexp = "formulas"
  )

  testthat::expect_output(
    print(x),
    regexp = "df"
  )

  testthat::expect_output(
    print(x),
    regexp = "preference order"
  )

  x <- collinear(
    df = vi_smol,
    response = "vi_categorical",
    predictors = vi_predictors_numeric,
    preference_order = NULL,
    quiet = TRUE
  )

  testthat::expect_output(
    print(x$vi_categorical),
    regexp = "formula"
  )

  #number of terms
  x$vi_categorical$selection <- x$vi_categorical$selection[1:6]

  testthat::expect_output(
    print(x$vi_categorical),
    regexp = "formula"
  )
})

testthat::test_that("`summary.collinear_selection()` works", {
  testthat::skip_on_cran()

  library(collinear)

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  x <- collinear(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric,
    f = f_auto,
    quiet = TRUE
  )

  testthat::expect_output(
    y <- summary(object = x),
    regexp = "response: "
  )

  testthat::expect_true(
    names(y) == "vi_numeric"
  )

  testthat::expect_true(
    all(x$vi_numeric$selection %in% y$vi_numeric)
  )
})

testthat::test_that("`summary.collinear_selection()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #normal usage
  x <- collinear(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  testthat::expect_output(
    z <- summary(x),
    regexp = "response"
  )

  testthat::expect_true(
    all(x$selection %in% z)
  )
})

#' Check and validate argument `predictors`
#'
#' @description
#' Validates the argument \code{predictors} by ensuring that all provided predictors are in \code{df} and don't intersect with \code{responses}, if any.
#'
#' @inheritParams collinear
#' @inheritParams validate_arg_quiet
#'
#' @return character vector: predictor names
#' @examples
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' x <- validate_arg_predictors(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )
#'
#' attributes(x)$validated
#'
#' @autoglobal
#' @family argument_validation
#' @export
validate_arg_predictors <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  #if already validated, return it
  if (
    isTRUE(attr(x = predictors, which = "validated")) &&
      all(predictors %in% colnames(df))
  ) {
    return(predictors)
  }

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_predictors()",
    function_name = function_name
  )

  #if df is NULL, stop
  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #if predictors is NULL, use colnames(df)
  if (is.null(predictors)) {
    predictors <- setdiff(
      x = colnames(df),
      y = responses
    )
  }

  #remove response from predictors
  predictors <- setdiff(
    x = predictors,
    y = responses
  )

  #identify wrongly named predictors
  predictors.missing <- setdiff(
    x = predictors,
    y = colnames(df)
  )

  if (length(predictors.missing) > 0) {
    if (length(predictors.missing) == length(predictors)) {
      warning(
        "\n",
        function_name,
        ": none of the 'predictors' are column names of 'df'.",
        call. = FALSE
      )

      return(NULL)
    }

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": these 'predictors' are not column names of 'df' and will be ignored:\n - ",
        paste(
          predictors.missing,
          collapse = "\n - "
        )
      )
    }
  }

  #getting predictors in df only
  predictors <- intersect(
    x = predictors,
    y = colnames(df)
  )

  attr(
    x = predictors,
    which = "validated"
  ) <- TRUE

  predictors
}

testthat::test_that("`validate_arg_predictors()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #no arguments
  testthat::expect_error(
    predictors <- validate_arg_predictors(
      df = NULL,
      responses = NULL,
      predictors = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #without predictors
  #without response
  #must contain only numeric columns
  testthat::expect_no_message(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = NULL,
      predictors = NULL,
      quiet = FALSE
    )
  )

  testthat::expect_true(
    attributes(predictors)$validated
  )

  testthat::expect_true(
    length(predictors) == ncol(vi_smol)
  )

  testthat::expect_true(
    all(predictors %in% colnames(vi_smol))
  )

  #validated predictors from previous test
  testthat::expect_no_message(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = NULL,
      predictors = predictors,
      quiet = FALSE
    )
  )

  testthat::expect_true(
    attributes(predictors)$validated
  )

  testthat::expect_true(
    length(predictors) == ncol(vi_smol)
  )

  testthat::expect_true(
    all(predictors %in% colnames(vi_smol))
  )

  #wrong predictors
  testthat::expect_warning(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = NULL,
      predictors = c("hola", "adios"),
      quiet = FALSE
    ),
    regexp = " none of the 'predictors' are column names of 'df'"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.null(predictors)
  )

  #wrong and right predictors
  testthat::expect_message(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = NULL,
      predictors = c("hola", "adios", vi_predictors),
      quiet = FALSE
    ),
    regexp = "these 'predictors' are not column names of 'df' and will be ignored"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.character(predictors)
  )

  testthat::expect_true(
    all(predictors %in% colnames(vi_smol))
  )

  #without predictors
  #with response
  #must contain all df columns but the response
  testthat::expect_no_message(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = "vi_numeric",
      quiet = FALSE
    )
  )

  testthat::expect_true(
    !("vi_numeric" %in% predictors)
  )

  #with predictors
  #with response
  #must contain all predictors
  testthat::expect_no_message(
    predictors <- validate_arg_predictors(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors,
      quiet = FALSE
    )
  )

  testthat::expect_true(
    all(predictors %in% vi_predictors)
  )
})

#' Find valid categorical variables in a dataframe
#'
#' @description
#' Identifies valid and invalid character or factor variables. Invalid categorical predictors are those with a single category, or as many categories as cases (full-cardinality).
#'
#'
#' @inheritParams identify_valid_variables
#' @return list:
#' \itemize{
#'   \item \code{valid}: character vector with valid categorical predictor names.
#'   \item \code{invalid}: character vector with invalid categorical predictor names due to degenerate cardinality (1 or \code{nrow(df)} categories).
#' }
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' #create an invalid categorical
#' vi_smol$invalid_categorical <- "a"
#'
#' x <- identify_categorical_variables(
#'   df = vi_smol,
#'   responses = "vi_categorical",
#'   predictors = vi_predictors
#' )
#'
#' x$valid
#' x$invalid
#'
#' @autoglobal
#' @family data_types
#' @author Blas M. Benito, PhD
#' @export
identify_categorical_variables <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_categorical_variables()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  if (!is.null(responses)) {
    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  vars_string <- if (!is.null(predictors) && !is.null(responses)) {
    "variables"
  } else if (!is.null(predictors)) {
    "predictors"
  } else if (!is.null(responses)) {
    "responses"
  } else {
    "variables"
  }

  predictors <- c(responses, predictors)

  if (is.null(predictors) || length(predictors) == 0) {
    stop(
      "\n",
      function_name,
      ": there are no ",
      vars_string,
      " to identify.",
      call. = FALSE
    )
  }

  out_list <- list(
    valid = NULL,
    invalid = NULL
  )

  #subset character or factor
  predictors_categorical_all <- predictors[
    vapply(
      X = df[, predictors, drop = FALSE],
      FUN = function(x) {
        is.character(x) || is.factor(x)
      },
      FUN.VALUE = logical(1)
    )
  ] |>
    stats::na.omit()

  if (length(predictors_categorical_all) == 0) {
    return(out_list)
  }

  n <- nrow(df)

  #keep predictors with unique length different than one or nrow(df)
  length_unique <- vapply(
    X = df[, predictors_categorical_all, drop = FALSE],
    FUN = function(x) {
      length(unique(x))
    },
    FUN.VALUE = integer(1)
  )

  predictors_categorical_valid <- predictors_categorical_all[
    !(length_unique %in% c(1, n))
  ]

  predictors_categorical_invalid <- setdiff(
    x = predictors_categorical_all,
    y = predictors_categorical_valid
  )

  if (
    quiet == FALSE &&
      length(predictors_categorical_invalid) > 0
  ) {
    message(
      "\n",
      function_name,
      ": invalid categorical ",
      vars_string,
      " due to degenerate cardinality:\n - ",
      paste(
        predictors_categorical_invalid,
        collapse = "\n - "
      )
    )
  }

  if (length(predictors_categorical_valid) > 0) {
    out_list$valid <- predictors_categorical_valid
  }

  if (length(predictors_categorical_invalid) > 0) {
    out_list$invalid <- predictors_categorical_invalid
  }

  out_list
}

testthat::test_that("`cor_select()` works", {
  testthat::skip_on_cran()
  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  # mixed types ----
  testthat::expect_message(
    x <- cor_select(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      quiet = FALSE
    ),
    regexp = "from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors[1:10])
  )

  testthat::expect_true(
    length(vi_predictors[1:10]) > length(x)
  )

  #custom preference order
  preference_order <- c(
    "swi_mean",
    "topo_elevation"
  )

  testthat::expect_message(
    x <- cor_select(
      df = vi_smol,
      predictors = vi_predictors_numeric,
      preference_order = preference_order,
      quiet = FALSE
    ),
    regexp = paste0(
      "ranking ",
      length(vi_predictors_numeric) - length(preference_order),
      " 'predictors' from lower to higher multicollinearity"
    )
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors_numeric)
  )

  testthat::expect_true(
    length(vi_predictors_numeric) > length(x)
  )

  testthat::expect_true(
    preference_order[1] == x[1]
  )

  #automated preference order
  preference_order_df <- preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  x <- cor_select(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    preference_order = preference_order_df,
    quiet = TRUE
  )

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors_numeric)
  )

  testthat::expect_true(
    length(vi_predictors_numeric) > length(x)
  )

  testthat::expect_true(
    all(preference_order_df$predictor[1] == x[1])
  )

  # categorical only ----
  x <- cor_select(
    df = vi_smol,
    predictors = vi_predictors_categorical[1:4],
    quiet = TRUE
  )

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors_categorical[1:4])
  )

  testthat::expect_true(
    length(vi_predictors_categorical[1:4]) > length(x)
  )

  testthat::expect_message(
    y <- cor_select(
      df = vi_smol,
      predictors = x,
      quiet = FALSE
    ),
    regexp = "maximum pairwise correlation is <="
  ) |>
    suppressMessages()

  # edge cases ----

  #no df
  testthat::expect_error(
    x <- cor_select(
      df = NULL,
      predictors = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #predictors only
  testthat::expect_error(
    x <- cor_select(
      df = NULL,
      predictors = vi_predictors
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #few rows
  testthat::expect_error(
    x <- cor_select(
      df = vi_smol[1, ],
      predictors = vi_predictors
    ),
    regexp = "argument 'df' has fewer than 3 rows"
  )

  #null max cor
  testthat::expect_error(
    x <- cor_select(
      df = vi_smol[, 1:5],
      predictors = NULL,
      max_cor = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'max_cor' cannot be NULL"
  )

  #no predictors
  x <- cor_select(
    df = vi_smol[, 1:5],
    predictors = NULL,
    quiet = TRUE
  )

  testthat::expect_true(
    all(x %in% colnames(vi_smol)[1:5])
  )

  #single predictor
  testthat::expect_message(
    x <- cor_select(
      df = vi_smol,
      predictors = vi_predictors_categorical[1]
    ),
    regexp = "only one valid predictor"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors_categorical[1])
  )

  testthat::expect_true(
    length(vi_predictors_categorical[1]) == length(x)
  )
})

testthat::test_that("`collinear()` works", {
  testthat::skip_on_cran()
  #load data
  data(vi, vi_smol, vi_responses, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #DEFAULT CALL ----
  #Error: collinear::collinear(): argument 'df' cannot be NULL
  testthat::expect_error(
    x <- collinear(),
    regexp = "'df' cannot be NULL"
  )

  #DF ONLY ----

  ##fewer than 3 rows ----
  testthat::expect_error(
    x <- collinear(
      df = vi_smol[1:2, ],
      predictors = vi_predictors_numeric
    ),
    regexp = "has fewer than 3 rows"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  ##fewer than 10 rows ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol[1:9, ],
      predictors = vi_predictors_numeric
    ),
    regexp = "has fewer than 10 rows"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  x <- collinear(
    df = vi_smol[1:9, ],
    predictors = vi_predictors_numeric,
    max_cor = 0.7,
    max_vif = NULL,
    quiet = TRUE
  )

  ##fewer than 30 rows ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol[1:29, ],
      predictors = vi_predictors_numeric
    ),
    regexp = "has fewer than 30 rows"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  ##more than 30 rows ----

  #max_cor and max_vif
  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      predictors = vi_predictors_numeric,
      max_cor = NULL,
      max_vif = NULL
    ),
    regexp = "setting 'max_cor' to"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      predictors = vi_predictors_numeric,
      max_cor = NULL,
      max_vif = NULL
    ),
    regexp = "setting 'max_vif' to"
  ) |>
    suppressMessages()

  #max_cor and max_vif invalid
  f_test <- function() {
    collinear(
      df = vi_smol,
      predictors = vi_predictors_numeric,
      max_cor = 2,
      max_vif = 20
    )
  }

  testthat::expect_message(
    x <- f_test(),
    regexp = "argument 'max_cor' is outside its valid range"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- f_test(),
    regexp = "argument 'max_vif' is outside its valid range"
  ) |>
    suppressMessages()

  #RESPONSE ----

  ##response only ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol[, c(vi_responses[1:2], vi_predictors_numeric)],
      responses = vi_responses[1:2],
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "processing response 'vi_numeric'"
  ) |>
    suppressMessages()

  testthat::expect_true(
    all(vi_responses[1:2] %in% names(x))
  )

  #check that the given response is never in selections
  for (i in vi_responses[1:2]) {
    testthat::expect_true(
      !i %in% x[[i]]$selection
    )
  }

  #PREDICTORS ----

  ##numeric predictors ----
  testthat::expect_message(
    x <- collinear(
      df = vi,
      responses = NULL,
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "selected predictors"
  ) |>
    suppressMessages()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  #test lack of formulas because there is no response
  testthat::expect_true(
    !any(c("formula", "formulas") %in% names(x))
  )

  ##categorical predictors ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = NULL,
      predictors = vi_predictors_categorical[1:4],
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  #test lack of formulas because there is no response
  testthat::expect_true(
    !any(c("formula", "formulas") %in% names(x))
  )

  ##mixed predictors ----
  x <- collinear(
    df = vi_smol,
    responses = NULL,
    predictors = c(vi_predictors_numeric[1:3], vi_predictors_categorical[1:2]),
    encoding_method = NULL,
    preference_order = NULL,
    f = NULL,
    quiet = TRUE
  ) |>
    suppressWarnings()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  #test lack of formulas because there is no response
  testthat::expect_true(
    !any(c("formula", "formulas") %in% names(x))
  )

  #PREDICTORS + RESPONSE ----

  ##numeric numeric ----
  testthat::expect_message(
    x <- collinear(
      df = vi,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  testthat::expect_true(
    all(c("linear", "smooth") %in% names(x$vi_numeric$formulas))
  )

  ##categorical numeric ----
  testthat::expect_message(
    x <- collinear(
      df = vi,
      responses = "vi_categorical",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  ##categorical categorical ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = "vi_categorical",
      predictors = vi_predictors_categorical[1:4],
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    inherits(x = x, what = "collinear_output")
  )

  ##multiple responses ----
  #all selections are the same
  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = vi_responses,
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "processing response"
  ) |>
    suppressMessages()

  #check that the given response does not appear in the selections
  for (i in names(x)) {
    testthat::expect_true(
      !i %in% x[[i]]$selection
    )
  }

  #checking formulas
  testthat::expect_true(
    all(c("linear", "smooth") %in% names(x$vi_numeric$formulas))
  )

  testthat::expect_true(
    all(c("linear", "smooth") %in% names(x$vi_counts$formulas))
  )

  testthat::expect_true(
    all(c("linear", "smooth") %in% names(x$vi_binomial$formulas))
  )

  testthat::expect_true(
    all(c("classification") %in% names(x$vi_factor$formulas))
  )

  testthat::expect_true(
    all(c("classification") %in% names(x$vi_categorical$formulas))
  )

  #TARGET ENCODING ----
  f_test <- function() {
    collinear(
      df = vi_smol,
      responses = c("vi_numeric", "vi_categorical"),
      predictors = vi_predictors_categorical[1:4],
      encoding_method = "loo",
      preference_order = NULL,
      f = NULL,
      quiet = FALSE
    )
  }

  testthat::expect_message(
    x <- f_test(),
    regexp = "using response 'vi_numeric' to encode these categorical predictors"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- f_test(),
    regexp = "processing response 'vi_categorical'"
  ) |>
    suppressMessages()

  #check that a categorical predictor was converted to numeric for "vi_numeric"
  testthat::expect_true(
    !is.numeric(vi_smol[["soil_type"]]) &&
      !is.numeric(x$vi_categorical$df[["soil_type"]]) &&
      is.numeric(x$vi_numeric$df[["soil_type"]])
  )

  # PREFERENCE ORDER ----

  ## no target encoding ----

  ### invalid character vector ----
  preference_order <- c(
    "hola",
    "adios"
  )

  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = preference_order,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "invalid values in argument 'preference_order'"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = preference_order,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    !all(preference_order %in% x$vi_numeric$selection)
  )

  testthat::expect_true(
    all(vi_predictors_numeric %in% x$vi_numeric$preference_order$predictor)
  )

  ### valid character vector ----

  #uncorrelated, must show up in selection
  preference_order <- c(
    "soil_soc",
    "growing_season_temperature"
  )

  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = preference_order,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    all(preference_order %in% x$vi_numeric$selection)
  )

  testthat::expect_true(
    all(vi_predictors_numeric %in% x$vi_numeric$preference_order$predictor)
  )

  ### valid dataframe ----
  preference_df <- preference_order(
    df = vi_smol,
    responses = c("vi_numeric", "vi_categorical"),
    predictors = vi_predictors_numeric,
    f = f_auto,
    quiet = TRUE
  )

  x <- collinear(
    df = vi_smol,
    responses = c("vi_numeric", "vi_categorical"),
    predictors = vi_predictors_numeric,
    encoding_method = NULL,
    preference_order = preference_df,
    f = NULL,
    max_cor = 0.75,
    max_vif = 5,
    quiet = TRUE
  )

  testthat::expect_true(
    unique(x$vi_categorical$preference_order$response) == "vi_categorical"
  )

  ### invalid data frame ----
  preference_df <- data.frame(
    response = "hola",
    f = NA
  )

  testthat::expect_error(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = preference_df,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = TRUE
    ),
    regexp = "dataframe 'preference_order' must have these columns"
  )

  ### NULL response ----
  #same as above!
  preference <- c(
    "soil_soc",
    "growing_season_temperature"
  )

  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = NULL,
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = preference,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    all(preference %in% x$result$selection)
  )

  testthat::expect_true(
    preference[1] == x$result$preference_order$predictor[1]
  )

  testthat::expect_true(
    preference[2] == x$result$preference_order$predictor[2]
  )

  ### f_auto ----
  testthat::expect_message(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = f_auto,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "selected function 'f_numeric_glm"
  ) |>
    suppressMessages()

  testthat::expect_true(
    x$vi_numeric$selection[1] == x$vi_numeric$preference$predictor[1]
  )

  testthat::expect_true(
    x$vi_numeric$preference$f[1] == "f_numeric_glm"
  )

  ### f_numeric_rf ----
  x <- collinear(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = vi_predictors_numeric,
    encoding_method = NULL,
    preference_order = NULL,
    f = f_numeric_rf,
    max_cor = 0.75,
    max_vif = 5,
    quiet = TRUE
  )

  testthat::expect_true(
    x$vi_numeric$selection[1] == x$vi_numeric$preference$predictor[1]
  )

  testthat::expect_true(
    x$vi_numeric$preference_order$f[1] == "f_numeric_rf"
  )

  ### bad function name ----
  my_f <- function() {
    return(NULL)
  }

  testthat::expect_error(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = my_f,
      max_cor = 0.75,
      max_vif = 5,
      quiet = TRUE
    ),
    regexp = "the function 'f' must have the argument 'df'"
  )

  ### character function name ----
  testthat::expect_error(
    x <- collinear(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = vi_predictors_numeric,
      encoding_method = NULL,
      preference_order = NULL,
      f = "my_f",
      max_cor = 0.75,
      max_vif = 5,
      quiet = TRUE
    ),
    regexp = "must be a uquoted function name"
  )

  #passing a correlation matrix
  time_no_matrix <- system.time(
    expr = {
      x <- collinear(
        df = vi_smol,
        responses = "vi_numeric",
        predictors = vi_predictors_categorical[1:10],
        encoding_method = NULL,
        preference_order = NULL,
        f = NULL,
        max_cor = 0.75,
        max_vif = 5,
        quiet = TRUE
      )
    }
  )

  m <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_categorical[1:10],
    quiet = TRUE
  )

  time_matrix <- system.time(
    expr = {
      x <- collinear(
        df = vi_smol,
        responses = "vi_numeric",
        predictors = vi_predictors_categorical[1:10],
        encoding_method = NULL,
        preference_order = NULL,
        f = NULL,
        max_cor = 0.75,
        max_vif = 5,
        quiet = TRUE,
        m = m
      )
    }
  )

  testthat::expect_true(
    time_no_matrix[3] > time_matrix[3]
  )

  #losing one column
  df <- vi_smol
  df$logical <- TRUE

  testthat::expect_message(
    x <- collinear(
      df = df,
      responses = "vi_numeric",
      predictors = c("logical", vi_predictors_numeric),
      encoding_method = NULL,
      preference_order = NULL,
      f = NULL,
      max_cor = 0.75,
      max_vif = 5,
      quiet = FALSE
    ),
    regexp = "invalid logical variables due to constant values"
  ) |>
    suppressMessages()

  testthat::expect_true(
    !"logical" %in% colnames(x$vi_numeric$df)
  )

  #cross validation
  time_no_cv <- system.time(
    expr = {
      x <- collinear(
        df = vi_smol,
        responses = "vi_numeric",
        predictors = vi_predictors_numeric,
        encoding_method = NULL,
        preference_order = NULL,
        f = f_auto,
        max_cor = 0.75,
        max_vif = 5,
        quiet = TRUE
      )
    }
  )

  time_cv <- system.time(
    expr = {
      x <- collinear(
        df = vi_smol,
        responses = "vi_numeric",
        predictors = vi_predictors_numeric,
        encoding_method = NULL,
        preference_order = NULL,
        f = f_auto,
        max_cor = 0.75,
        max_vif = 5,
        quiet = TRUE,
        cv_iterations = 100,
        cv_training_fraction = 0.5
      )
    }
  )

  testthat::expect_true(
    time_cv[3] > time_no_cv[3]
  )
})

testthat::test_that("`identify_valid_variables()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  x <- identify_valid_variables(
    df = vi_smol,
    predictors = vi_predictors
  )

  testthat::expect_true(
    all(x$numeric %in% vi_predictors_numeric)
  )

  testthat::expect_true(
    all(x$categorical %in% vi_predictors_categorical)
  )

  x <- identify_valid_variables(
    df = vi_smol,
    responses = vi_predictors
  )

  testthat::expect_true(
    all(x$numeric %in% vi_predictors_numeric)
  )

  testthat::expect_true(
    all(x$categorical %in% vi_predictors_categorical)
  )

  x <- identify_valid_variables(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = vi_predictors
  )

  testthat::expect_true(
    all(x$numeric %in% c("vi_numeric", vi_predictors_numeric))
  )

  testthat::expect_true(
    all(x$categorical %in% vi_predictors_categorical)
  )
})

testthat::test_that("`identify_zero_variance_variables()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  vi_smol$zv_1 <- 1

  vi_smol$zv_2 <- runif(
    n = nrow(vi_smol),
    min = 0,
    max = 1e-04
  )

  testthat::expect_message(
    x <- identify_zero_variance_variables(
      df = vi_smol,
      predictors = c(
        vi_predictors,
        "zv_1",
        "zv_2"
      )
    ),
    regexp = "invalid predictors due to near-zero variance"
  )

  testthat::expect_true(
    is.character(x),
  )

  testthat::expect_true(
    length(x) == 2
  )

  testthat::expect_true(
    all(c("zv_1", "zv_2") %in% x)
  )

  ##
  testthat::expect_message(
    x <- identify_zero_variance_variables(
      df = vi_smol,
      responses = c(
        vi_predictors,
        "zv_1",
        "zv_2"
      )
    ),
    regexp = "invalid responses due to near-zero variance"
  )

  testthat::expect_true(
    is.character(x),
  )

  testthat::expect_true(
    length(x) == 2
  )

  testthat::expect_true(
    all(c("zv_1", "zv_2") %in% x)
  )

  ##
  testthat::expect_message(
    x <- identify_zero_variance_variables(
      df = vi_smol,
      responses = c(
        "zv_1",
        "zv_2"
      ),
      predictors = vi_predictors
    ),
    regexp = "invalid variables due to near-zero variance"
  )

  testthat::expect_true(
    is.character(x),
  )

  testthat::expect_true(
    length(x) == 2
  )

  testthat::expect_true(
    all(c("zv_1", "zv_2") %in% x)
  )

  ##
  x <- identify_zero_variance_variables(
    df = vi_smol,
    predictors = vi_predictors_numeric
  )

  testthat::expect_true(
    is.null(x)
  )

  x <- identify_zero_variance_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #without predictors
  testthat::expect_error(
    x <- identify_zero_variance_variables(
      df = vi_smol
    ),
    regexp = "there are no variables to identify"
  )

  #edge cases
  testthat::expect_error(
    x <- identify_zero_variance_variables(
      df = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #non overlapping predictors?
  testthat::expect_error(
    x <- identify_zero_variance_variables(
      df = vi_smol,
      predictors = c("hola", "adios")
    ),
    regexp = "there are no variables to identify"
  ) |>
    suppressWarnings()

  x <- identify_zero_variance_variables(
    df = vi_smol,
    predictors = vi_predictors_categorical
  )

  testthat::expect_null(
    x
  )
})

testthat::test_that("`identify_categorical_variables()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  df <- vi_smol
  df$zero_cardinality <- "a"
  df$full_cardinality <- as.character(seq_len(nrow(df)))
  df$logical <- TRUE

  predictors <- c(
    vi_predictors,
    "zero_cardinality",
    "full_cardinality",
    "logical"
  )

  #identify all categoricals
  testthat::expect_message(
    x <- identify_categorical_variables(
      df = df,
      predictors = predictors,
      quiet = FALSE
    ),
    regexp = "invalid categorical predictors due to degenerate cardinality"
  )

  testthat::expect_true(
    is.list(x) && all(c("valid", "invalid") %in% names(x))
  )

  testthat::expect_true(
    all(c("zero_cardinality", "full_cardinality") %in% x$invalid)
  )

  testthat::expect_false(
    all(c("zero_cardinality", "full_cardinality") %in% x$valid)
  )

  testthat::expect_false(
    ("logical" %in% x$valid) && ("logical" %in% x$invalid)
  )

  ##
  #identify all categoricals
  testthat::expect_message(
    x <- identify_categorical_variables(
      df = df,
      responses = c("zero_cardinality", "full_cardinality"),
      predictors = vi_predictors_categorical,
      quiet = FALSE
    ),
    regexp = "invalid categorical variables due to degenerate cardinality"
  )

  testthat::expect_message(
    x <- identify_categorical_variables(
      df = df,
      responses = c("zero_cardinality", "full_cardinality"),
      quiet = FALSE
    ),
    regexp = "invalid categorical responses due to degenerate cardinality"
  )

  testthat::expect_true(
    is.list(x) && all(c("valid", "invalid") %in% names(x))
  )

  testthat::expect_true(
    all(c("zero_cardinality", "full_cardinality") %in% x$invalid)
  )

  testthat::expect_false(
    all(c("zero_cardinality", "full_cardinality") %in% x$valid)
  )

  testthat::expect_false(
    ("logical" %in% x$valid) && ("logical" %in% x$invalid)
  )

  #no predictors
  x <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = FALSE
  )

  testthat::expect_null(
    x$valid
  )

  testthat::expect_null(
    x$invalid
  )

  testthat::expect_error(
    x <- identify_categorical_variables(
      df = vi_smol,
      quiet = FALSE
    ),
    regexp = "there are no variables to identify"
  )

  testthat::expect_null(
    x$valid
  )

  testthat::expect_null(
    x$invalid
  )
})

testthat::test_that("`preference_order()` works", {
  testthat::skip_on_cran()

  expected_colnames <- c(
    "response",
    "predictor",
    "f",
    "metric",
    "score",
    "rank"
  )

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #several responses
  responses <- c(
    "vi_numeric",
    "vi_counts",
    "vi_binomial",
    "vi_categorical",
    "vi_factor"
  )

  #empty args

  testthat::expect_error(
    x <- preference_order(
      df = NULL,
      responses = NULL,
      predictors = NULL,
      f = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #default behavior without response and f
  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      responses = NULL,
      predictors = vi_predictors_numeric,
      f = NULL,
      quiet = FALSE
    ),
    regexp = "predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    unique(x$metric) == "1 - R-squared"
  )

  testthat::expect_true(
    unique(x$response) == "none"
  )

  #no predictors
  x <- preference_order(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = NULL,
    f = f_auto,
    quiet = TRUE
  )

  #all df column names but vi_numeric
  testthat::expect_true(
    sum(colnames(vi_smol) %in% x$predictor) == ncol(vi_smol) - 1
  )

  testthat::expect_true(
    !"vi_numeric" %in% x$predictor
  )

  testthat::expect_true(
    "vi_numeric" %in% x$response
  )

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors,
      f = f_auto,
      cv_iterations = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'cv_iterations' must be a positive integer."
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors,
      f = f_auto,
      cv_iterations = 1,
      cv_training_fraction = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'cv_training_fraction' must be a numeric between 0.1 and 1"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors,
      f = f_auto,
      cv_iterations = 1,
      cv_training_fraction = 1,
      seed = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'seed' is invalid, resetting it to '1'"
  ) |>
    suppressMessages()

  #full use case

  #test that one message per response is printed
  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = c("vi_numeric", "vi_binomial"),
      predictors = vi_predictors_numeric,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "processing response 'vi_numeric'"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = c("vi_numeric", "vi_binomial"),
      predictors = vi_predictors_numeric,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "processing response 'vi_binomial'"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    all(c("vi_numeric", "vi_binomial") %in% x$response)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  #numeric response

  ##all types
  x <- preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors,
    f = f_numeric_rf,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_numeric_rf"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "f_numeric_rf"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_numeric_rf"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  #count response

  ## all types
  x <- preference_order(
    df = vi_smol,
    response = "vi_counts",
    predictors = vi_predictors_numeric,
    f = f_count_glm,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_count_glm"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  #wrong f function
  testthat::expect_error(
    x <- preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric,
      f = f_categorical_rf,
      quiet = FALSE
    ),
    regexp = "column 'y' of dataframe 'df' must be character or factor"
  ) |>
    suppressMessages()

  #binomial response
  x <- preference_order(
    df = vi_smol,
    response = "vi_binomial",
    predictors = vi_predictors_numeric,
    f = f_binomial_glm,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_binomial_glm"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_binomial",
      predictors = vi_predictors,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "f_binomial_rf"
  ) |>
    suppressMessages()

  #categorical response and predictors
  x <- preference_order(
    df = vi_smol,
    response = "vi_categorical",
    predictors = vi_predictors_categorical,
    f = f_categorical_rf,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_categorical_rf"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_categorical)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_categorical",
      predictors = vi_predictors_categorical,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "f_categorical_rf"
  ) |>
    suppressMessages()

  #categorical response and categorical and numeric predictors
  x <- preference_order(
    df = vi_smol,
    response = "vi_categorical",
    predictors = vi_predictors_numeric,
    f = f_categorical_rf,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    unique(x$f) == "f_categorical_rf"
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric)
  )

  testthat::expect_true(
    all(colnames(x) %in% expected_colnames)
  )

  testthat::expect_message(
    x <- preference_order(
      df = vi_smol,
      response = "vi_categorical",
      predictors = vi_predictors_numeric,
      f = f_auto,
      quiet = FALSE
    ),
    regexp = "f_categorical_rf"
  ) |>
    suppressMessages()

  #custom function
  f_rsquared <- function(df, ...) {
    stats::cor(
      x = df$x,
      y = df$y,
      use = "complete.obs"
    )^2
  }

  x <- preference_order(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = vi_predictors_numeric[1:10],
    f = f_rsquared
  )

  testthat::expect_true(
    all(x$metric == "custom")
  )

  # namespace-qualified built-in should still resolve metric correctly (issue #16)
  x <- preference_order(
    df = vi_smol,
    responses = "vi_counts",
    predictors = vi_predictors_numeric[1:5],
    f = collinear::f_count_gam,
    quiet = TRUE
  ) |>
    suppressMessages()

  testthat::expect_false(
    unique(x$metric) == "custom"
  )

  # NA values in response column should not crash preference_order (issue #18)
  vi_smol_inf <- vi_smol
  vi_smol_inf$vi_numeric[1:3] <- Inf

  x <- preference_order(
    df = vi_smol_inf,
    responses = "vi_numeric",
    predictors = vi_predictors_numeric[1:5],
    f = f_numeric_glm,
    quiet = TRUE
  ) |>
    suppressMessages()

  testthat::expect_true(is.data.frame(x))
  testthat::expect_true(nrow(x) == 5)
})

testthat::test_that("`validate_arg_df()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #no arguments
  testthat::expect_error(
    x <- validate_arg_df(
      df = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      responses = NULL,
      predictors = NULL
    ),
    regexp = "arguments 'responses' and 'predictors' are NULL, skipping validation of column values"
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  #few rows
  testthat::expect_error(
    x <- validate_arg_df(
      df = vi_smol[1:2, ]
    ),
    regexp = "has fewer than 3 rows"
  )

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol[1:9, ]
    ),
    regexp = "has fewer than 10 rows"
  ) |>
    suppressMessages()

  testthat::expect_true(
    nrow(x) == 9
  )

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol[1:29, ],
      predictors = vi_predictors_numeric
    ),
    regexp = "has fewer than 30 rows"
  )

  testthat::expect_true(
    nrow(x) == 29
  )

  testthat::expect_error(
    x <- validate_arg_df(
      df = vi_smol[, 0],
      predictors = vi_predictors_numeric
    ),
    regexp = "argument 'df' has zero columns"
  )

  testthat::expect_error(
    x <- validate_arg_df(
      df = vi_smol[, vi_predictors_numeric[1], drop = FALSE],
      predictors = vi_predictors_numeric[1]
    ),
    regexp = "argument 'df' has one valid column, multicollinearity analysis cannot be performed"
  )

  testthat::expect_error(
    x <- validate_arg_df(
      df = stats::lm
    ),
    regexp = "cannot coerce argument 'df' to class 'data.frame'"
  ) |>
    suppressMessages()

  testthat::expect_error(
    x <- validate_arg_df(
      df = FALSE
    ),
    regexp = "cannot coerce argument 'df' to class 'data.frame'"
  ) |>
    suppressMessages()

  #logical columns
  vi_smol$logical <- sample(
    c(TRUE, FALSE),
    size = nrow(vi_smol),
    replace = TRUE
  )

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      responses = "vi_numeric",
      predictors = c(vi_predictors_numeric, "logical"),
      quiet = FALSE
    ),
    regexp = "converted the following logical columns to numeric"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.numeric(x$logical)
  )

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      responses = "hola",
      predictors = "adios",
      quiet = FALSE
    ),
    regexp = "argument 'responses' does not contain column names of 'df'"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      responses = "hola",
      predictors = "adios",
      quiet = FALSE
    ),
    regexp = "arguments 'responses' and 'predictors' are NULL, skipping validation of column values"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  testthat::expect_warning(
    x <- validate_arg_df(
      df = vi_smol,
      responses = "hola",
      predictors = "adios",
      quiet = FALSE
    ),
    regexp = "none of the 'predictors' are column names of 'df'"
  ) |>
    suppressMessages() |>
    suppressWarnings()

  x <- validate_arg_df(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = vi_predictors_numeric,
    quiet = FALSE
  )

  testthat::expect_true(attributes(x)$validated)

  testthat::expect_true(
    all(c("vi_numeric", vi_predictors_numeric) %in% colnames(x))
  )

  #pass a validated data frame
  x <- validate_arg_df(
    df = x
  )

  testthat::expect_true(attributes(x)$validated)

  #only one predictor
  x <- validate_arg_df(
    df = vi_smol,
    responses = NULL,
    predictors = vi_predictors_numeric[1],
    quiet = FALSE
  )

  testthat::expect_true(attributes(x)$validated)

  testthat::expect_true(
    all(vi_predictors_numeric[1] %in% colnames(x))
  )

  #zero variance predictors
  vi_smol$zero_variance <- 1

  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      responses = NULL,
      predictors = c(vi_predictors_numeric, "zero_variance"),
      quiet = FALSE
    ),
    regexp = "invalid numeric predictors due to near-zero variance"
  ) |>
    suppressMessages()

  testthat::expect_false(
    "zero_variance" %in% colnames(x)
  )

  #categorical
  testthat::expect_message(
    x <- validate_arg_df(
      df = vi_smol,
      predictors = vi_predictors_categorical
    ),
    regexp = "converted the following character columns to factor"
  )
})

testthat::test_that("`cor_df()` works", {
  testthat::skip_on_cran()
  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #mixed types
  testthat::expect_warning(
    x <- cor_df(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      quiet = TRUE
    ),
    regexp = "may bias the multicollinearity analysis"
  )

  testthat::expect_true(
    "collinear_cor_df" %in% class(x)
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) > 0
  )

  testthat::expect_true(
    nrow(
      t(
        combn(
          x = vi_predictors[1:10],
          m = 2
        )
      )
    ) ==
      nrow(x)
  )

  testthat::expect_true(
    all(names(x) %in% c("x", "y", "correlation", "metric"))
  )

  testthat::expect_true(
    all(c("Pearson", "Cramer's V") %in% x$metric)
  )

  testthat::expect_true(
    all(x$correlation < 1)
  )

  # edge cases ----

  #no df
  testthat::expect_error(
    x <- cor_df(
      df = NULL,
      predictors = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #few rows
  testthat::expect_error(
    x <- cor_df(
      df = vi_smol[1, ],
      predictors = vi_predictors,
      quiet = TRUE
    ),
    regexp = "argument 'df' has fewer than 3 rows"
  )

  #no predictors
  x <- cor_df(
    df = vi_smol[, 1:5],
    predictors = NULL,
    quiet = TRUE
  )

  testthat::expect_true(
    "collinear_cor_df" %in% class(x)
  )

  testthat::expect_true(
    all(
      unique(
        c(
          x$x,
          x$y
        )
      ) %in%
        colnames(vi_smol)[1:5]
    )
  )

  #single predictor
  testthat::expect_message(
    x <- cor_df(
      df = vi_smol,
      predictors = vi_predictors[1],
      quiet = FALSE
    ),
    regexp = "only one valid predictor, returning one-row dataframe"
  ) |>
    suppressMessages()

  testthat::expect_true(
    "collinear_cor_df" %in% class(x)
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == 1
  )

  testthat::expect_true(
    all(names(x) %in% c("x", "y", "correlation", "metric"))
  )

  testthat::expect_true(x$x == vi_predictors[1])

  testthat::expect_true(x$y == vi_predictors[1])

  testthat::expect_equal(x$correlation, 1, tolerance = 1e-10)
})

testthat::test_that("`f_auto()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid
  vi_predictors_categorical <- identify_categorical_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  responses_vector <- c(
    "vi_numeric",
    "vi_counts",
    "vi_categorical",
    "vi_binomial",
    "vi_factor"
  )

  predictors_vector <- c(
    "vi_predictors",
    "vi_predictors_categorical",
    "vi_predictors_numeric"
  )

  collinear_functions <- ls(
    getNamespace("collinear"),
    all.names = TRUE
  )

  for (response in responses_vector) {
    for (predictor in predictors_vector) {
      testthat::expect_no_message(
        x <- f_auto(
          df = vi_smol,
          response = response,
          predictors = get(predictor),
          quiet = TRUE
        )
      )

      testthat::expect_true(
        x %in% collinear_functions
      )
    }
  }

  #edge cases
  testthat::expect_error(
    x <- f_auto(
      df = vi_smol,
      response = NULL,
      predictors = vi_predictors_numeric,
      quiet = FALSE
    ),
    regexp = "argument 'response' must not be NULL"
  ) |>
    suppressMessages()

  #edge cases
  testthat::expect_message(
    x <- f_auto(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric,
      quiet = FALSE
    ),
    regexp = "selected function"
  ) |>
    suppressMessages()
})

#' Find near-zero variance variables in a dataframe
#'
#'
#' @description
#' Returns the names of near-zero variance variables in a modelling dataframe.
#'
#' @inheritParams collinear
#' @param decimals (required, integer) Number of decimal places for the zero variance test. Smaller numbers will increase the number of variables detected as near-zero variance. Recommended values will depend on the range of the numeric variables in 'df'. Default: 4
#' @return character vector: names of near-zero variance columns.
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' #create zero and near variance predictors
#' vi_smol$zero_variance <- 1
#' vi_smol$near_zero_variance <- runif(
#'   n = nrow(vi_smol),
#'   min = 0,
#'   max = 0.0001
#'   )
#'
#'
#' #add to vi predictors
#' vi_predictors <- c(
#'   vi_predictors,
#'   "zero_variance",
#'   "near_zero_variance"
#' )
#'
#' #identify zero variance predictors
#' x <- identify_zero_variance_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )
#'
#' x
#'
#' @autoglobal
#' @family data_types
#' @author Blas M. Benito, PhD
#' @export
identify_zero_variance_variables <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  decimals = 4,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_zero_variance_variables()",
    ... = ...
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (!is.null(responses)) {
    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  vars_string <- if (!is.null(predictors) && !is.null(responses)) {
    "variables"
  } else if (!is.null(predictors)) {
    "predictors"
  } else if (!is.null(responses)) {
    "responses"
  } else {
    "variables"
  }

  predictors <- c(responses, predictors)

  if (is.null(predictors) || length(predictors) == 0) {
    stop(
      "\n",
      function_name,
      ": there are no ",
      vars_string,
      " to identify.",
      call. = FALSE
    )
  }

  decimals <- as.integer(decimals)

  df <- df[, predictors, drop = FALSE]

  predictors <- predictors[
    vapply(
      X = df,
      FUN = is.numeric,
      FUN.VALUE = logical(1)
    )
  ]

  if (length(predictors) == 0) {
    return(NULL)
  }

  df <- df[, predictors, drop = FALSE]

  #compute variance on valid values only
  predictors <- colnames(df)[
    round(
      sapply(
        X = df,
        FUN = function(x) {
          stats::var(
            x = x[is.finite(x)],
            na.rm = TRUE
          )
        }
      ),
      decimals
    ) ==
      0
  ]

  if (
    quiet == FALSE &&
      length(predictors) > 0
  ) {
    message(
      "\n",
      function_name,
      ": invalid ",
      vars_string,
      " due to near-zero variance:\n - ",
      paste(
        predictors,
        collapse = "\n - "
      )
    )
  }

  if (length(predictors) == 0) {
    predictors <- NULL
  }

  predictors
}

#' Find valid numeric variables in a dataframe
#'
#' @description
#' Identifies valid numeric variables and ignores those with constant values.
#'
#' @inheritParams identify_valid_variables
#' @return list:
#' \itemize{
#'   \item \code{valid}: character vector with valid numeric predictor names.
#'   \item \code{invalid}: character vector with invalid numeric predictor names due to near-zero variance.
#' }
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' x <- identify_numeric_variables(
#'   df = vi_smol,
#'   responses = "vi_numeric",
#'   predictors = vi_predictors
#' )
#'
#' #valid numeric predictors
#' x$valid
#'
#' #invalid due to zero variance (none here)
#' x$invalid
#'
#' @autoglobal
#' @family data_types
#' @author Blas M. Benito, PhD
#' @export
identify_numeric_variables <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  decimals = 4,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_numeric_variables()",
    ... = ...
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (!is.null(responses)) {
    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      responses = responses,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  vars_string <- if (!is.null(predictors) && !is.null(responses)) {
    "variables"
  } else if (!is.null(predictors)) {
    "predictors"
  } else if (!is.null(responses)) {
    "responses"
  } else {
    "variables"
  }

  predictors <- c(responses, predictors)

  if (is.null(predictors) || length(predictors) == 0) {
    stop(
      "\n",
      function_name,
      ": there are no ",
      vars_string,
      " to identify.",
      call. = FALSE
    )
  }

  out_list <- list(
    valid = NULL,
    invalid = NULL
  )

  #get numeric predictors
  predictors_numeric_all <- predictors[
    vapply(
      X = df[, predictors, drop = FALSE],
      FUN = is.numeric,
      FUN.VALUE = logical(1)
    )
  ]

  if (length(predictors_numeric_all) == 0) {
    return(out_list)
  }

  #ignore constant and near-zero variance predictors
  predictors_numeric_invalid <- identify_zero_variance_variables(
    df = df,
    predictors = predictors_numeric_all,
    quiet = TRUE,
    function_name = function_name
  )

  if (quiet == FALSE && length(predictors_numeric_invalid) > 0) {
    message(
      "\n",
      function_name,
      ": invalid numeric ",
      vars_string,
      " due to near-zero variance:\n - ",
      paste(
        predictors_numeric_invalid,
        collapse = "\n - "
      )
    )
  }

  predictors_numeric_valid <- setdiff(
    x = predictors_numeric_all,
    y = predictors_numeric_invalid
  )

  if (length(predictors_numeric_valid) > 0) {
    out_list$valid <- predictors_numeric_valid
  }

  if (length(predictors_numeric_invalid) > 0) {
    out_list$invalid <- predictors_numeric_invalid
  }

  out_list
}

testthat::test_that("`model_formula()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #null df
  testthat::expect_error(
    x <- model_formula(),
    regexp = "argument 'df' cannot be NULL"
  )

  testthat::expect_error(
    x <- model_formula(
      df = vi_smol,
      response = NULL
    ),
    regexp = "argument 'response' cannot be NULL"
  )

  #additive formula
  x <- model_formula(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:3]
  )

  testthat::expect_true(
    inherits(x = x, what = "formula")
  )

  #using a formula in a model
  m <- stats::lm(
    formula = x,
    data = vi_smol
  )

  testthat::expect_true(
    inherits(x = m, what = "lm")
  )

  #polynomial formula
  x <- model_formula(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:3],
    term_f = "poly",
    term_args = "degree = 3, raw = TRUE"
  )

  testthat::expect_true(
    inherits(x = x, what = "formula")
  )

  m <- stats::lm(
    formula = x,
    data = vi_smol
  )

  testthat::expect_true(
    inherits(x = m, what = "lm")
  )

  #gam formula
  x <- model_formula(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:3],
    term_f = "s"
  )

  testthat::expect_true(
    inherits(x = x, what = "formula")
  )

  m <- mgcv::gam(
    formula = x,
    data = vi_smol
  )

  testthat::expect_true(
    inherits(x = m, what = "gam")
  )

  #random effect
  testthat::expect_error(
    x <- model_formula(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[1:3],
      random_effects = 2
    ),
    regexp = "argument 'random_effects' must be a character string or vector"
  )

  testthat::expect_error(
    x <- model_formula(
      df = vi_smol,
      response = "vi_numeric",
      predictors = c(vi_predictors_numeric[1:3], "country_name"),
      random_effects = "country_name"
    ),
    regexp = "argument 'random_effects' must name variables not in argument 'predictors'"
  )

  x <- model_formula(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:3],
    random_effects = "country_name" #from vi_smol$country_name
  )

  testthat::expect_true(
    inherits(x = x, what = "formula")
  )

  # sf geometry column handling (issue #17)
  sf_df <- vi_smol[1:50, c("vi_numeric", vi_predictors_numeric[1:3])]
  sf_df$geometry <- I(rep(list(c(0, 0)), nrow(sf_df)))
  attr(sf_df, "sf_column") <- "geometry"

  x <- model_formula(
    df = sf_df,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:3],
    quiet = TRUE
  )

  testthat::expect_true(
    inherits(x = x, what = "formula")
  )
})

testthat::test_that("`vif_stats()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #general usage

  #from predictors dataframe
  x <- vif_stats(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    all(colnames(x) %in% c("method", "statistic", "value"))
  )

  #from m results
  m <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  x <- vif_stats(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = TRUE,
    m = m
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    all(colnames(x) %in% c("method", "statistic", "value"))
  )

  #edge cases
  testthat::expect_error(
    x <- vif_stats(),
    regexp = "argument 'df' cannot be NULL"
  )
})

testthat::test_that("`cor_matrix()` works", {
  testthat::skip_on_cran()
  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #nput from cor_df()

  testthat::expect_warning(
    df <- cor_df(
      df = vi_smol,
      predictors = vi_predictors[1:15],
      quiet = TRUE
    ),
    regexp = "may bias the multicollinearity analysis"
  )

  m <- cor_matrix(df = df)

  testthat::expect_true(
    "collinear_cor_matrix" %in% class(m)
  )

  testthat::expect_true(
    is.matrix(m)
  )

  testthat::expect_true(
    all(rownames(m) == colnames(m))
  )

  testthat::expect_true(
    all(rownames(m) %in% vi_predictors[1:15])
  )

  testthat::expect_true(
    all(diag(m) == 1)
  )

  #univariate case
  testthat::expect_message(
    df <- cor_df(
      df = vi_smol,
      predictors = vi_predictors[1]
    ),
    regexp = "only one valid predictor"
  ) |>
    suppressMessages()

  m <- cor_matrix(
    df = df,
    quiet = TRUE
  )

  testthat::expect_true(
    "collinear_cor_matrix" %in% class(m)
  )

  testthat::expect_true(
    is.matrix(m)
  )

  testthat::expect_true(
    all(rownames(m) == colnames(m))
  )

  testthat::expect_true(
    all(rownames(m) %in% vi_predictors[1:15])
  )

  testthat::expect_true(
    all(diag(m) == 1)
  )

  #few rows
  testthat::expect_error(
    x <- cor_matrix(
      df = vi_smol[1:2, ],
      predictors = vi_predictors[1:15]
    ),
    regexp = "has fewer than 3 rows"
  )

  testthat::expect_message(
    x <- cor_matrix(
      df = vi_smol[1:9, ],
      predictors = vi_predictors[1:15]
    ),
    regexp = "has fewer than 10 rows"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- cor_matrix(
      df = vi_smol[1:29, ],
      predictors = vi_predictors[1:15]
    ),
    regexp = "has fewer than 30 rows"
  ) |>
    suppressMessages()

  #check dots argument
  m1 <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_numeric[1:10],
    quiet = TRUE
  )

  attr(x = m1, which = "my_matrix") <- TRUE

  m2 <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_numeric[1:10],
    quiet = TRUE,
    m = m1
  )

  testthat::expect_true(
    attributes(m2)$my_matrix
  )

  testthat::expect_equal(
    object = m1,
    expected = m2
  )

  m3 <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_numeric[11:20],
    quiet = TRUE,
    m = m1
  )

  testthat::expect_null(
    attributes(m3)$my_matrix
  )

  testthat::expect_false(
    all(colnames(m1) %in% colnames(m3))
  )
})

testthat::test_that("`vif_select()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  testthat::expect_message(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      quiet = FALSE
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_warning(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      quiet = TRUE
    )
  ) |>
    suppressWarnings()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors[1:10])
  )

  testthat::expect_true(
    length(vi_predictors[1:10]) > length(x)
  )

  testthat::expect_message(
    y <- vif_select(
      df = vi_smol,
      predictors = x,
      quiet = FALSE
    ),
    regexp = "maximum VIF is <= 5, multicollinearity filtering is not required"
  ) |>
    suppressMessages()

  testthat::expect_true(
    all(x %in% y)
  )

  #custom preference order
  preference_order <- c(
    "swi_mean",
    "topo_elevation",
    "hola"
  )

  testthat::expect_message(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      preference_order = preference_order
    ),
    regexp = "'predictors' from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors[1:10])
  )

  testthat::expect_true(
    length(vi_predictors[1:10]) > length(x)
  )

  testthat::expect_true(
    all(preference_order[1] == x[1])
  )

  #automated preference order
  preference_order <- preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors[1:10],
    quiet = TRUE
  )

  testthat::expect_warning(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      preference_order = preference_order,
      quiet = TRUE
    )
  ) |>
    suppressWarnings()

  testthat::expect_true(
    is.character(x)
  )

  testthat::expect_true(
    all(x %in% vi_predictors[1:10])
  )

  testthat::expect_true(
    length(vi_predictors[1:10]) > length(x)
  )

  testthat::expect_true(
    all(preference_order$predictor[1] == x[1])
  )

  # edge cases ----

  #no df
  testthat::expect_error(
    x <- vif_select(
      df = NULL,
      predictors = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  testthat::expect_error(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors,
      max_vif = NULL
    ),
    regexp = "argument 'max_vif' cannot be NULL"
  )

  #no predictors
  x <- vif_select(
    df = vi_smol[, 1:5],
    predictors = NULL,
    preference_order = NULL,
    quiet = TRUE
  )

  testthat::expect_true(
    all(x %in% colnames(vi_smol)[1:5])
  )

  #single predictor

  testthat::expect_message(
    x <- vif_select(
      df = vi_smol,
      predictors = vi_predictors_numeric[1]
    ),
    regexp = "only one valid predictor in 'predictors', skipping multicollinearity filtering"
  ) |>
    suppressMessages()

  testthat::expect_true(
    x == vi_predictors_numeric[1]
  )
})

test_that("step_collinear works!", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  test_selection <- collinear(
    df = vi_smol,
    responses = "vi_numeric",
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  test_selection <- test_selection$vi_numeric$selection

  vi_formula <- model_formula(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric
  )

  rec <- recipes::recipe(
    formula = vi_formula,
    data = vi_smol
  ) |>
    step_collinear(
      recipes::all_predictors()
    )

  rec_prep <- recipes::prep(
    rec,
    training = vi_smol
  )

  baked <- recipes::bake(
    rec_prep,
    new_data = vi_smol
  )

  testthat::expect_true(is.data.frame(baked))

  testthat::expect_true(
    all(test_selection %in% colnames(baked))
  )

  testthat::expect_true(
    all(c("vi_numeric", rec_prep$steps[[1]]$selected) %in% names(baked))
  )
})

testthat::test_that("`vif()` works", {
  testthat::skip_on_cran()

  # numeric types ----
  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  m <- cor_matrix(
    df = vi_smol,
    predictors = vi_predictors_numeric,
    quiet = TRUE
  )

  v <- vif(m = m)

  testthat::expect_true(
    is.numeric(v)
  )

  testthat::expect_true(
    length(v) == length(vi_predictors_numeric)
  )

  testthat::expect_true(
    all(names(v) %in% vi_predictors_numeric)
  )

  #no input
  testthat::expect_error(
    v <- vif(m = NULL)
  )

  #matrix with wrong dimensions
  testthat::expect_error(
    v <- vif(m = m[1:nrow(m), 2:ncol(m)])
  )

  testthat::expect_error(
    v <- vif(m = 1)
  )

  #matrix without dimnames
  dimnames(m) <- NULL
  testthat::expect_error(
    v <- vif(m = m)
  )

  #perfect correlations
  m <- matrix(
    data = rep(x = 1, times = 100),
    ncol = 10,
    nrow = 10
  )

  dimnames(m) <- list(
    letters[1:10],
    letters[1:10]
  )

  class(m) <- c("collinear_cor_matrix", class(m))
})

testthat::test_that("`validate_arg_preference_order()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")
  vi_predictors_numeric <- identify_numeric_variables(
    df = vi_smol,
    predictors = vi_predictors
  )$valid

  #no arguments
  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = NULL,
      predictors = NULL,
      preference_order = NULL,
      function_name = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  # df only ----
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol[, vi_predictors_numeric],
      response = NULL,
      predictors = NULL,
      preference_order = NULL,
      quiet = FALSE
    ),
    regexp = "from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    unique(x$metric) == "1 - R-squared"
  )

  testthat::expect_true(
    unique(x$f) == "stats::cor()"
  )

  #pass validated arg
  x <- validate_arg_preference_order(
    df = vi_smol,
    response = NULL,
    predictors = vi_predictors_numeric,
    preference_order = x
  )

  #df and response, no predictors ----
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol[, c("vi_numeric", vi_predictors_numeric)],
      response = "vi_numeric",
      predictors = NULL,
      preference_order = NULL,
      quiet = FALSE
    ),
    regexp = "from lower to higher multicollinearity"
  ) |>
    suppressMessages()

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    all(vi_predictors_numeric %in% x$predictor)
  )

  #df and two responses
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol[, c("vi_numeric", "vi_binomial", vi_predictors_numeric)],
      response = c("vi_numeric", "vi_binomial"),
      predictors = vi_predictors_numeric,
      preference_order = vi_predictors_numeric,
      quiet = FALSE
    ),
    regexp = "argument 'response' must be of length 1, using response"
  ) |>
    suppressMessages()

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  #auto
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[1:10],
      preference_order = "auto",
      quiet = FALSE
    ),
    regexp = "value 'auto' for the argument 'preference_order' is deprecated"
  ) |>
    suppressMessages()

  #preference_order vector ----
  x <- validate_arg_preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:10],
    preference_order = vi_predictors_numeric[1:10],
    quiet = FALSE
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric[1:10])
  )

  testthat::expect_true(
    all(vi_predictors_numeric[1:10] %in% x$predictor)
  )

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  testthat::expect_true(
    is.na(unique(x$f))
  )

  testthat::expect_true(
    unique(x$metric) == "user_preference"
  )

  #incomplete preference order vector ----
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[1:10],
      preference_order = vi_predictors_numeric[10:5],
      quiet = FALSE
    ),
    regexp = "4 'predictors' from lower to higher multicollinearity"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric[1:10])
  )

  testthat::expect_true(
    x$predictor[1] == vi_predictors_numeric[10]
  )

  testthat::expect_true(
    all(vi_predictors_numeric[1:10] %in% x$predictor)
  )

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  testthat::expect_true(
    all(unique(x$metric) %in% c("user_preference", "1 - R-squared"))
  )

  #preference_order dataframe ----
  preference_df <- preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:12],
    quiet = TRUE
  )

  #missing one predictor
  x <- validate_arg_preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:13],
    preference_order = preference_df,
    quiet = FALSE
  )

  testthat::expect_true(
    all(vi_predictors_numeric[1:13] %in% x$predictor)
  )

  #wrong response
  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_binomial",
      predictors = vi_predictors_numeric[1:12],
      preference_order = preference_df,
      quiet = FALSE
    ),
    regexp = "argument 'response' does not match the column 'response' of the dataframe 'preference_order'"
  ) |>
    suppressMessages()

  #wrong column names
  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[1:10],
      preference_order = preference_df[, c("response", "predictor")],
      quiet = FALSE
    ),
    regexp = "dataframe 'preference_order' must have these columns"
  ) |>
    suppressMessages()

  #wrong predictors
  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[13:20],
      preference_order = preference_df,
      quiet = FALSE
    ),
    regexp = "does not contain any 'predictors'"
  ) |>
    suppressMessages()

  x <- validate_arg_preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:10],
    preference_order = preference_df,
    quiet = FALSE
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric[1:10])
  )

  testthat::expect_true(
    x$predictor[1] == preference_df[1, "predictor"]
  )

  testthat::expect_true(
    all(vi_predictors_numeric[1:10] %in% x$predictor)
  )

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  testthat::expect_true(
    unique(x$f) == "f_numeric_glm"
  )

  #filtering responses ----
  preference_df <- preference_order(
    df = vi_smol,
    responses = c("vi_numeric", "vi_binomial"),
    predictors = vi_predictors_numeric[1:12],
    quiet = TRUE
  )

  testthat::expect_true(
    all(c("vi_numeric", "vi_binomial") %in% preference_df$response)
  )

  x <- validate_arg_preference_order(
    df = vi_smol,
    response = "vi_numeric",
    predictors = vi_predictors_numeric[1:10],
    preference_order = preference_df,
    function_name = NULL,
    quiet = FALSE
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric[1:10])
  )

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  testthat::expect_true(
    length(setdiff(preference_df$predictor, vi_predictors_numeric[1:10])) == 2
  )

  #wrong response ----
  preference_df <- preference_order(
    df = vi_smol,
    responses = c("vi_numeric", "vi_binomial"),
    predictors = vi_predictors_numeric[1:12],
    quiet = TRUE
  )

  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "wrong_response",
      predictors = vi_predictors_numeric[1:10],
      preference_order = preference_df,
      function_name = NULL,
      quiet = TRUE
    ),
    regexp = "dataframe 'preference_order' contains more than one response"
  )

  #several responses ----
  testthat::expect_message(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = c("vi_numeric", "vi_binomial"),
      predictors = vi_predictors_numeric[1:10],
      preference_order = preference_df,
      function_name = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'response' must be of length 1"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors_numeric[1:10])
  )

  testthat::expect_true(
    unique(x$response) == "vi_numeric"
  )

  #no response ----
  preference_df <- preference_order(
    df = vi_smol,
    response = NULL,
    predictors = vi_predictors_numeric[1:12],
    quiet = TRUE
  )

  testthat::expect_error(
    x <- validate_arg_preference_order(
      df = vi_smol,
      response = "vi_numeric",
      predictors = vi_predictors_numeric[1:12],
      preference_order = preference_df,
      function_name = NULL,
      quiet = FALSE
    ),
    regexp = "argument 'response' does not match the column 'response' of the dataframe 'preference_order'"
  )
})

#' Print single selection results from `collinear`
#'
#' @param x (required, sub-list in output of [collinear()]) Object to print. Default: NULL
#' @param n (optional, integer) Maximum printed vector length. Default: 5.
#' @param ... Ignored, kept for consistency with generic.
#' @method print collinear_selection
#' @family S3_methods
#' @autoglobal
#' @export
print.collinear_selection <- function(
  x = NULL,
  n = 5,
  ...
) {
  #function to shorten formulas
  short_formula <- function(f = NULL, n = 5) {
    terms <- attr(terms(f), "term.labels")
    nterms <- length(terms)

    n <- max(n, 1)
    n <- min(n, nterms)

    #adding one term if needed
    if (nterms == (n + 1)) {
      n <- nterms
    }

    lhs <- deparse(f[[2]])

    rhs <- paste0(terms[1:n], collapse = " + ")

    f_string <- paste(lhs, rhs, sep = " ~ ")

    notice <- if (n < nterms) {
      paste0(
        "+ ... (",
        nterms - n,
        " terms omitted)"
      )
    } else {
      NULL
    }

    paste(f_string, notice)
  }

  # response ----
  if (!is.null(x$response)) {
    msg <- paste0(
      "response: ",
      x$response
    )

    msg_length <- nchar(msg)

    underline <- paste0(rep(x = "-", times = nchar(msg)), collapse = "")

    cat(
      " -",
      msg,
      fill = TRUE
    )

    cat("  ", underline)

    cat("\n")
  }

  # df ----

  cat("\n")

  cat(" + df:", fill = TRUE)

  cat("   - rows:", nrow(x$df), fill = TRUE)

  cat("   - cols:", ncol(x$df))

  cat("\n")

  # preference ----
  if (!is.null(x$preference_order)) {
    cat("\n")

    cat(" + preference order:", fill = TRUE)

    cat("   + df:", fill = TRUE)

    cat("     - rows:", nrow(x$preference_order), fill = TRUE)

    cat("     - cols:", ncol(x$preference_order))

    cat("\n")

    selection <- x$preference_order$predictor[1:n]

    omitted <- length(x$preference_order$df$predictor) - n

    symbol <- ifelse(
      test = length(selection) == 1,
      yes = " -",
      no = "   +"
    )

    cat(
      symbol,
      "preference:\n     -",
      paste(
        selection,
        collapse = "\n     - "
      )
    )

    if (omitted > 0) {
      cat(
        paste0(
          "\n     - ... (",
          omitted,
          " ommited)"
        )
      )
    }

    cat("\n")

    cat("   + f:", unique(x$preference_order$f), fill = TRUE)
  }

  # selection ----
  if (!is.null(x$selection)) {
    cat("\n")

    selection <- stats::na.omit(x$selection[1:n])

    omitted <- length(x$selection) - n

    symbol <- ifelse(
      test = length(selection) == 1,
      yes = " -",
      no = " +"
    )

    cat(
      symbol,
      "selection:\n   -",
      paste(
        selection,
        collapse = "\n   - "
      )
    )

    if (omitted > 0) {
      cat(
        paste0(
          "\n   - ... (",
          omitted,
          " ommited)"
        )
      )
    }

    cat("\n")
  }

  # formulas ----
  if (!is.null(x$formulas)) {
    cat("\n")

    if (length(x$formulas) > 1) {
      cat(" + formulas:", fill = TRUE)
    } else {
      cat(" - formula:", fill = TRUE)
    }

    for (i in names(x$formulas)) {
      cat(
        paste0("   - ", i, ": ", short_formula(f = x$formulas[[i]], n = n)),
        "\n"
      )
    }
  }

  cat("\n\n")

  invisible(x)
}

#' Signed pairwise correlation matrix
#'
#' @description
#' Computes a square matrix of pairwise correlations for a set of numeric and/or categorical predictors.
#'
#' If \code{df} is already a correlation dataframe generated by [cor_df()]), the function transforms it into a correlation matrix. Otherwise, [cor_df()] is used internally to compute pairwise correlations before generating the matrix.
#'
#' Supports parallel computation via \code{future::plan()} and optional progress reporting via \code{progressr::handlers()}.
#'
#'
#' @inheritParams collinear
#' @param df (required; dataframe, tibble, or sf) A dataframe with predictors or the output of [cor_df()]. Default: NULL.
#' @return correlation matrix
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' predictors <- c(
#'   "koppen_zone", #character
#'   "soil_type", #factor
#'   "topo_elevation", #numeric
#'   "soil_temperature_mean" #numeric
#' )
#'
#' #from dataframe with predictors
#' x <- cor_matrix(
#'   df = vi_smol,
#'   predictors = predictors
#' )
#'
#' x
#'
#' #from correlation dataframe
#' x <- cor_df(
#'   df = vi_smol,
#'   predictors = predictors
#' ) |>
#'   cor_matrix()
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_assessment
#' @author Blas M. Benito, PhD
#' @export
cor_matrix <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_matrix()",
    function_name = dots$function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  if (
    "collinear_cor_matrix" %in%
      class(dots$m) &&
      all(predictors %in% colnames(dots$m))
  ) {
    if (length(predictors) > length(colnames(dots$m))) {
      m <- dots$m
      m <- m[predictors, predictors]
      class(m) <- c("collinear_cor_matrix", class(m))
      return(m)
    }

    return(dots$m)
  }

  #if df with predictors, compute correlation dataframe
  if (!"collinear_cor_df" %in% class(df)) {
    df <- cor_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #create all possible pairs
  df <- rbind(
    df[, c("x", "y", "correlation")],
    data.frame(
      x = df$y,
      y = df$x,
      correlation = df$correlation
    )
  )

  #rows and col names
  variables <- sort(
    unique(
      c(df$x, df$y)
    )
  )

  #empty square matrix
  m <- matrix(
    data = NA,
    nrow = length(variables),
    ncol = length(variables)
  )

  #named vector to map row/column names to indices
  index_map <- stats::setNames(
    object = seq_along(variables),
    nm = variables
  )

  #vectorized indexing to fill in the matrix
  m[
    cbind(
      index_map[df$x],
      index_map[df$y]
    )
  ] <- df$correlation

  #dim names
  dimnames(m) <- list(
    variables,
    variables
  )

  #replace NA in diag with 1
  diag(m) <- 1

  class(m) <- c("collinear_cor_matrix", class(m))

  m
}

#' Compute signed pairwise correlations dataframe
#'
#' @description
#' Computes pairwise correlations between predictors using appropriate methods for different variable types:
#' \itemize{
#'   \item **Numeric vs. Numeric**: Pearson correlation via \code{stats::cor()}.
#'   \item **Numeric vs. Categorical**: Target-encodes the categorical variable  using the numeric variable as reference via [target_encoding_lab()] with leave-one-out method, then computes Pearson correlation.
#'   \item **Categorical vs. Categorical**: Cramer's V via [cor_cramer()] as a measure of association. See [cor_cramer()] for important notes on mixing Pearson correlation and Cramer's V in multicollinearity analysis.
#' }
#'
#' Parallelization via \code{future::plan()} and progress bars via \code{progressr::handlers()} are supported but only beneficial for large datasets with categorical predictors. Numeric-only correlations do not use parallelization or progress bars. Example: With 16 workers, 30k rows (dataframe [spatialData::vi]), 49 numeric and 12 categorical predictors (see [spatialData::vi_predictors]), parallelization achieves a 5.4x speedup (147s → 27s).
#'
#' @inheritParams collinear
#'
#' @return dataframe with columns:
#' \itemize{
#'   \item \code{x}: character, first predictor name.
#'   \item \code{y}: character, second predictor name.
#'   \item \code{correlation}: numeric, Pearson correlation (numeric vs. numeric and numeric vs. categorical) or Cramer's V (categorical vs. categorical).
#' }
#'
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' #predictors
#' predictors = c(
#'   "koppen_zone", #character
#'   "soil_type", #factor
#'   "topo_elevation", #numeric
#'   "soil_temperature_mean" #numeric
#' )
#'
#' x <- cor_df(
#'   df = vi_smol,
#'   predictors = predictors
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_assessment
#' @export
cor_df <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_df()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  predictors <- validate_arg_predictors(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df.ncol <- ncol(df)

  df <- validate_arg_df(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #revalidate predictors if any columns were removed
  if (ncol(df) < df.ncol) {
    attributes(predictors)$validated <- NULL

    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #identify predictors types
  predictors <- identify_valid_variables(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #univariate case
  if (length(c(predictors$numeric, predictors$categorical)) == 1) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": only one valid predictor, returning one-row dataframe."
      )
    }

    temp <- c(predictors$numeric, predictors$categorical)

    out_df <- data.frame(
      x = temp,
      y = temp,
      correlation = 1,
      metric = ifelse(
        test = length(predictors$categorical) == 1,
        yes = "cramer_v",
        no = "pearson"
      )
    )

    class(out_df) <- c("collinear_cor_df", class(out_df))

    return(out_df)
  }

  #warning about cramers v
  if (
    length(predictors$categorical) >= 2L &&
      length(predictors$numeric) >= 1L
  ) {
    cardinality <- vapply(
      X = df[predictors$categorical],
      FUN = function(x) length(unique(x)),
      integer(1L)
    )

    vars_above_2 <- sum(cardinality > 2L)

    if (vars_above_2 > 0L) {
      msg <- paste0(
        "\n",
        function_name,
        ": ",
        vars_above_2,
        " categorical predictors have cardinality > 2 and may bias the multicollinearity analysis. Applying target encoding to convert them to numeric will solve this issue."
      )

      if (isFALSE(quiet)) {
        message(msg)
      } else {
        warning(msg, call. = FALSE)
      }
    }
  }

  #initialize output dataframes
  numerics_df <- NULL
  categoricals_df <- NULL

  #categoricals
  df_num_cat <- NULL
  df_cat_cat <- NULL

  #num vs cat
  if (
    all(
      c(
        length(predictors$categorical),
        length(predictors$numeric)
      ) >
        0
    )
  ) {
    #numerics_vs_categoricals
    df_num_cat <- expand.grid(
      x = predictors$numeric,
      y = predictors$categorical,
      correlation = NA_real_,
      metric = "Pearson",
      type = 1,
      stringsAsFactors = FALSE
    )
  }

  #cat vs cat
  if (length(predictors$categorical) > 1) {
    pairs_cat_cat <- t(
      utils::combn(
        x = predictors$categorical,
        m = 2
      )
    )

    df_cat_cat <- data.frame(
      x = pairs_cat_cat[, 1],
      y = pairs_cat_cat[, 2],
      correlation = NA_real_,
      metric = "Cramer's V",
      type = 2,
      stringsAsFactors = FALSE
    )
  }

  #df to iterate over
  categoricals_df <- rbind(
    df_num_cat,
    df_cat_cat
  )

  #compute iterations for categorical vars
  iterations_categorical <- ifelse(
    test = !is.null(categoricals_df) && nrow(categoricals_df) > 0,
    yes = nrow(categoricals_df),
    no = 0
  )

  if (iterations_categorical > 0) {
    p <- progressr::progressor(
      steps = iterations_categorical
    )

    categoricals_df$correlation <- future.apply::future_apply(
      X = categoricals_df,
      MARGIN = 1,
      FUN = function(x) {
        p()

        df.x <- data.frame(
          x = df[[x[1]]],
          y = df[[x[2]]]
        ) |>
          stats::na.omit()

        #num_vs_cat
        if (x[5] == "1") {
          attr(
            x = df.x,
            which = "validated"
          ) <- TRUE

          #target encode
          df.x <- target_encoding_lab(
            df = df.x,
            response = "x",
            predictors = "y",
            encoding_method = "loo",
            overwrite = TRUE,
            quiet = TRUE,
            function_name = function_name
          )

          #compute correlation
          score <- stats::cor(
            x = df.x$x,
            y = df.x$y,
            use = "complete.obs",
            method = "pearson"
          )
        } else {
          #cat vs cat
          score <- cor_cramer(
            x = df.x$x,
            y = df.x$y,
            check_input = FALSE,
            function_name = function_name
          )
        }

        score
      }, #end of lambda function
      future.seed = TRUE
    )

    #remove type
    categoricals_df$type <- NULL
  }

  #numerics
  if (length(predictors$numeric) > 1) {
    numerics_matrix <- stats::cor(
      x = df[, predictors$numeric, drop = FALSE],
      use = "complete.obs",
      method = "pearson"
    )

    upper_indices <- which(
      x = upper.tri(numerics_matrix),
      arr.ind = TRUE
    )

    numerics_df <- data.frame(
      x = rownames(numerics_matrix)[upper_indices[, 1]],
      y = colnames(numerics_matrix)[upper_indices[, 2]],
      correlation = numerics_matrix[upper_indices],
      metric = "Pearson",
      stringsAsFactors = FALSE
    )
  }

  out_df <- rbind(
    numerics_df,
    categoricals_df
  )

  #arrange by correlation values
  out_df <- out_df[
    order(
      abs(out_df$correlation),
      decreasing = TRUE
    ),
    ,
    drop = FALSE
  ]

  rownames(out_df) <- NULL

  class(out_df) <- c("collinear_cor_df", class(out_df))

  out_df
}

#' Compute variance inflation factors dataframe
#'
#' @description
#'
#' Computes the pairwise correlation matrix between all pairs of predictors via [cor_df()] and [cor_matrix()], applies [vif()] to the resulting matrix to compute Variance Inflation Factors, and returns the result as a dataframe.
#'
#' @inheritSection collinear Variance Inflation Factors
#'
#' @inheritParams collinear
#' @return dataframe with columns:
#' \itemize{
#'   \item \code{predictor}: Character, predictor name.
#'   \item \code{vif}: Numeric, variance inflation factor
#' }
#'
#' @examples
#'
#' data(vi_smol, package = "spatialData")
#'
#' # ## OPTIONAL: parallelization setup
#' # ## irrelevant when all predictors are numeric
#' # ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' # ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' #predictors
#' predictors = c(
#'   "koppen_zone", #character
#'   "soil_type", #factor
#'   "topo_elevation", #numeric
#'   "soil_temperature_mean" #numeric
#' )
#'
#' x <- vif_df(
#'   df = vi_smol,
#'   predictors = predictors
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#'
#' @autoglobal
#' @family multicollinearity_assessment
#' @inherit vif_select references
#' @export
vif_df <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::vif_df()",
    function_name = dots$function_name
  )

  if (!inherits(dots$m, "collinear_cor_matrix")) {
    df <- validate_arg_df_not_null(
      df = df,
      function_name = function_name
    )

    quiet <- validate_arg_quiet(
      quiet = quiet,
      function_name = function_name
    )

    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )

    df.ncol <- ncol(df)

    df <- validate_arg_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )

    #revalidate predictors if any columns were removed
    if (ncol(df) < df.ncol) {
      attributes(predictors)$validated <- NULL

      predictors <- validate_arg_predictors(
        df = df,
        predictors = predictors,
        quiet = quiet,
        function_name = function_name
      )
    }

    #if no predictors
    if (length(predictors) == 1) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": only one valid predictor, returning one-row dataframe."
        )
      }

      return(
        data.frame(
          variable = predictors,
          vif = 0
        )
      )
    }
  }

  #compute correlation matrix
  m <- cor_matrix(
    df = df,
    predictors = predictors,
    function_name = function_name,
    quiet = quiet,
    m = dots$m
  )

  out <- vif(
    m = m,
    quiet = quiet,
    function_name = function_name
  ) |>
    data.frame(
      stringsAsFactors = FALSE
    )

  #format dataframe
  colnames(out) <- "vif"
  out$predictor <- colnames(m)
  rownames(out) <- NULL

  class(out) <- c("collinear_vif_df", class(out))

  out
}

testthat::test_that("`vif_df()` works", {
  testthat::skip_on_cran()

  data(vi_smol, vi_predictors, package = "spatialData")

  testthat::expect_message(
    x <- vif_df(
      df = vi_smol,
      predictors = vi_predictors[1:10]
    ),
    regexp = "converted the following character columns to factor"
  ) |>
    suppressMessages()

  testthat::expect_message(
    x <- vif_df(
      df = vi_smol,
      predictors = vi_predictors[1:10]
    ),
    regexp = "may bias the multicollinearity analysis"
  ) |>
    suppressMessages()

  testthat::expect_warning(
    x <- vif_df(
      df = vi_smol,
      predictors = vi_predictors[1:10],
      quiet = TRUE
    ),
    regexp = "may bias the multicollinearity analysis"
  )

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    all(names(x) %in% c("predictor", "vif"))
  )

  testthat::expect_true(
    nrow(x) == length(vi_predictors[1:10])
  )

  # edge cases ----

  #no arguments
  testthat::expect_error(
    x <- vif_df(
      df = NULL,
      predictors = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  #single predictor
  testthat::expect_message(
    x <- vif_df(
      df = vi_smol,
      predictors = vi_predictors[1]
    ),
    regexp = "only one valid predictor"
  ) |>
    suppressMessages()

  testthat::expect_true(
    is.data.frame(x)
  )

  testthat::expect_true(
    nrow(x) == 1
  )

  testthat::expect_equal(x$vif, 0, tolerance = 1e-10)
})

#' Compute Cramer's V between categorical observations and predictions
#'
#' @description
#' Internal function to compute the Cramer's V of categorical observations versus categorical model predictions. Please read the help file of [cor_cramer()] for further details.
#'
#'
#' @param o (required; character vector) categorical observations. Default: NULL
#' @param p (required; character vector) categorical predictions. Default: NULL
#' @inheritParams collinear
#' @return numeric: Cramer's V
#' @export
#' @autoglobal
#' @family modelling_tools
#' @examples
#' score_cramer(
#'  o = c("a", "a", "b", "c", "c"),
#'  p = c("a", "b", "b", "c", "c")
#'  )
#'
score_cramer <- function(
  o = NULL,
  p = NULL,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::score_cramer()",
    ... = ...
  )

  out <- tryCatch(
    {
      cor_cramer(
        x = o,
        y = p,
        check_input = FALSE,
        function_name = function_name
      )
    },
    error = function(e) {
      stop(
        "\n",
        function_name,
        ": ",
        conditionMessage(e),
        call. = FALSE
      )
    }
  )

  out
}

#' Encode categories as response means
#'
#'
#' @inheritParams target_encoding_lab
#' @param predictor (required; string) Name of the categorical predictor to encode. Default: NULL
#' @param encoded_name (optional, string) Name of the encoded predictor. Default: NULL
#' @param smoothing (optional; integer) Groups smaller than this number have their means pulled towards the mean of the response across all cases. Ignored by [target_encoding_rank()] and [target_encoding_loo()]. Default: 0
#'
#'
#' @inherit target_encoding_lab return
#'
#' @examples
#'
#' #  loading example data
#' data(vi_smol, package = "spatialData")
#'
#' #mean encoding
#' #-------------
#'
#' df <- target_encoding_mean(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictor = "soil_type", #categorical
#'   encoded_name = "soil_type_encoded"
#' )
#'
#' if(interactive()){
#'
#'   plot(
#'     x = df$soil_type_encoded,
#'     y = df$vi_numeric,
#'     xlab = "encoded variable",
#'     ylab = "response"
#'   )
#'
#' }
#'
#'
#' #rank encoding
#' #----------
#'
#' df <- target_encoding_rank(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictor = "soil_type",
#'   encoded_name = "soil_type_encoded"
#' )
#'
#' if(interactive()){
#'
#'   plot(
#'     x = df$soil_type_encoded,
#'     y = df$vi_numeric,
#'     xlab = "encoded variable",
#'     ylab = "response"
#'   )
#'
#' }
#'
#'
#' #leave-one-out
#' #-------------
#'
#' df <- target_encoding_loo(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictor = "soil_type",
#'   encoded_name = "soil_type_encoded"
#' )
#'
#' if(interactive()){
#'
#'   plot(
#'     x = df$soil_type_encoded,
#'     y = df$vi_numeric,
#'     xlab = "encoded variable",
#'     ylab = "response"
#'   )
#'
#' }
#'
#' @export
#' @autoglobal
#' @family target_encoding
#' @rdname target_encoding_methods
target_encoding_mean <- function(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = 0,
  ...
) {
  if (is.null(encoded_name)) {
    encoded_name <- paste0(
      predictor,
      "__encoded"
    )
  }

  #mean encoding when smoothing > 0
  if (smoothing == 0) {
    df[[encoded_name]] <- stats::ave(
      x = df[[response]],
      df[[predictor]],
      FUN = function(x) mean(x, na.rm = TRUE)
    )
  } else {
    #global response mean
    global_response_mean <- mean(
      x = df[[response]],
      na.rm = TRUE
    )

    #encoding
    df[[encoded_name]] <- stats::ave(
      x = df[[response]],
      df[[predictor]],
      FUN = function(x) {
        n <- length(x)
        mean_x <- mean(x, na.rm = TRUE)
        (n * mean_x + smoothing * global_response_mean) / (n + smoothing)
      }
    )
  }

  df
}

testthat::test_that("`target_encoding_mean()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  #target_encoding_mean
  df <- target_encoding_mean(
    df = vi_smol,
    response = "vi_numeric",
    predictor = "soil_type",
    encoded_name = "soil_type_encoded_1",
    smoothing = 0
  )

  testthat::expect_true(
    class(df[["soil_type"]]) == "factor"
  )

  testthat::expect_true(
    class(df[["soil_type_encoded_1"]]) == "numeric"
  )

  df <- target_encoding_mean(
    df = df,
    response = "vi_numeric",
    predictor = "soil_type",
    encoded_name = "soil_type_encoded_2",
    smoothing = 30
  )

  testthat::expect_true(
    class(df[["soil_type_encoded_2"]]) == "numeric"
  )

  testthat::expect_true(
    cor(df$soil_type_encoded_1, df$soil_type_encoded_2) < 1
  )

  df <- target_encoding_mean(
    df = vi_smol,
    response = "vi_numeric",
    predictor = "soil_type",
    encoded_name = NULL,
    smoothing = 30
  )

  testthat::expect_true(
    "soil_type__encoded" %in% colnames(df)
  )
})

#' Removes \code{geometry} Column From \code{sf} Dataframes
#'
#' Remove geometry column from sf objects
#'
#' @inheritParams collinear
#'
#' @return dataframe
#' @family argument_validation
#' @autoglobal
#' @author Blas M. Benito, PhD
#' @export
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' #creating fake geometry column without sf loaded
#' vi_smol$geometry <- NA
#' attr(
#'   x = vi_smol,
#'   which = "sf_column"
#'   ) <- "geometry"
#'
#' #check new attribute
#' attributes(vi_smol)$sf_column
#'
#' #drop geometry column
#' df <- drop_geometry_column(
#'   df = vi_smol
#'   )
#'
#' #checking that the geometry was droppped
#' "geometry" %in% colnames(df)
#' attributes(df)$sf_column
#'
drop_geometry_column <- function(
  df = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::drop_geometry_column()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #remove geometry column from df
  sf.column <- attributes(df)$sf_column

  if (!is.null(sf.column)) {
    if (quiet == FALSE) {
      message("\n", function_name, ": dropping geometry column from 'df'.")
    }

    df <- as.data.frame(df)
    df[[sf.column]] <- NULL
    attr(df, "sf_column") <- NULL
  }

  df
}

testthat::test_that("`drop_geometry_column()` works", {
  testthat::skip_on_cran()
  data(vi_smol, package = "spatialData")

  vi_smol

  #creating fake geometry column without sf loaded
  vi_smol$geometry <- NA

  attr(
    x = vi_smol,
    which = "sf_column"
  ) <- "geometry"

  #check new attribute
  testthat::expect_true(
    attributes(vi_smol)$sf_column == "geometry"
  )

  testthat::expect_true(
    "geometry" %in% colnames(vi_smol)
  )

  #drop geometry column
  testthat::expect_message(
    vi_smol <- drop_geometry_column(
      df = vi_smol,
      quiet = FALSE
    ),
    regexp = "dropping geometry column from 'df'"
  )

  #checking that the geometry was droppped
  testthat::expect_false(
    "geometry" %in% colnames(vi_smol)
  )

  testthat::expect_true(
    is.null(attributes(vi_smol)$sf_column)
  )
})

#' Check and constrain argument \code{max_cor}
#'
#' @inheritParams collinear_select
#' @inheritParams validate_arg_quiet
#'
#' @return numeric or NULL
#' @autoglobal
#' @export
#' @family argument_validation
#' @examples
#' x <- validate_arg_max_cor(
#'   max_cor = 1.5, #wrong value
#'   quiet = FALSE
#' )
#'
#' x
#' attributes(x)$validated
validate_arg_max_cor <- function(
  max_cor = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_max_cor()",
    function_name = function_name
  )

  max_cor_default <- 0.70

  if (isTRUE(attr(x = max_cor, which = "validated"))) {
    return(max_cor)
  }

  if (is.null(max_cor)) {
    return(NULL)
  }

  if (is.numeric(max_cor)) {
    if (length(max_cor) > 1) {
      max_cor <- max_cor[1]

      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": argument 'max_cor' must be of length one, using value '",
          max_cor,
          "'."
        )
      }
    }
  } else {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'max_cor' is non-numeric, resetting it to '",
        max_cor_default,
        "'."
      )
    }

    max_cor <- max_cor_default
  }

  if (max_cor > 1 || max_cor < 0) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'max_cor' is outside its valid range (>=0.1 to <=1), resetting it to '",
        max_cor_default,
        "'."
      )
    }

    max_cor <- max_cor_default
  }

  attr(
    x = max_cor,
    which = "validated"
  ) <- TRUE

  max_cor
}

testthat::test_that("`validate_arg_max_cor()` works", {
  testthat::skip_on_cran()

  x <- validate_arg_max_cor(
    max_cor = NULL,
    quiet = FALSE
  )

  testthat::expect_null(x)

  testthat::expect_message(
    x <- validate_arg_max_cor(
      max_cor = c(0.5, 0.6),
      quiet = FALSE
    ),
    regexp = "argument 'max_cor' must be of length one, using value '0.5'"
  )

  testthat::expect_equal(as.numeric(x), 0.5, tolerance = 1e-10)

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_no_message(
    x <- validate_arg_max_cor(
      max_cor = x,
      quiet = FALSE
    )
  )

  testthat::expect_message(
    x <- validate_arg_max_cor(
      max_cor = "hola",
      quiet = FALSE
    ),
    regexp = "argument 'max_cor' is non-numeric, resetting it to '0.7'"
  )

  testthat::expect_equal(as.numeric(x), 0.7, tolerance = 1e-10)

  testthat::expect_message(
    x <- validate_arg_max_cor(
      max_cor = 1.2,
      quiet = FALSE
    ),
    regexp = "is outside its valid range"
  )

  testthat::expect_equal(as.numeric(x), 0.7, tolerance = 1e-10)

  testthat::expect_true(
    attributes(x)$validated
  )
})

#' Check and complete argument `preference_order`
#'
#' @description
#' Internal function to validate the argument \code{preference_order} in [cor_select()], [vif_select()], [collinear_select()], [collinear()], and [collinear()]. Predictors not in \code{preference_order} are ranked from lower to higher sum of Pearson correlations with all other predictors.
#'
#' @inheritParams collinear
#' @inheritParams validate_arg_quiet
#' @inheritParams target_encoding_lab
#'
#' @return character vector: ranked variable names
#' @export
#' @family argument_validation
#' @autoglobal
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' #input arguments must be validated first
#' df <- validate_arg_df(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = vi_predictors_numeric,
#'   quiet = TRUE
#' )
#'
#' response <- validate_arg_responses(
#'   df = df,
#'   responses = "vi_numeric"
#' )
#'
#' predictors <- validate_arg_predictors(
#'   df = df,
#'   response = response,
#'   predictors = vi_predictors_numeric[1:10]
#' )
#'
#'
#'
#' #no preference order
#' #no response
#' #ranks predictor from lower to higher multicollinearity
#' y <- validate_arg_preference_order(
#'   df = df,
#'   predictors = predictors,
#'   preference_order = NULL
#' )
#'
#' y
#' attributes(y)$validated
#'
#'
#' #validate character vector
#' y <- validate_arg_preference_order(
#'   df = df,
#'   predictors = predictors,
#'   preference_order = c(
#'     "swi_max",
#'     "swi_min",
#'     "swi_deviance" #does not exist
#'   )
#' )
#'
#' y
#' attributes(y)$validated
#'
#' #validate output of preference order
#' x <- preference_order(
#'   df = df,
#'   responses = response,
#'   predictors = predictors
#' )
#'
#' x
#'
#' y <- validate_arg_preference_order(
#'   df = df,
#'   response = response,
#'   predictors = predictors,
#'   preference_order = x
#' )
#'
#' y
#' attributes(y)$validated
validate_arg_preference_order <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  preference_order = NULL,
  quiet = FALSE,
  function_name = NULL,
  ...
) {
  if (isTRUE(attr(x = preference_order, which = "validated"))) {
    return(preference_order)
  }

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_preference_order()",
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      responses = response,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  #character
  if (is.character(preference_order)) {
    #remove "auto"
    if ("auto" %in% preference_order) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": value 'auto' for the argument 'preference_order' is deprecated, ignoring it."
        )
      }

      preference_order <- setdiff(
        x = preference_order,
        y = "auto"
      )
    }

    preference_order <- intersect(
      x = preference_order,
      y = predictors
    )

    preference_order <- setdiff(
      x = preference_order,
      y = response
    )

    if (length(preference_order) == 0) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": invalid values in argument 'preference_order', ignoring it."
        )
      }

      preference_order <- NULL
    } else {
      #generate dataframe
      preference_order <- data.frame(
        response = rep(
          x = ifelse(
            test = is.null(response),
            yes = "none",
            no = response
          ),
          times = length(preference_order)
        ),
        predictor = preference_order,
        f = NA,
        metric = "user_preference",
        score = seq(
          from = 1,
          to = 0,
          length.out = length(preference_order)
        ),
        rank = seq_len(length(preference_order))
      )
    }
  }

  #dataframe
  if (is.data.frame(preference_order)) {
    #error if not expected structure
    expected_columns <- c(
      "response",
      "predictor",
      "f",
      "metric",
      "score",
      "rank"
    )

    if (!all(expected_columns %in% colnames(preference_order))) {
      stop(
        "\n",
        function_name,
        ": dataframe 'preference_order' must have these columns: '",
        paste(expected_columns, collapse = "', '"),
        "'.",
        call. = FALSE
      )
    }

    #subset preference order for the given predictors
    preference_order <- preference_order[
      preference_order$predictor %in% predictors,
    ]

    if (nrow(preference_order) == 0) {
      stop(
        "\n",
        function_name,
        ": column 'preference_order$predictor' does not contain any 'predictors'.",
        call. = FALSE
      )
    }

    if (is.null(response)) {
      unique_df_response <- unique(preference_order$response)

      if (length(unique_df_response) > 1) {
        stop(
          "\n",
          function_name,
          ": dataframe 'preference_order' contains more than one response and there is no valid 'response' argument to filter it.",
          call. = FALSE
        )
      }
    } else {
      if (response %in% preference_order$response) {
        preference_order <- preference_order[
          preference_order$response == response,
        ]
      } else {
        stop(
          "\n",
          function_name,
          ": argument 'response' does not match the column 'response' of the dataframe 'preference_order'",
          call. = FALSE
        )
      }
    }
  }

  if (is.null(preference_order)) {
    #generate empty dataframe
    preference_order <- data.frame(
      response = ifelse(
        test = is.null(response),
        yes = "none",
        no = response
      ),
      predictor = NA,
      f = NA,
      metric = NA,
      score = NA,
      rank = NA
    ) |>
      stats::na.omit()
  }

  #order missing predictors by their multicollinearity
  if (nrow(preference_order) < length(predictors)) {
    predictors_missing <- setdiff(
      x = predictors,
      y = preference_order$predictor
    )

    if (length(predictors_missing) == 1) {
      preference_order_default <- data.frame(
        response = ifelse(
          test = is.null(response),
          yes = "none",
          no = response
        ),
        predictor = predictors_missing,
        f = NA,
        metric = "user_preference",
        score = 0,
        rank = nrow(preference_order) + 1
      )
    } else if (length(predictors_missing) > 1) {
      preference_order_default <- preference_order(
        df = df,
        responses = response,
        predictors = predictors_missing,
        f = NULL,
        quiet = quiet,
        function_name = function_name
      )
    }

    preference_order <- rbind(
      preference_order,
      preference_order_default
    )
  }

  rownames(preference_order) <- NULL

  preference_order$rank <- seq_len(length.out = nrow(preference_order))

  attr(
    x = preference_order,
    which = "validated"
  ) <- TRUE

  preference_order
}

#' @rdname target_encoding_methods
#' @family target_encoding
#' @autoglobal
#' @export
target_encoding_loo <- function(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = NULL,
  ...
) {
  if (is.null(encoded_name)) {
    encoded_name <- paste0(
      predictor,
      "__encoded"
    )
  }

  #add id column to facilitate reordering
  df$id.. <- seq_len(nrow(df))

  #order data by predictor levels
  #to facilitate next block
  df <- df[order(df[[predictor]]), ]

  #leave one out
  #by group
  #sum all cases of the response
  #subtract the value of the current row
  #divide by n-1
  df$encoded <- unlist(
    lapply(
      X = split(
        x = df,
        f = df[[predictor]]
      ),
      FUN = function(x) {
        (sum(
          x = x[[response]],
          na.rm = TRUE
        ) -
          x[[response]]) /
          (nrow(x) - 1)
      }
    )
  )

  #fill groups with NaN or NA with the global mean
  df[is.na(df$encoded), "encoded"] <- mean(
    x = df[[response]],
    na.rm = TRUE
  )

  df[is.nan(df$encoded), "encoded"] <- mean(
    x = df[[response]],
    na.rm = TRUE
  )

  #rename encoded column
  names(df)[names(df) == "encoded"] <- encoded_name

  df <- df[order(df[["id.."]]), ]

  df
}

#' Quantify association between categorical variables
#'
#' @description
#'
#' Cramer's V extends the chi-squared test to quantify how strongly the categories of two variables co-occur. The value ranges from 0 to 1, where 0 indicates no  association and 1 indicates perfect association.
#'
#' This function implements a bias-corrected version of Cramer's V, which adjusts for sample size and is more accurate for small samples. However, this bias correction means that even for binary variables, Cramer's V will no equal the Pearson correlation (the standard, uncorrected Cramer's V does match Pearson for binary data).
#'
#' As the number of categories increases, Cramer's V and Pearson correlation measure increasingly different aspects of association and should not be directly compared.
#'
#' If you intend to combine these measures in a multicollinearity analysis, interpret them with care. It is often preferable to convert non-numeric variables to numeric form (for example, via target encoding) before assessing multicollinearity.
#'
#' @param x (required; vector) Values of a categorical variable (character or vector). Converted to character if numeric or logical. Default: NULL
#' @param y (required; vector) Values of a categorical variable (character or vector). Converted to character if numeric or logical. Default: NULL
#' @param check_input (required; logical) If FALSE, disables data checking for a slightly faster execution. Default: TRUE
#' @inheritParams collinear
#'
#' @return numeric: Cramer's V
#'
#' @examples
#'
#' # perfect one-to-one association
#' cor_cramer(
#'   x = c("a", "a", "b", "c"),
#'   y = c("a", "a", "b", "c")
#' )
#'
#' # still perfect: labels differ but mapping is unique
#' cor_cramer(
#'   x = c("a", "a", "b", "c"),
#'   y = c("a", "a", "b", "d")
#' )
#'
#' # high but < 1: mostly aligned, one category of y repeats
#' cor_cramer(
#'   x = c("a", "a", "b", "c"),
#'   y = c("a", "a", "b", "b")
#' )
#'
#' # appears similar by position, but no association by distribution
#' # (x = "a" mixes with y = "a" and "b")
#' cor_cramer(
#'   x = c("a", "a", "a", "c"),
#'   y = c("a", "a", "b", "b")
#' )
#'
#' # numeric inputs are coerced to character internally
#' cor_cramer(
#'   x = c(1, 1, 2, 3),
#'   y = c(1, 1, 2, 2)
#' )
#'
#' # logical inputs are also coerced to character
#' cor_cramer(
#'   x = c(TRUE, TRUE, FALSE, FALSE),
#'   y = c(TRUE, TRUE, FALSE, FALSE)
#' )
#'
#' @autoglobal
#' @family multicollinearity_assessment
#' @author Blas M. Benito, PhD
#' @references
#' \itemize{
#'  \item Cramér, H. (1946). Mathematical Methods of Statistics. Princeton: Princeton University Press, page 282 (Chapter 21. The two-dimensional case). ISBN 0-691-08004-6
#' }
#' @export
cor_cramer <- function(
  x = NULL,
  y = NULL,
  check_input = TRUE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_cramer()",
    ... = ...
  )

  #data checks
  if (check_input == TRUE) {
    # Check if 'x' is not NULL
    if (is.null(x)) {
      stop(
        "\n",
        function_name,
        ": argument 'x' cannot be NULL.",
        call. = FALSE
      )
    }

    # Check if 'y' is not NULL
    if (is.null(y)) {
      stop(
        "\n",
        function_name,
        ": argument 'y' cannot be NULL.",
        call. = FALSE
      )
    }

    #x and y have the same length
    if (length(x) != length(y)) {
      stop(
        "\n",
        function_name,
        ": arguments 'x' and 'y' must be of the same length.",
        call. = FALSE
      )
    }
  }

  #to dataframe to remove NA
  xy.df <- data.frame(
    x = as.character(x),
    y = as.character(y)
  ) |>
    stats::na.omit()

  # contingency table of 'x' and 'y'
  xy.table <- table(
    xy.df$x,
    xy.df$y
  )

  # chi-squared test with Monte Carlo simulation
  #for p-value estimation
  xy.chi <- stats::chisq.test(
    xy.table,
    simulate.p.value = TRUE
  )$statistic |>
    suppressWarnings()

  xy.chi <- tryCatch(
    {
      suppressWarnings(
        stats::chisq.test(
          xy.table,
          simulate.p.value = TRUE
        )$statistic
      )
    },
    error = function(e) {
      stop(
        function_name <- validate_arg_function_name(
          default_name = "stats::chisq.test()",
          function_name = function_name
        ),
        ": chi-squared test failed: ",
        e$message,
        call. = FALSE
      )
    }
  )

  #columns of xy.table
  xy.table.cols <- ncol(xy.table)

  #rows of xy.table
  xy.table.rows <- nrow(xy.table)

  #total sample size
  xy.table.sum <- sum(xy.table)

  #bias corrected Cramer's V
  v <- sqrt(
    max(
      c(
        0,
        (xy.chi / xy.table.sum) -
          ((xy.table.cols - 1) * (xy.table.rows - 1)) /
            (xy.table.sum - 1)
      )
    ) /
      min(
        c(
          (xy.table.cols -
            ((xy.table.cols - 1)^2 /
              (xy.table.sum - 1))) -
            1,
          (xy.table.rows -
            ((xy.table.rows - 1)^2 /
              (xy.table.sum - 1))) -
            1
        )
      )
  )

  #Cramer's V with no bias correction
  #kept here for reference
  # min_dim <- min(dim(xy.table))
  # v <- sqrt(xy.chi / (xy.table.sum * (min_dim - 1)))

  #remove names from the output
  names(v) <- NULL

  abs(v)
}

testthat::test_that("`cor_cramer()` works", {
  testthat::skip_on_cran()
  data(vi_smol, package = "spatialData")

  testthat::expect_error(
    x <- cor_cramer(
      x = NULL,
      y = vi_smol$koppen_zone
    ),
    regexp = "argument 'x' cannot be NULL"
  )

  testthat::expect_error(
    x <- cor_cramer(
      x = vi_smol$koppen_zone,
      y = NULL,
    ),
    regexp = "argument 'y' cannot be NULL"
  )

  testthat::expect_error(
    x <- cor_cramer(
      x = vi_smol$koppen_zone,
      y = vi_smol$soil_type[1:10],
    ),
    regexp = "arguments 'x' and 'y' must be of the same length"
  )

  x <- cor_cramer(
    x = vi_smol$soil_type,
    y = vi_smol$koppen_zone
  )

  testthat::expect_true(
    is.numeric(x)
  )

  testthat::expect_true(
    x >= 0 && x <= 1
  )

  # perfect one-to-one association
  x <- cor_cramer(
    x = c("a", "a", "b", "c"),
    y = c("a", "a", "b", "c")
  )

  testthat::expect_equal(x, 1, tolerance = 1e-10)

  # still perfect: labels differ but mapping is unique
  x <- cor_cramer(
    x = c("a", "a", "b", "c"),
    y = c("a", "a", "b", "d")
  )

  testthat::expect_equal(x, 1, tolerance = 1e-10)

  # high but < 1: mostly aligned, one category of y repeats
  x <- cor_cramer(
    x = c("a", "a", "b", "c"),
    y = c("a", "a", "b", "b")
  )

  testthat::expect_true(
    x < 1 & x > 0
  )

  # appears similar by position, but no association by distribution
  # (x = "a" mixes with y = "a" and "b")
  x <- cor_cramer(
    x = c("a", "a", "a", "c"),
    y = c("a", "a", "b", "b")
  )

  testthat::expect_equal(x, 0, tolerance = 1e-6)

  # numeric inputs are coerced to character internally
  x <- cor_cramer(
    x = c(1, 1, 2, 3),
    y = c(1, 1, 2, 2)
  )

  testthat::expect_true(
    x < 1 & x > 0
  )

  # logical inputs are also coerced to character
  x <- cor_cramer(
    x = c(TRUE, TRUE, FALSE, FALSE),
    y = c(TRUE, TRUE, FALSE, FALSE)
  )

  testthat::expect_equal(x, 1, tolerance = 1e-10)
})

#' Check and validate argument \code{quiet}
#'
#' @description
#' Internal function to validate the logical argument \code{quiet}, which triggers messaging when \code{FALSE}
#'
#' @inheritParams collinear
#' @param function_name (optional, character string) Name of the function performing the argument check. Default: NULL
#' @return logical
#' @examples
#'
#' x <- validate_arg_quiet(
#'   quiet = TRUE
#'   )
#'
#' attributes(x)$validated
#'
#' @autoglobal
#' @family argument_validation
#' @export
validate_arg_quiet <- function(
  quiet = FALSE,
  function_name = NULL
) {
  if (isTRUE(attr(x = quiet, which = "validated"))) {
    return(quiet)
  }

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_quiet()",
    function_name = function_name
  )

  if (is.logical(quiet)) {
    return(quiet)
  }

  message(
    "\n",
    function_name,
    ": argument 'quiet' must be logical, setting it to FALSE."
  )

  quiet <- FALSE

  attr(
    x = quiet,
    which = "validated"
  ) <- TRUE

  quiet
}

testthat::test_that("`validate_arg_quiet()` works", {
  testthat::skip_on_cran()

  testthat::expect_true(
    validate_arg_quiet(
      quiet = TRUE
    )
  )

  testthat::expect_false(
    validate_arg_quiet(
      quiet = FALSE
    )
  )

  testthat::expect_false(
    validate_arg_quiet()
  )

  testthat::expect_message(
    x <- validate_arg_quiet(
      quiet = "a"
    ),
    regexp = "argument 'quiet' must be logical"
  )

  testthat::expect_true(
    x == FALSE
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  x <- validate_arg_quiet(
    quiet = x
  )

  testthat::expect_true(
    attributes(x)$validated
  )
})

#' Build model formulas from response and predictors
#'
#' @description
#' Generates model formulas from a dataframe, a response name, and a vector of predictors that can be the output of a multicollinearity management function such as [collinear_select()] and the likes. Intended to help fit exploratory models from the result of a multicollinearity analysis.
#'
#' The types of formulas it can generate are:
#' \itemize{
#'   \item additive: \code{y ~ x + z}
#'   \item polynomial: \code{y ~ poly(x, ...) + poly(z, ...)}
#'   \item GAM: \code{y ~ s(x) + s(z)}
#'   \item random effect: \code{y ~ x + (1 \ z)}
#' }
#'
#' @inheritParams collinear
#' @inheritParams f_auto
#'
#' @param term_f (optional; string). Name of function to apply to each term in the formula, such as "s" for \code{mgcv::s()} or any other smoothing function, "poly" for \code{stats::poly()}. Default: NULL
#'
#' @param term_args (optional; string). Arguments of the function applied to each term. For example, for "poly" it can be "degree = 2, raw = TRUE". Default: NULL
#'
#' @param random_effects (optional, string or character vector). Names of variables to be used as random effects. Each element is added to the final formula as \code{+(1 | random_effect_name)}. Default: NULL
#'
#' @return list if \code{predictors} is a list or length of \code{response} is higher than one, and character vector otherwise.
#' @export
#' @autoglobal
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' #reduce collinearity
#' x <- collinear_select(
#'   df = vi_smol,
#'   predictors = vi_predictors_numeric
#' )
#'
#' #additive formula
#' y <- model_formula(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = x
#' )
#'
#' y
#'
#' #using a formula in a model
#' m <- stats::lm(
#'  formula = y,
#'  data = vi_smol
#'  )
#'
#' summary(m)
#'
#' #classification formula (character response)
#' y <- model_formula(
#'   df = vi_smol,
#'   response = "vi_categorical",
#'   predictors = x
#' )
#'
#' y
#'
#'
#' #polynomial formula (3rd degree)
#' y <- model_formula(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = x,
#'   term_f = "poly",
#'   term_args = "degree = 3, raw = TRUE"
#' )
#'
#' y
#'
#' #gam formula
#' y <- model_formula(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = x,
#'   term_f = "s"
#' )
#'
#' y
#'
#' #random effect
#' y <- model_formula(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = x,
#'   random_effects = "country_name" #from vi_smol$country_name
#' )
#'
#' y
#' @family modelling_tools
model_formula <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  term_f = NULL,
  term_args = NULL,
  random_effects = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::model_formula()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  df <- drop_geometry_column(
    df = df,
    quiet = quiet,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  if (is.null(response)) {
    stop(
      "\n",
      function_name,
      ": argument 'response' cannot be NULL.",
      call. = FALSE
    )
  }

  predictors <- validate_arg_predictors(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  if (is.null(predictors)) {
    stop(
      "\n",
      function_name,
      ": argument 'predictors' cannot be NULL.",
      call. = FALSE
    )
  }

  predictors_types <- identify_valid_variables(
    df = df,
    predictors = predictors
  )

  #prepare terms formula
  term_comma <- ","
  if (!is.null(term_f)) {
    #remove (
    term_f <- gsub(
      pattern = "\\(",
      replacement = "",
      x = term_f
    )

    term_f <- paste0(term_f, "(")

    if (is.null(term_args)) {
      term_args <- ")"
      term_comma <- NULL
    } else {
      term_args <- gsub(
        pattern = "\\)",
        replacement = "",
        x = term_args
      )

      term_args <- gsub(
        pattern = "^,",
        replacement = "",
        x = term_args
      )

      term_args <- paste0(term_args, ")")
    }
  } else {
    term_args <- NULL
    term_comma <- NULL
  }

  #random effects

  if (!is.null(random_effects)) {
    if (!is.character(random_effects)) {
      stop(
        "\n",
        function_name,
        ": argument 'random_effects' must be a character string or vector.",
        call. = FALSE
      )
    }

    random_effects <- setdiff(
      x = random_effects,
      y = unique(unlist(predictors))
    )

    if (length(random_effects) == 0) {
      stop(
        "\n",
        function_name,
        ": argument 'random_effects' must name variables not in argument 'predictors'.",
        call. = FALSE
      )
    }

    random_effects <- gsub(
      pattern = "\\)",
      replacement = "",
      x = random_effects
    )

    random_effects <- gsub(
      pattern = "\\(",
      replacement = "",
      x = random_effects
    )

    random_effects <- gsub(
      pattern = "\\+",
      replacement = "",
      x = random_effects
    )

    random_effects <- gsub(
      pattern = "\\|",
      replacement = "",
      x = random_effects
    )

    random_effects <- gsub(
      pattern = "1",
      replacement = "",
      x = random_effects
    )

    random_effects <- gsub(
      pattern = "\\s",
      replacement = "",
      x = random_effects
    )

    random_effects <- paste(
      "+",
      paste(
        "(1 |",
        random_effects,
        ")",
        collapse = " + "
      )
    )
  }

  #out list
  terms <- vector()

  #iterate over responses
  for (predictor.i in predictors) {
    if (predictor.i %in% predictors_types$numeric) {
      predictor.i <- paste0(
        term_f,
        predictor.i,
        term_comma,
        term_args
      )
    }

    terms <- c(terms, predictor.i)
  }

  out <- paste0(
    response,
    " ~ ",
    paste0(
      terms,
      collapse = " + "
    ),
    random_effects
  ) |>
    stats::as.formula()

  environment(out) <- parent.frame()

  out
}

#' Check and validate arguments `response` and `responses`
#'
#' @description Internal function validate the arguments \code{response} and \code{responses}. It checks that its value exists as a column name of \code{df},
#'
#' @inheritParams collinear
#' @inheritParams f_auto
#' @inheritParams validate_arg_quiet
#' @param max_responses (required, integer or NULL) Maximum number of responses to consider. Default: NULL
#' @return character string: response name
#' @examples
#'
#' data(vi_smol, package = "spatialData")
#'
#' x <- validate_arg_responses(
#'   df = vi_smol,
#'   responses = "vi_numeric"
#' )
#'
#' attributes(x)$validated
#'
#' @autoglobal
#' @family argument_validation
#' @export
validate_arg_responses <- function(
  df = NULL,
  responses = NULL,
  max_responses = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_responses()",
    function_name = function_name
  )

  #if df is NULL, stop
  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #return responses if NULL
  if (is.null(responses)) {
    return(NULL)
  }

  #max_responses <-
  if (is.null(max_responses)) {
    max_responses <- Inf
  } else {
    if (!is.numeric(max_responses)) {
      stop(
        "\n",
        function_name,
        ": argument 'max_responses' must be a integer.",
        call. = FALSE
      )
    }

    max_responses <- as.integer(max_responses)
  }

  #arg name for messages
  if (max_responses == 1) {
    arg_name <- "response"
  } else {
    arg_name <- "responses"
  }

  #return responses if validated
  if (
    isTRUE(attr(x = responses, which = "validated")) &&
      all(responses %in% colnames(df)) &&
      length(responses) <= max_responses
  ) {
    return(responses)
  }

  #check wrong responses
  wrong_responses <- setdiff(
    x = responses,
    y = colnames(df)
  )

  if (length(wrong_responses) > 0L) {
    if (all(responses %in% wrong_responses)) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": argument '",
          arg_name,
          "' does not contain column names of 'df' and will be ignored."
        )
      }

      return(NULL)
    }

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": the following values of argument '",
        arg_name,
        "' are not column names of 'df' and will be ignored:\n - ",
        paste(
          wrong_responses,
          collapse = "\n - "
        )
      )
    }
  }

  responses <- intersect(
    x = responses,
    y = colnames(df)
  )

  #check length
  if (length(responses) == 0) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument '",
        arg_name,
        "' does not contain column names of 'df' and will be ignored."
      )
    }

    return(NULL)
  }

  if (length(responses) > max_responses) {
    responses <- responses[1:max_responses]

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument '",
        arg_name,
        "' must be of length ",
        max_responses,
        ", using ",
        arg_name,
        ": \n - ",
        paste(
          responses,
          collapse = "\n - "
        )
      )
    }
  }

  attr(
    x = responses,
    which = "validated"
  ) <- TRUE

  responses
}

testthat::test_that("`validate_arg_responses()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  #no arguments
  testthat::expect_error(
    x <- validate_arg_responses(
      df = NULL,
      responses = NULL,
      function_name = NULL,
      quiet = FALSE
    )
  )

  #valid use case
  x <- validate_arg_responses(
    df = vi_smol,
    responses = "vi_numeric"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  #validated input
  x <- validate_arg_responses(
    df = vi_smol,
    responses = x
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  #several responses and max_responses

  #invalid
  testthat::expect_error(
    x <- validate_arg_responses(
      df = vi_smol,
      responses = c("vi_numeric", "vi_categorical", "vi_binomial"),
      max_responses = "hola"
    ),
    regexp = "argument 'max_responses' must be a integer"
  )

  #valid use case
  x <- validate_arg_responses(
    df = vi_smol,
    responses = c("vi_numeric", "vi_categorical", "vi_binomial"),
    max_responses = 3
  )

  testthat::expect_true(
    all(c("vi_numeric", "vi_categorical", "vi_binomial") %in% x)
  )

  #invalid use case
  testthat::expect_message(
    x <- validate_arg_responses(
      df = vi_smol,
      responses = c("vi_numeric", "vi_categorical", "vi_binomial"),
      max_responses = 2
    ),
    regexp = "must be of length"
  )

  testthat::expect_true(
    all(c("vi_numeric", "vi_categorical") %in% x)
  )

  #response is NULL
  x <- validate_arg_responses(
    df = vi_smol,
    responses = NULL
  )

  testthat::expect_true(
    is.null(x)
  )

  #response not in df
  testthat::expect_message(
    x <- validate_arg_responses(
      df = vi_smol,
      responses = "hola"
    ),
    regexp = "argument 'responses' does not contain column names of 'df'"
  )

  testthat::expect_true(
    is.null(x)
  )

  testthat::expect_message(
    x <- validate_arg_responses(
      df = vi_smol,
      responses = c("hola", "vi_numeric")
    ),
    regexp = "the following values of argument 'responses' are not column names of 'df' and will be ignored"
  )

  testthat::expect_true(
    x == "vi_numeric"
  )

  #wrong response and max_responses
  testthat::expect_message(
    x <- validate_arg_responses(
      df = vi_smol,
      responses = "hola",
      max_responses = 1
    ),
    regexp = "argument 'response' does not contain column names of 'df'"
  )

  testthat::expect_true(
    is.null(x)
  )

  #all null
  testthat::expect_error(
    response <- validate_arg_responses(
      df = NULL,
      responses = NULL
    )
  )
})

#' R-squared of Poisson GLM predictions vs. observations
#'
#' @description
#' Fits a Poisson GLM model \code{y ~ x} with the numeric response \code{y} and the numeric predictor \code{x} using \code{stats::glm()} and returns the R-squared of the observations against the predictions (see [score_r2()]).
#'
#' Supports cross-validation via the arguments arguments \code{cv_training_fraction} (numeric between 0 and 1) and \code{cv_iterations} (integer between 1 and \code{n}) introduced via ellipsis (\code{...}). See [preference_order()] for further details.
#'
#' @param df (required, dataframe) with columns:
#' \itemize{
#'   \item "x": (numeric, character, factor) predictor.
#'   \item "y" (integer) counts response.
#' }
#' @inheritParams f_numeric_glm
#' @return numeric or numeric vector: R-squared
#' @examples
#'
#' data(vi_smol, package = "spatialData")
#'
#' df <- data.frame(
#'   y = vi_smol[["vi_counts"]],
#'   x = vi_smol[["swi_max"]]
#' )
#'
#' #no cross-validation
#' f_count_glm(df = df)
#'
#' #cross-validation
#' f_count_glm(
#'   df = df,
#'   cv_training_fraction = 0.5,
#'   cv_iterations = 10
#'   )
#'
#' #categorical predictor
#' df <- data.frame(
#'   y = vi_smol[["vi_counts"]],
#'   x = vi_smol[["koppen_zone"]]
#' )
#'
#' f_count_glm(df = df)
#'
#' @family preference_order_functions
#' @autoglobal
#' @export
f_count_glm <- function(
  df,
  ...
) {
  dots <- list(...)

  cv_training_fraction <- if (is.null(dots$cv_training_fraction)) {
    1
  } else {
    dots$cv_training_fraction
  }

  cv_iterations <- if (is.null(dots$cv_iterations)) 1 else dots$cv_iterations

  function_name <- validate_arg_function_name(
    default_name = "collinear::f_count_glm()",
    function_name = dots$function_name
  )

  df <- stats::na.omit(object = df)

  #check column names in df
  if (!all(c("x", "y") %in% names(df))) {
    stop(
      "\n",
      function_name,
      ": dataframe 'df' must have the column names 'x' and 'y'.",
      call. = FALSE
    )
  }

  if (
    !is.numeric(df[["y"]]) ||
      !is.integer(df[["y"]])
  ) {
    stop(
      "\n",
      function_name,
      ": column 'y' of dataframe 'df' must be numeric integer.",
      call. = FALSE
    )
  }

  scores <- rep(x = NA_real_, times = cv_iterations)

  #iterations
  for (i in seq_len(cv_iterations)) {
    #data split
    if (cv_training_fraction < 1) {
      train_indices <- sample(
        x = nrow(df),
        size = floor(nrow(df) * cv_training_fraction),
        replace = FALSE
      )

      train_df <- df[train_indices, , drop = FALSE]
      test_df <- df[-train_indices, , drop = FALSE]

      if (!is.numeric(df[["x"]])) {
        test_df <- test_df[test_df[["x"]] %in% train_df[["x"]], ]
      }
    } else {
      train_df <- df
      test_df <- df
    }

    if (nrow(test_df) < 2) {
      next
    }

    scores[i] <- tryCatch(
      {
        #train
        m <- stats::glm(
          formula = y ~ x,
          data = train_df,
          family = stats::poisson(
            link = "log"
          )
        )

        #predict
        p <- stats::predict(
          object = m,
          newdata = test_df,
          type = "response"
        )

        #evaluate
        score <- score_r2(
          o = test_df[["y"]],
          p = p,
          function_name = function_name
        )
      },
      error = function(e) {
        return(NA)
      }
    )
  }

  scores
}

testthat::test_that("f_count_glm() works", {
  testthat::skip_on_cran()
  data(vi_smol, package = "spatialData")

  response <- "vi_counts"
  numeric_predictor <- "swi_mean"
  categorical_predictor <- "koppen_zone"
  factor_predictor <- "vi_factor"

  #numeric predictor ----

  #perfect result
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[response]]
  )

  testthat::expect_true(
    f_count_glm(df = df) > 0.93
  )

  #imperfect result
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[numeric_predictor]]
  )

  testthat::expect_true(
    f_count_glm(df = df) < 1
  )

  #categorical response and predictor ----
  df <- data.frame(
    y = vi_smol[[categorical_predictor]],
    x = vi_smol[[categorical_predictor]]
  )

  testthat::expect_error(
    f_count_glm(df = df),
    regexp = "column 'y' of dataframe 'df' must be numeric"
  )

  #categorical predictor ----
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[categorical_predictor]]
  )

  testthat::expect_true(
    f_count_glm(df = df) < 1
  )

  #wrong column names ----
  df <- data.frame(
    a = vi_smol[["vi_categorical"]],
    b = vi_smol[["koppen_zone"]]
  )

  testthat::expect_error(
    f_count_glm(df = df),
    regexp = "dataframe 'df' must have the column names 'x' and 'y'."
  )

  #NA in data
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[categorical_predictor]]
  )
  df$x[1:5] <- NA
  df$y[6:10] <- NA

  x <- f_count_glm(df = df)

  testthat::expect_true(
    is.numeric(x)
  )

  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[numeric_predictor]]
  )
  df$x[1:5] <- NA
  df$y[6:10] <- NA

  x <- f_count_glm(df = df)

  testthat::expect_true(
    is.numeric(x)
  )

  #cross validation ----

  #one iteration
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[numeric_predictor]]
  )

  x0 <- f_count_glm(df = df)

  set.seed(1)

  x1 <- f_count_glm(
    df = df,
    cv_training_fraction = 1,
    cv_iterations = 1
  )

  testthat::expect_equal(x0, x1, tolerance = 1e-10)

  x2 <- f_count_glm(
    df = df,
    cv_training_fraction = 0.9,
    cv_iterations = 100
  )

  testthat::expect_true(
    length(x2) == 100
  )

  x3 <- f_count_glm(
    df = df,
    cv_training_fraction = 0.5,
    cv_iterations = 100
  )

  testthat::expect_true(
    length(x3) == 100
  )

  testthat::expect_true(
    mean(x3) < mean(x2)
  )

  x4 <- f_count_glm(
    df = df,
    cv_training_fraction = 0.25,
    cv_iterations = 100
  )

  testthat::expect_true(
    length(x4) == 100
  )

  testthat::expect_true(
    mean(x4) < mean(x2)
  )

  set.seed(2)

  x5 <- f_count_glm(
    df = df,
    cv_training_fraction = 0.25,
    cv_iterations = 100
  )

  testthat::expect_true(
    mean(x4) != mean(x5)
  )

  set.seed(2)

  x6 <- f_count_glm(
    df = df,
    cv_training_fraction = 0.25,
    cv_iterations = 100
  )

  testthat::expect_equal(mean(x5), mean(x6), tolerance = 1e-10)

  #categorical predictor
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[categorical_predictor]]
  )

  x_categorical <- f_count_glm(
    df = df,
    cv_training_fraction = 0.5,
    cv_iterations = 100
  )

  testthat::expect_true(
    length(x_categorical) == 100
  )

  #factor predictor
  df <- data.frame(
    y = vi_smol[[response]],
    x = vi_smol[[factor_predictor]]
  )

  x_factor <- f_count_glm(
    df = df,
    cv_training_fraction = 0.5,
    cv_iterations = 100
  )

  testthat::expect_true(
    length(x_factor) == 100
  )
})

#' Toy dataframe with varying levels of multicollinearity
#'
#' Dataframe with known relationship between responses and predictors useful to illustrate multicollinearity concepts.
#'
#' Columns:
#' \itemize{
#'   \item \code{y}: response variable generated from \code{a * 0.75 + b * 0.25 + noise}.
#'   \item \code{a}: most important predictor of \code{y}, uncorrelated with \code{b}.
#'   \item \code{b}: second most important predictor of \code{y}, uncorrelated with \code{a}.
#'   \item \code{c}: generated from \code{a + noise}.
#'   \item \code{d}: generated from \code{(a + b)/2 + noise}.
#' }
#'
#' These are variance inflation factors of the predictors in \code{toy}.
#' variable vif
#'  b       4.062
#'  d       6.804
#'  c       13.263
#'  a       16.161
#'
#' @usage data(toy)
#'
#' @format dataframe with 2000 rows and 5 columns.
#' @family example_data
"toy"

#' Dataframe resulting from experiment to test the automatic selection of multicollinearity thresholds
#'
#' A dataframe summarizing 10,000 experiments validating the adaptive multicollinearity threshold system in [collinear()]. Each row records input data characteristics and the resulting multicollinearity metrics after filtering.
#'
#' The source data is a synthetic dataframe with 500 columns and 10,000 rows generated using \code{distantia::zoo_simulate()} with correlated time series (\code{independent = FALSE}, \code{seasons = 0}).
#'
#' Each iteration randomly subsets 10-100 predictors and 30-100 rows per predictor, then applies [collinear()] with automatic threshold configuration to assess:
#' \itemize{
#'   \item Whether output VIF stays bounded between ~2.5 and ~7.5
#'   \item How the system adapts to different correlation structures
#'   \item How predictor retention scales with input size
#' }
#'
#' @format A dataframe with 10,000 rows and 9 variables:
#' \describe{
#'   \item{input_rows}{Number of rows in the input data subset.}
#'   \item{input_predictors}{Number of predictors in the input data subset.}
#'   \item{output_predictors}{Number of predictors retained after filtering.}
#'   \item{input_cor_q75}{75th percentile of pairwise correlations in the input data.}
#'   \item{output_cor_q75}{75th percentile of pairwise correlations in the selected predictors.}
#'   \item{input_cor_max}{Maximum pairwise correlation in the input data.}
#'   \item{output_cor_max}{Maximum pairwise correlation in the selected predictors.}
#'   \item{input_vif_max}{Maximum VIF in the input data.}
#'   \item{output_vif_max}{Maximum VIF in the selected predictors.}
#' }
#' @family experiments
#' @usage data(experiment_adaptive_thresholds)
#' @examples
#' data(experiment_adaptive_thresholds)
#' str(experiment_adaptive_thresholds)
"experiment_adaptive_thresholds"

#' Dataframe with results of experiment comparing correlation and VIF thresholds
#'
#' A dataframe summarizing 10,000 experiments comparing the output of [cor_select()] and [vif_select()]. Each row records the input sampling parameters and the resulting feature-selection metrics.
#'
#' The source data is a synthetic dataframe with 500 columns and 10,000 rows generated using \code{distantia::zoo_simulate()} with correlated time series (\code{independent = FALSE}).
#'
#' Each iteration randomly subsets 10-50 predictors and 30-100 rows per predictor, applies [cor_select()] with a random \code{max_cor} threshold, then finds the \code{max_vif} value that maximizes Jaccard similarity between the two selections.
#'
#' @format A dataframe with 10,000 rows and 6 variables:
#' \describe{
#'   \item{input_rows}{Number of rows in the input data subset.}
#'   \item{input_predictors}{Number of predictors in the input data subset.}
#'   \item{output_predictors}{Number of predictors selected by [vif_select()] at the best-matching \code{max_vif}.}
#'   \item{max_cor}{Maximum allowed pairwise correlation supplied to [cor_select()].}
#'   \item{max_vif}{VIF threshold at which [vif_select()] produced the highest Jaccard similarity with [cor_select()] for the given \code{max_cor}.}
#'   \item{out_selection_jaccard}{Jaccard similarity between the predictors selected by [cor_select()] and [vif_select()].}
#' }
#' @family experiments
#' @usage data(experiment_cor_vs_vif)
#' @examples
#' data(experiment_cor_vs_vif)
#' str(experiment_cor_vs_vif)
"experiment_cor_vs_vif"


#' GAM describing the relationship between correlation and VIF thresholds
#'
#' A fitted generalized additive model describing \code{max_vif} as a function of \code{max_cor} in [experiment_cor_vs_vif].
#'
#' The model parameters (basis dimension \code{k} and weight exponent) were selected via optimization, filtering for models in the top 90\% of R-squared and bottom 10\% of effective degrees of freedom to balance fit quality and parsimony.
#'
#' The final model uses squared Jaccard similarity as weights to emphasize cases with high agreement between [cor_select()] and [vif_select()].
#'
#'
#' Model performance:
#' \itemize{
#'   \item Adjusted R-squared: 0.834
#'   \item Deviance explained: 83.4\%
#'   \item Effective degrees of freedom for smooth: ~6
#' }
#'
#' @format A \code{\link[mgcv]{gam}} object.
#' @family experiments
#' @usage data(gam_cor_to_vif)
#' @source Generated internally from \link{experiment_cor_vs_vif}.
#' @examples
#' data(gam_cor_to_vif)
#' plot(gam_cor_to_vif, shade = TRUE)
"gam_cor_to_vif"


#' Prediction of the model `gam_cor_to_vif` across correlation values
#'
#' Dataframe with predicted VIF threshold corresponding to a given correlation threshold..
#'
#' Values were generated by applying \code{mgcv::predict.gam()} to the fitted model [gam_cor_to_vif] and rounding to three decimal places.
#'
#' @format A dataframe with 901 rows and 2 numeric columns:
#' \describe{
#'   \item{max_cor}{Maximum allowed pairwise correlation, from 0.10 to 1.00 in steps of 0.001.}
#'   \item{max_vif}{Predicted VIF threshold corresponding to each \code{max_cor}.}
#' }
#'
#' @family experiments
#' @usage data(prediction_cor_to_vif)
#' @examples
#' data(prediction_cor_to_vif)
#' head(prediction_cor_to_vif)
#' plot(max_vif ~ max_cor, data = prediction_cor_to_vif, type = "l")
"prediction_cor_to_vif"

#' Compute R-squared between numeric observations and predictions
#'
#' @description
#' Internal function to compute the R-squared of observations versus predictions via \code{stats::cor()}. Used within [f_numeric_glm()], [f_numeric_gam()], [f_numeric_rf()], [f_count_glm()], and [f_count_gam()].
#'
#' @param o (required, numeric vector) Observations. Default: NULL
#' @param p (required, numeric vector) Predictions. Default: NULL
#' @inheritParams collinear
#'
#' @return numeric: Pearson R-squared
#' @export
#' @autoglobal
#' @family modelling_tools
#' @examples
#'   score_r2(
#'     o = c(1, 1, 1, 0.5, 0.5, 0, 0),
#'     p = c(1, 0.8, 0.7, 0.6, 0.5, 0.1, 0)
#'   )
score_r2 <- function(
  o = NULL,
  p = NULL,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::score_r2()",
    ... = ...
  )

  out <- tryCatch(
    {
      stats::cor(
        x = p,
        y = o,
        use = "complete.obs",
        method = "pearson"
      )^2
    },
    error = function(e) {
      stop(
        "\n",
        function_name,
        ": ",
        conditionMessage(e),
        call. = FALSE
      )
    }
  )

  out
}

testthat::test_that("`score_r2()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  #perfect prediction
  x <- score_r2(
    o = vi_smol$vi_numeric,
    p = vi_smol$vi_numeric
  )

  testthat::expect_equal(x, 1, tolerance = 1e-10)

  #random prediction
  x <- score_r2(
    o = vi_smol$vi_binomial,
    p = runif(n = nrow(vi_smol))
  )

  testthat::expect_true(
    x < 1
  )

  #error
  testthat::expect_error(
    x <- score_r2(
      o = NULL,
      p = vi_smol$vi_binomial
    )
  )

  testthat::expect_error(
    x <- score_r2(
      o = vi_smol$vi_binomial,
      p = NULL
    )
  )
})

#' @title Generate sample weights for imbalanced responses
#' @description Computes case weights adding to one for response variables of these types:
#'
#' \itemize{
#'   \item numeric binomial (1 and 0).
#'   \item logical (TRUE and FALSE): converted to numeric internally.
#'   \item categorical (character or factor)
#' }
#'
#' Values NA, Inf, -Inf, and NaN are invalid for numeric and logical variables and will result in errors. For categorical variables, these are converted to their respective categories ("NA", "Inf", "-Inf", and "NaN") with their assigned case weights.
#'
#' All returned weights sum to one.
#'
#' @param x (required, integer, character, or factor vector) Values of a binomial, categorical, or factor variable. Default: NULL
#' @inheritParams collinear
#' @return numeric vector: case weights
#' @examples
#'  #numeric vector
#'  y <- case_weights(
#'    x = c(0, 0, 1, 1)
#'    )
#'
#'  #logical vector
#'  y <- case_weights(
#'    x = c(TRUE, TRUE, FALSE, FALSE)
#'    )
#'
#'  #character vector
#'  y <- case_weights(
#'    x = c("a", "a", "b", "c")
#'    )
#' @family modelling_tools
#' @autoglobal
#' @export
case_weights <- function(
  x = NULL,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::case_weights()",
    ... = ...
  )

  if (is.null(x)) {
    stop(
      "\n",
      function_name,
      ": argument 'x' cannot be NULL.",
      call. = FALSE
    )
  }

  if (is.logical(x)) {
    x <- as.numeric(x)
  }

  if (is.numeric(x) && any(!is.finite(x))) {
    stop(
      "\n",
      function_name,
      ": NA, Inf, -Inf, or NaN values are not allowed when 'x' is numeric or logical.",
      call. = FALSE
    )
  }

  #convert NA, Inf, -Inf, or NaN to character
  if (is.character(x)) {
    x <- paste(x)
  }

  #weights per class as inverse of the counts
  weights <- 1 / table(as.factor(x))

  #full vector of weights
  out <- weights[as.character(x)]

  #normalize to one
  out <- out / sum(out)

  if (length(out) != length(x)) {
    stop(
      "\n",
      function_name,
      ": something went wrong when computing case weights, the input length is ",
      length(x),
      ", but the output length is ",
      length(out),
      ".",
      call. = FALSE
    )
  }

  out
}

testthat::test_that("`case_weights()` works", {
  testthat::skip_on_cran()

  tol <- 0.000001

  #algorithmic changes
  #numeric vector
  y <- case_weights(
    x = c(0, 0, 1, 1)
  )

  testthat::expect_true(
    all(y - 0.25 < tol)
  )

  testthat::expect_true(
    sum(y) - 1 < tol
  )

  #logical vector
  y <- case_weights(
    x = c(TRUE, TRUE, FALSE, FALSE)
  )

  testthat::expect_true(
    all(y - 0.25 < tol)
  )

  testthat::expect_true(
    sum(y) - 1 < tol
  )

  #character vector
  y <- case_weights(
    x = c("a", "a", "b", "c")
  )

  testthat::expect_true(
    sum(y) - 1 < tol
  )

  testthat::expect_true(
    min(y) - 0.1666667 < tol
  )

  testthat::expect_true(
    max(y) - 0.3333333 < tol
  )

  #other behaviors
  testthat::expect_error(
    weights <- case_weights(
      x = NULL
    ),
    regexp = "argument 'x' cannot be NULL"
  )

  #works with logical
  weights <- case_weights(
    x = c(TRUE, TRUE, FALSE, FALSE)
  )

  testthat::expect_true(
    all(weights - 0.25 < tol)
  )

  testthat::expect_true(
    sum(weights) - 1 < tol
  )

  #controlled values
  weights <- case_weights(
    x = c(1, 1, 2, 2)
  )

  #error
  testthat::expect_error(
    weights <- case_weights(
      x = c(NA, TRUE, FALSE, FALSE)
    ),
    regexp = "NA, Inf, -Inf, or NaN values are not allowed when 'x' is numeric or logical"
  )

  testthat::expect_true(
    all(weights - 0.25 < tol)
  )

  testthat::expect_true(
    sum(weights) - 1 < tol
  )

  #errors
  testthat::expect_error(
    weights <- case_weights(
      x = c(NA, NaN, Inf, -Inf, 2, 2)
    ),
    regexp = "NA, Inf, -Inf, or NaN values are not allowed when 'x' is numeric or logical"
  )

  data(vi, package = "spatialData")

  #binomial weights
  weights <- case_weights(
    x = vi$vi_binomial
  )

  testthat::expect_true(
    length(unique(weights)) == 2
  )

  testthat::expect_true(
    sum(weights) - 1 < tol
  )

  #multinomial weights
  weights <- case_weights(
    x = vi$vi_categorical
  )

  testthat::expect_true(
    all(names(weights) %in% unique(vi$vi_categorical))
  )

  testthat::expect_true(
    sum(weights) - 1 < tol
  )

  #non valid values in multinomial
  vi$vi_categorical[1:4] <- c(NA, Inf, -Inf, NaN)

  weights <- case_weights(
    x = vi$vi_categorical
  )

  testthat::expect_true(
    all(c("NA", "Inf", "-Inf", "NaN") %in% names(weights))
  )
})

#' Rank predictors by importance or multicollinearity
#'
#' @description
#' Generates a valid input for the argument \code{preference_order} of the functions [vif_select()], [cor_select()], [collinear_select()], and [collinear()]. This argument helps preserve important predictors during multicollinearity filtering.
#'
#' The function works in two different ways:
#' \itemize{
#'   \item When \code{f} is NULL, it ranks the predictors from lower to higher multicollinearity, computed as one minus the average Pearson correlation between the given predictor against all others. This option is useful when the goal is to limit redundancy in a large dataset and there is not an specific model to train in mind.
#'   \item When \code{responses} and \code{f} are not NULL, it ranks the predictors by the strength of their association with a response based on the evaluation of univariate models. This is the best possible option when the end-goal is training a model.
#' }
#'
#' The argument \code{f} (requires a valid \code{resopnses} argument) defines how the strength of association between the response and each predictor is computed. By default it calls [f_auto()], which uses [f_auto_rules()] to select a suitable function depending on the types of the response and the predictors. This option is designed to provide sensible, general-purpose defaults optimized for speed and stability rather than any specific modeling approach.
#'
#' For more fine-tuned control, the package offers the following \code{f} functions (see [f_functions()]):
#'
#' \itemize{
#'   \item **Numeric response**:
#'   \itemize{
#'    \item [f_numeric_glm()]: Pearson's R-squared of response versus the predictions of a Gaussian GLM.
#'    \item [f_numeric_gam()]: GAM model fitted with \code{mgcv::gam()}.
#'    \item [f_numeric_rf()]: Random Forest model fitted with \code{ranger::ranger()}.
#'   }
#'   \item **Integer counts response**:
#'   \itemize{
#'     \item [f_count_glm()]: Pearson's R-squared of a Poisson GLM.
#'     \item [f_count_gam()]: Poisson GAM.
#'     \item [f_count_rf()]: Random Forest model fitted with \code{ranger::ranger()}.
#'   }
#'   \item **Binomial response (1 and 0)**:
#'   \itemize{
#'     \item [f_binomial_glm()]: AUC of Quasibinomial GLM with weighted cases.
#'     \item [f_binomial_gam()]: AUC of Quasibinomial GAM with weighted cases.
#'     \item [f_binomial_rf()]: AUC of a Random Forest model with weighted cases.
#'   }
#'   \item **Categorical response**:
#'   \itemize{
#'     \item [f_categorical_rf()]: Cramer's V of the response against the predictions of a classification Random Forest model.
#'   }
#' }
#'
#' These functions accept a cross-validation setup via the arguments \code{cv_iterations} and \code{cv_training_fraction}.
#'
#' Additionally, the argument \code{f} accepts any custom function taking a dataframe with the columns "x" (predictor) and "y" (response) and returning a numeric indicator of association.
#'
#' Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).
#'
#' Accepts a character vector of response variables as input for the argument \code{responses}. When more than one response is provided, the output is a named list of preference data frames.
#'
#' @inheritParams collinear
#'
#' @param f (optional: function name) Unquoted function name without parenthesis (see [f_functions]). By default calls to [f_auto()], which selects a suitable function depending on the nature of the response and predictors. Set to NULL if \code{responses = NULL}. If NULL, predictors are ranked from lower to higher multicollinearity. Default: \code{f_auto}
#'
#' @param cv_training_fraction (optional, numeric) Value between 0.1 and 1 defining the training faction used in cross-validation. If 1 (default), no cross-validation is performed, and the resulting metric is computed from all observations and predictions. Automatically set to 1 when \code{cv_iterations = 1}. Default: 1
#'
#' @param cv_iterations (optional, integer) Number of cross-validation iterations to perform. The recommended range lies between 30 and 100. In general, smaller datasets and large values of \code{cv_training_fraction} require more iterations to achieve stability. Automatically set to 1 when \code{cv_training_fraction = 1}. Default: 1
#'
#' @param seed (optional, integer) Random seed, required for reproducibility when using cross-validation or random forest models. Default: 1
#'
#'
#' @family preference_order_functions
#' @return dataframe:
#' \itemize{
#'   \item \code{response}: character, response name, if any, or \code{"none"} otherwise.
#'
#'   \item \code{predictor}: character, name of the predictor.
#'
#'   \item \code{f}: name of the function used to compute the preference order. If argument \code{f} is NULL, the value "stats::cor()" is added to this column.
#'
#'   \item \code{metric}: name of the metric used to assess strength of association. Usually one of "R-squared", "AUC" (Area Under the ROC Curve), or "Cramer's V". If \code{f} is a custom function not in [f_functions()], then \code{metric} is set to "custom". If \code{f} is NULL, then "1 - R-squared" is returned in this column.
#'
#'   \item \code{score}: value of the metric returned by \code{f} to assess the association between the \code{response} and each given \code{predictor}.
#'
#' \item \code{rank}: integer value indicating the rank of the predictor.

#' }
#' @examples
#' #load example data
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' ##OPTIONAL: parallelization setup
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ##OPTIONAL: progress bar
#' ##does not work in R examples
#' # progressr::handlers(global = TRUE)
#'
#' #ranking predictors from lower to higher multicollinearity
#' #------------------------------------------------
#' x <- preference_order(
#'   df = vi_smol,
#'   responses = NULL, #default value
#'   predictors = vi_predictors_numeric[1:10],
#'   f = NULL #must be explicit
#' )
#'
#' x
#'
#' #automatic selection of ranking function
#' #------------------------------------------------
#' x <- preference_order(
#'   df = vi_smol,
#'   responses = c("vi_numeric", "vi_categorical"),
#'   predictors = vi_predictors_numeric[1:10],
#'   f = f_auto
#'   )
#'
#' x
#'
#' #user selection of ranking function
#' #------------------------------------------------
#' #Poisson GLM for a integer counts response
#' x <- preference_order(
#'   df = vi_smol,
#'   responses = "vi_binomial",
#'   predictors = vi_predictors_numeric[1:10],
#'   f = f_binomial_glm
#' )
#'
#' x
#'
#' #cross-validation
#' #------------------------------------------------
#' x <- preference_order(
#'   df = vi_smol,
#'   responses = "vi_binomial",
#'   predictors = vi_predictors_numeric[1:10],
#'   f = f_binomial_glm,
#'   cv_training_fraction = 0.5,
#'   cv_iterations = 10
#' )
#'
#' x
#'
#' #custom pairwise correlation function
#' #------------------------------------------------
#' #custom functions need the ellipsis argument
#' f_rsquared <- function(df, ...){
#'     stats::cor(
#'       x = df$x,
#'       y = df$y,
#'       use = "complete.obs"
#'     )^2
#' }
#'
#' x <- preference_order(
#'   df = vi_smol,
#'   responses = "vi_numeric",
#'   predictors = vi_predictors_numeric[1:10],
#'   f = f_rsquared
#' )
#'
#' x
#'
#' #resetting to sequential processing
#' #future::plan(future::sequential)
#' @autoglobal
#' @author Blas M. Benito, PhD
#' @export
preference_order <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  f = f_auto,
  cv_training_fraction = 1,
  cv_iterations = 1,
  seed = 1,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::preference_order()",
    function_name = dots$function_name
  )

  quiet <- validate_arg_quiet(
    function_name = function_name,
    quiet = quiet
  )

  #check input dataframe
  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #managing multiple responses
  responses <- validate_arg_responses(
    df = df,
    responses = responses,
    quiet = quiet,
    function_name = function_name
  )

  predictors <- validate_arg_predictors(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df.ncol <- ncol(df)

  df <- validate_arg_df(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #revalidate predictors if any columns were removed
  if (ncol(df) < df.ncol) {
    attributes(responses)$validated <- NULL
    attributes(predictors)$validated <- NULL

    response <- validate_arg_responses(
      df = df,
      responses = responses,
      max_responses = 1,
      quiet = quiet,
      function_name = function_name
    )

    predictors <- validate_arg_predictors(
      df = df,
      responses = responses,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #check f
  if (
    is.null(responses) &&
      !is.null(f)
  ) {
    f <- NULL
  }

  f <- validate_arg_f(
    f = f,
    f_name = sub(".*::", "", deparse(substitute(f))),
    function_name = function_name
  )

  #DEFAULT
  if (is.null(f)) {
    if (is.null(responses)) {
      responses <- "none"
    }

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": ranking ",
        length(predictors),
        " 'predictors' from lower to higher multicollinearity."
      )
    }

    m <- cor_matrix(
      df = df,
      predictors = predictors,
      function_name = function_name,
      quiet = quiet,
      m = dots$m
    ) |>
      abs()

    diag(x = m) <- 0

    m.colmeans <- sort(
      x = 1 - colMeans(m),
      decreasing = TRUE
    ) |>
      round(digits = 4)

    m.names <- names(m.colmeans)

    names(m.colmeans) <- NULL

    out <- data.frame(
      response = sort(
        rep(
          x = responses,
          times = length(m.names) * length(responses)
        )
      ),

      predictor = rep(
        x = m.names,
        times = length(responses)
      ),

      f = rep(
        x = "stats::cor()",
        times = length(m.names) * length(responses)
      ),

      metric = rep(
        x = "1 - R-squared",
        times = length(m.names) * length(responses)
      ),

      score = rep(
        x = m.colmeans,
        times = length(responses)
      ),

      rank = rep(
        x = seq_len(length.out = length(m.names)),
        times = length(responses)
      )
    )

    return(out)
  }

  if (
    !is.numeric(cv_iterations) ||
      cv_iterations < 1 ||
      cv_iterations != floor(cv_iterations)
  ) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'cv_iterations' must be a positive integer, setting it to '1'"
      )
    }

    cv_iterations <- 1
    cv_training_fraction <- 1
  }

  #check cross validation arguments
  if (
    !is.numeric(cv_training_fraction) ||
      cv_training_fraction < 0.1 ||
      cv_training_fraction > 1
  ) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'cv_training_fraction' must be a numeric between 0.1 and 1 (inclusive), setting it to '1'."
      )
    }

    cv_training_fraction <- 1
    cv_iterations <- 1
  }

  cv_training_fraction <- cv_training_fraction[1]
  cv_iterations <- as.integer(cv_iterations[1])

  if (
    !is.numeric(seed) ||
      seed != floor(seed)
  ) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'seed' is invalid, resetting it to '1'.",
        call. = FALSE
      )
    }

    seed <- 1L
  }

  f_df <- f_functions()[, c("name", "metric")]
  colnames(f_df) <- c("f", "metric")

  #output list
  out_list <- list()

  #iterating over responses
  for (response in responses) {
    set.seed(seed)

    if (
      quiet == FALSE &&
        length(responses) > 1
    ) {
      msg <- paste0(
        function_name,
        ": processing response '",
        response,
        "'."
      )

      msg_length <- nchar(msg)

      message("\n", msg)
      message(rep(x = "-", times = nchar(msg)))
    }

    #use f_auto to get function name
    if (
      all(c("df", "response", "predictors", "quiet") %in% names(formals(f)))
    ) {
      #get function name
      f_name <- f(
        df = df,
        response = response,
        predictors = predictors,
        quiet = quiet,
        function_name = function_name
      )

      #validate to function
      f.response <- validate_arg_f(
        f = get(f_name),
        f_name = f_name,
        function_name = function_name
      )
    } else {
      f.response <- f
    }

    #dataframe to store results
    preference <- data.frame(
      response = rep(
        x = response,
        times = length(predictors)
      ),
      predictor = predictors,
      f = rep(
        x = attributes(f.response)$name,
        times = length(predictors)
      )
    )

    #progress bar
    p <- progressr::progressor(
      steps = nrow(preference)
    )

    #computing preference order
    preference$score <- future.apply::future_lapply(
      X = preference$predictor,
      FUN = function(x) {
        p()

        out <- f.response(
          df = stats::na.omit(data.frame(
            y = df[[response]],
            x = df[[x]]
          )),
          function_name = function_name,
          cv_training_fraction = cv_training_fraction,
          cv_iterations = cv_iterations
        ) |>
          mean(na.rm = TRUE)

        if (is.na(out)) {
          warning(
            "\n",
            function_name,
            ": function '",
            attributes(f.response)$name,
            "' produced NA scores for response '",
            response,
            " and predictor '",
            x,
            "."
          )
        }

        out
      },
      future.seed = TRUE
    ) |>
      unlist() |>
      suppressWarnings()

    #reorder preference
    preference <- preference[
      order(
        preference$score,
        decreasing = TRUE
      ),
    ]

    rownames(preference) <- NULL

    #join metric if function is in f_functions()
    if (unique(preference[["f"]]) %in% unique(f_df[["f"]])) {
      preference <- merge(
        x = preference,
        y = f_df,
        by = "f"
      )
    } else {
      preference$metric <- "custom"
    }

    preference$rank <- seq_len(length.out = nrow(preference))

    out_list[[response]] <- preference
  } #end of loop

  out_df <- do.call(
    what = "rbind",
    args = out_list
  )

  rownames(out_df) <- NULL

  out_df <- out_df[, c("response", "predictor", "f", "metric", "score", "rank")]

  out_df$score <- round(
    x = out_df$score,
    digits = 4
  )

  out_df
}

#' @title Tidymodels recipe step for multicollinearity filtering
#'
#' @description
#'
#' Adds a step to a recipe created by \code{recipes::recipe()]+} to apply multicollinearity filtering via [collinear()].
#'
#' This function requires the \code{recipes} package to be installed.
#'
#'Unlike \code{collinear()}, this wrapper does not perform target encoding, and the default value for the argument \code{quiet} is \code{TRUE}.
#'
#' @param recipe (required, recipe) A recipe object to which this step will be added.
#' @param ... (optional, selector functions) Variables to consider. Typically [recipes::all_predictors()] or [recipes::all_numeric_predictors()].
#' @param role (optional, character) Not used by this step since no new variables are created. Default: NA
#' @param trained (optional, logical) Indicates if the step has been trained. Default: FALSE
#' @param options (optional, list) Named list of arguments passed to [collinear::collinear()]. Common options include:
#'   \itemize{
#'     \item \code{max_cor}: Maximum correlation threshold. If NULL (default), automatically set based on median correlation of predictors.
#'     \item \code{max_vif}: Maximum VIF threshold. If NULL (default), automatically set to match the auto-computed max_cor.
#'     \item \code{preference_order}: Vector of predictor names in priority order.
#'     \item \code{f}: Function to compute preference order (default: \code{f_auto})
#'     \item \code{quiet}: Suppress messages (default: TRUE)
#'   }
#'   Note: \code{encoding_method} is not supported in this step. The automatic threshold selection adapts to each dataset's correlation structure.
#' @param selected (character vector) Predictors retained after filtering. Populated during training and used during baking. Default: NULL
#' @param skip (optional, logical) Trigger to skip this step when baking. Default: FALSE.
#' @param keep_original_cols (optional, logical) Whether to keep original columns. Default: FALSE.
#' @param id (optional, character) Unique identifier for this step.
#'
#' @return Updated recipe with new step.
#'
#' @examples
#' \dontrun{
#' if(requireNamespace("recipes", quietly = TRUE) &&
#'     requireNamespace("parsnip", quietly = TRUE) &&
#'     requireNamespace("workflows", quietly = TRUE)
#'     ){
#'
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' # model formula
#' vi_formula <- collinear::model_formula(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = vi_predictors_numeric
#' )
#'
#' # recipe
#' vi_recipe <- recipes::recipe(
#'   formula = vi_formula,
#'   data = vi_smol
#'   ) |>
#'   #multicollinearity filtering
#'   collinear::step_collinear(
#'     recipes::all_predictors(),
#'     options = list(
#'       max_cor = 0.7,
#'       max_vif = 5,
#'       f = collinear::f_numeric_glm
#'       )
#'   ) |>
#'   #normalization
#'   recipes::step_normalize(
#'     recipes::all_predictors()
#'     )
#'
#' # define linear regression model
#' vi_model <- parsnip::linear_reg() |>
#'   parsnip::set_engine("lm")
#'
#' # create and fit workflow
#' vi_workflow <- workflows::workflow() |>
#'   workflows::add_recipe(vi_recipe) |>
#'   workflows::add_model(vi_model) |>
#'   workflows::fit(data = vi_smol)
#'
#' vi_workflow
#'
#' }
#' }
#' @family multicollinearity_filtering
#' @autoglobal
#' @export
step_collinear <- function(
  recipe,
  ...,
  role = NA,
  trained = FALSE,
  options = list(),
  selected = NULL,
  skip = FALSE,
  keep_original_cols = FALSE,
  id = recipes::rand_id("collinear")
) {
  if (!requireNamespace("recipes", quietly = TRUE)) {
    stop(
      "collinear::step_collinear(): Package 'recipes' is required but not installed.",
      call. = FALSE
    )
  }

  # recipes::add_step(
  #   recipe,
  #   step_collinear_new(
  #     terms = recipes::ellipse_check(...),
  #     role = role,
  #     trained = trained,
  #     options = options,
  #     selected = selected,
  #     skip = skip,
  #     keep_original_cols = keep_original_cols,
  #     id = id
  #   )
  # )

  recipes::add_step(
    recipe,
    step_collinear_new(
      terms = rlang::enquos(...), # ← Changed: store quosures
      role = role,
      trained = trained,
      options = options,
      selected = selected,
      skip = skip,
      keep_original_cols = keep_original_cols,
      id = id
    )
  )
}

step_collinear_new <- function(
  terms,
  role,
  trained,
  options,
  selected,
  skip,
  keep_original_cols,
  id
) {
  structure(
    list(
      terms = terms,
      role = role,
      trained = trained,
      options = options,
      selected = selected,
      skip = skip,
      keep_original_cols = keep_original_cols,
      id = id
    ),
    class = c(
      "step_collinear",
      "recipes_step"
    )
  )
}

#' @rdname step_collinear
#' @param x (required, step_collinear object) The step to be trained. Default: NULL
#' @param training (required, data.frame) The training dataset used to estimate quantities.
#' @param info (optional, data.frame) Preprocessed information about variables in `training`. Default: NULL
#' @param ... (optional) Additional arguments (currently ignored).
#' @importFrom recipes prep
#' @method prep step_collinear
#' @export
prep.step_collinear <- function(
  x = NULL,
  training = NULL,
  info = NULL,
  ...
) {
  if (!requireNamespace("recipes", quietly = TRUE)) {
    stop(
      "collinear::step_collinear(): Package 'recipes' is required but not installed.",
      call. = FALSE
    )
  }

  # outcomes from the recipe
  responses <- recipes::recipes_eval_select(
    quos = list(rlang::quo(recipes::all_outcomes())),
    data = training,
    info = info
  )

  # predictors selected by the step
  predictors <- recipes::recipes_eval_select(
    quos = x$terms,
    data = training,
    info = info
  )

  # remove outcomes if any selector accidentally included them
  predictors <- setdiff(predictors, responses)

  # allow overrides via options
  if (!is.null(x$options$responses)) {
    responses <- x$options$responses
  }

  if (!is.null(x$options$predictors)) {
    predictors <- x$options$predictors
  }

  #keep outcome
  df <- training[, c(responses, predictors), drop = FALSE]

  # df <- training[, predictors, drop = FALSE]

  #inject responses and predictors into options
  x$options$responses <- responses
  x$options$predictors <- predictors
  x$options$encoding_method <- NULL

  # Set quiet = TRUE as default unless user specified otherwise
  if (is.null(x$options$quiet)) {
    x$options$quiet <- TRUE
  }

  result <- do.call(
    what = collinear::collinear,
    args = c(
      list(df = df),
      x$options
    )
  )

  if (length(result) == 0 || is.null(result[[1]])) {
    stop(
      "collinear::step_collinear(): multicollinearity filtering failed,  function collinear::collinear() returned empty results.",
      call. = FALSE
    )
  }

  # In prep.step_collinear, around line 129
  selected = if (is.null(result[[1]]$response)) {
    result[[1]]$selection
  } else {
    c(result[[1]]$response, result[[1]]$selection)
  }

  step_collinear_new(
    terms = x$terms,
    role = x$role,
    trained = TRUE,
    options = x$options,
    selected = selected,
    skip = x$skip,
    keep_original_cols = x$keep_original_cols,
    id = x$id
  )
}

#' @rdname step_collinear
#' @param object (required, \code{step_collinear} object) The trained step. Default: NULL
#' @param new_data (required, data.frame) New data to apply the step to. Default: NULL
#' @param ... (optional) Additional arguments (currently ignored).
#' @importFrom recipes bake
#' @method bake step_collinear
#' @export
bake.step_collinear <- function(
  object = NULL,
  new_data = NULL,
  ...
) {
  if (!requireNamespace("recipes", quietly = TRUE)) {
    stop(
      "collinear::step_collinear(): Package 'recipes' is required but not installed.",
      call. = FALSE
    )
  }

  if (is.null(object$selected)) {
    warning(
      "collinear::bake.step_collinear(): function 'step_collinear()' was not properly trained, returning all variables.",
      call. = FALSE
    )
    return(new_data)
  }

  recipes::check_new_data(
    req = object$selected,
    object = object,
    new_data = new_data
  )

  new_data[, object$selected, drop = FALSE]
}

#' Compute summary statistics for absolute pairwise correlations
#'
#' Computes the the minimum, mean, maximum, and quantiles 0.05, 0.25, median (0.5), 0.75, and 0.95 on the absolute values of the column "correlation" in the output of [cor_df()].
#'
#' @inheritParams cor_matrix
#' @returns dataframe with columns \code{method} (with value "correlation"), \code{statistic} and \code{value}
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' x <- cor_stats(
#'   df = vi_smol,
#'   predictors = vi_predictors_numeric
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_assessment
#' @export
cor_stats <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_stats()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #cor_df dataframe
  if (!"collinear_cor_df" %in% class(df)) {
    df <- cor_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  values <- abs(stats::na.omit(df$correlation))

  stats <- c(
    "n" = length(values),

    "minimum" = min(values),

    "quantile_0.05" = stats::quantile(
      x = values,
      probs = 0.05,
      names = FALSE
    ),

    "quantile_0.25" = stats::quantile(
      x = values,
      probs = 0.25,
      names = FALSE
    ),

    "mean" = mean(values),

    "median" = stats::median(
      x = values
    ),

    "quantile_0.75" = stats::quantile(
      x = values,
      probs = 0.75,
      names = FALSE
    ),

    "quantile_0.95" = stats::quantile(
      x = values,
      probs = 0.95,
      names = FALSE
    ),

    "maximum" = max(values)
  ) |>
    round(digits = 4)

  stats_names <- names(stats)
  names(stats) <- NULL

  out <- data.frame(
    method = "correlation",
    statistic = stats_names,
    value = stats
  )

  out
}

#' VIF Statistics
#'
#' Computes the the minimum, mean, maximum, and quantiles 0.05, 0.25, median (0.5), 0.75, and 0.95 of the column "vif" in the output of [vif_df()].
#'
#' @inheritParams collinear
#' @returns dataframe with columns \code{method} with the value "vif", \code{statistic} with the statistic name, and \code{value}.
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' # ## OPTIONAL: parallelization setup
#' # ## irrelevant when all predictors are numeric
#' # ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' # ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' x <- vif_stats(
#'   df = vi_smol,
#'   predictors = vi_predictors_numeric
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_assessment
#' @export
vif_stats <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::vif_stats()",
    function_name = dots$function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  #cor_df dataframe
  if (!"collinear_vif_df" %in% class(df)) {
    df <- vif_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name,
      m = dots$m
    )
  }

  values <- stats::na.omit(df$vif)

  stats <- c(
    "n" = length(values),

    "minimum" = min(values),

    "quantile_0.05" = stats::quantile(
      x = values,
      probs = 0.05,
      names = FALSE
    ),

    "quantile_0.25" = stats::quantile(
      x = values,
      probs = 0.25,
      names = FALSE
    ),

    "mean" = mean(values),

    "median" = stats::median(
      x = values
    ),

    "quantile_0.75" = stats::quantile(
      x = values,
      probs = 0.75,
      names = FALSE
    ),

    "quantile_0.95" = stats::quantile(
      x = values,
      probs = 0.95,
      names = FALSE
    ),

    "maximum" = max(values)
  ) |>
    round(digits = 4)

  stats_names <- names(stats)
  names(stats) <- NULL

  out <- data.frame(
    method = "vif",
    statistic = stats_names,
    value = stats
  )

  out
}

#' Ensure that argument `df` is not `NULL`
#'
#' @description
#' Internal function to validate the default value of the argument \code{df}.
#' @inheritParams collinear
#' @inheritParams validate_arg_quiet
#' @return dataframe
#' @autoglobal
#' @family argument_validation
#' @export
#' @examples
#' data(vi_smol, package = "spatialData")
#' df <- validate_arg_df_not_null(
#'   df = vi_smol
#'   )
validate_arg_df_not_null <- function(
  df = NULL,
  function_name = NULL
) {
  if (isTRUE(attr(x = df, which = "validated"))) {
    return(df)
  }

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_df_not_null()",
    function_name = function_name
  )

  if (is.null(df)) {
    stop(
      "\n",
      function_name,
      ": argument 'df' cannot be NULL.",
      call. = FALSE
    )
  }

  df
}

testthat::test_that("`validate_arg_df_not_null()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  testthat::expect_error(
    x <- validate_arg_df_not_null(
      df = NULL
    ),
    regexp = "argument 'df' cannot be NULL"
  )

  x <- validate_arg_df_not_null(
    df = vi_smol
  )

  testthat::expect_true(
    all.equal(vi_smol, x)
  )

  df <- vi_smol

  attr(x = df, which = "validated") <- TRUE

  x <- validate_arg_df_not_null(
    df = df
  )
})

#' Print all collinear selection results of `collinear()`
#'
#' @param x (required, list of class \code{class.collinear_output}) Object to print. Default: NULL
#' @param n (optional, integer) Maximum printed vector length. Default: 5.
#' @param ... Ignored, kept for consistency with generic.
#' @method print collinear_output
#' @family S3_methods
#' @autoglobal
#' @export
print.collinear_output <- function(
  x = NULL,
  n = 5,
  ...
) {
  if (length(names(x)) > 1) {
    cat("Results\n")
    cat("===================\n")
    cat("\n")
  } else {
    cat("Result\n")
    cat("===================\n")
  }

  lapply(
    X = x,
    FUN = print,
    n = n
  )

  invisible()
}

#' Check and validate argument \code{encoding_method}
#'
#' @description
#' Internal function to validate the argument \code{encoding_method} of [target_encoding_lab()].
#'
#'
#' @inheritParams target_encoding_lab
#' @inheritParams validate_arg_quiet
#'
#' @returns character
#' @family argument_validation
#' @examples
#' x <- validate_arg_encoding_method(
#'   encoding_method = "wrong_method"
#'   )
#' @export
#' @autoglobal
validate_arg_encoding_method <- function(
  encoding_method = "loo",
  overwrite = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  valid_methods <- c(
    "loo",
    "mean",
    "rank"
  )

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_encoding_method()",
    function_name = function_name
  )

  if (is.null(encoding_method)) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'encoding_method' is NULL, skipping target encoding."
      )
    }

    return(NULL)
  }

  if (isTRUE(attr(x = encoding_method, which = "validated"))) {
    return(encoding_method)
  }

  encoding_method <- intersect(
    x = encoding_method,
    y = valid_methods
  )

  if (
    is.null(encoding_method) ||
      length(encoding_method) == 0
  ) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'encoding_method' is not valid, resetting it to '",
        valid_methods[1],
        "'."
      )
    }

    encoding_method <- valid_methods[1]
  }

  if (
    is.logical(overwrite) && overwrite == TRUE && length(encoding_method) > 1
  ) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": only one encoding method allowed when 'overwrite = TRUE', using method: '",
        valid_methods[1],
        "'."
      )
    }

    encoding_method <- valid_methods[1]
  }

  attr(
    x = encoding_method,
    which = "validated"
  ) <- TRUE

  encoding_method
}

testthat::test_that("`validate_arg_encoding_method()` works", {
  testthat::skip_on_cran()

  testthat::expect_message(
    x <- validate_arg_encoding_method(
      encoding_method = "hola"
    ),
    regexp = "argument 'encoding_method' is not valid, resetting it to 'loo'"
  )

  testthat::expect_true(
    x == "loo"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  testthat::expect_no_message(
    x <- validate_arg_encoding_method(
      encoding_method = "hola",
      quiet = TRUE
    )
  )

  testthat::expect_true(
    x == "loo"
  )

  testthat::expect_true(
    attributes(x)$validated
  )

  #passing a validated input
  x <- validate_arg_encoding_method(
    encoding_method = x,
    quiet = TRUE
  )

  #passing a NULL
  testthat::expect_message(
    x <- validate_arg_encoding_method(
      encoding_method = NULL
    ),
    regexp = "argument 'encoding_method' is NULL, skipping target encoding"
  )

  #with overwrite
  testthat::expect_message(
    x <- validate_arg_encoding_method(
      encoding_method = c("loo", "mean"),
      overwrite = TRUE
    ),
    regexp = "only one encoding method allowed when 'overwrite = TRUE', using method: 'loo'"
  )
})

#' Summarize all results of `collinear()`
#'
#' @param object (required, list of class \code{collinear_output}) Object to summarize. Default: NULL
#' @param ... Ignored, kept for consistency with generic.
#' @return list:
#' If \code{object} was created with \code{responses = NULL}, a sublist named "result" containing a vector with the selected predictors. Otherwise, a list named after each response containing the corresponding variable selection.
#' @method summary collinear_output
#' @family S3_methods
#' @autoglobal
#' @export
summary.collinear_output <- function(
  object = NULL,
  ...
) {
  #print summary of selections
  lapply(
    X = object,
    FUN = summary
  )
}

#' Check and prepare argument `df`
#'
#' @description
#' Internal function to validate the integrity of the argument \code{df}. It ensures that the dataframe has suitable dimensions for a multicollinearity analysis, transforms logical columns to numeric, character columns to factors, and converts \code{NaN}, \code{Inf} and \code{-Inf} to NA. Additionally, it checks the values of \code{responses} and \code{predictors} if these arguments are provided.
#'
#' @inheritParams collinear
#' @inheritParams f_auto
#' @inheritParams validate_arg_quiet
#' @return dataframe
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' df <- validate_arg_df(
#'   df = vi_smol,
#'   responses = "vi_numeric",
#'   predictors = vi_predictors_numeric,
#'   quiet = FALSE
#' )
#'
#' attributes(vi)$validated
#' @autoglobal
#' @family argument_validation
#' @export
validate_arg_df <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  quiet = FALSE,
  function_name = NULL
) {
  if (isTRUE(attr(x = df, which = "validated"))) {
    return(df)
  }

  function_name <- validate_arg_function_name(
    default_name = "collinear::validate_arg_df()",
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  df <- drop_geometry_column(
    df = df,
    quiet = quiet,
    function_name = function_name
  )

  #handle coercion to df
  if (is.data.frame(df) == FALSE) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'df' is not a dataframe, attempting coercion."
      )
    }

    df <- tryCatch(
      {
        out <- as.data.frame(df)

        if (all(dim(out)) == 1) {
          stop()
        }
      },
      error = function(e) {
        stop(
          "\n",
          function_name,
          ": cannot coerce argument 'df' to class 'data.frame'.",
          call. = FALSE
        )
      }
    )
  }

  #stop if no columns
  if (ncol(df) == 0) {
    stop(
      "\n",
      function_name,
      ": argument 'df' has zero columns.",
      call. = FALSE
    )
  }

  if (ncol(df) == 1) {
    stop(
      "\n",
      function_name,
      ": argument 'df' has one valid column, multicollinearity analysis cannot be performed.",
      call. = FALSE
    )
  }

  #error if not enough rows
  if (nrow(df) < 3) {
    stop(
      function_name,
      ": argument 'df' has fewer than 3 rows, multicollinearity analysis cannot be performed.",
      call. = FALSE
    )

    #warning if not enough rows
  } else if (nrow(df) < 10) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'df' has fewer than 10 rows, multicollinearity analysis may be unreliable due to insufficient sample size."
      )
    }

    #message if not enough rows
  } else if (nrow(df) < 30) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": argument 'df' has fewer than 30 rows, results of multicollinearity filtering may be statistically fragile."
      )
    }
  }

  responses <- validate_arg_responses(
    df = df,
    responses = responses,
    quiet = quiet,
    function_name = function_name
  )

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      responses = responses,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (all(is.null(c(responses, predictors)))) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": arguments 'responses' and 'predictors' are NULL, skipping validation of column values."
      )
    }

    return(df)
  }

  #subset valid columns
  selected_columns <- unique(c(responses, predictors))

  #general data validation
  df <- df[, selected_columns, drop = FALSE]

  column_order <- selected_columns

  #identify predictors types
  column_types <- identify_valid_variables(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  if (
    length(
      c(
        column_types$numeric,
        column_types$categorical
      )
    ) ==
      0
  ) {
    stop(
      "\n",
      function_name,
      ": argument 'df' has no valid columns.",
      call. = FALSE
    )
  }

  #logicals to numeric
  if (length(column_types$logical) > 0) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": converted the following logical columns to numeric:\n -",
        paste0(column_types$logical, collapse = "\n - ")
      )
    }

    #convert logical to numeric
    df <- rapply(
      object = df,
      f = as.numeric,
      classes = c("logical"),
      how = "replace"
    )

    #update identification of predictors
    column_types <- identify_valid_variables(
      df = df,
      predictors = selected_columns,
      function_name = function_name
    )

    selected_columns <- c(
      column_types$numeric,
      column_types$categorical
    )
  }

  #transform character categoricals to factors
  if (length(column_types$categorical) > 0) {
    #identify character categoricals
    char_cols <- column_types$categorical[
      vapply(
        X = df[, column_types$categorical, drop = FALSE],
        FUN = is.character,
        logical(1)
      )
    ]

    if (length(char_cols) > 0) {
      if (quiet == FALSE) {
        message(
          "\n",
          function_name,
          ": converted the following character columns to factor:\n - ",
          paste0(char_cols, collapse = "\n - ")
        )
      }

      #convert to factor
      df[char_cols] <- lapply(
        X = df[char_cols],
        FUN = as.factor
      )
    }
  }

  #replace invalid numeric values
  if (length(column_types$numeric) > 0) {
    df_numeric <- df[, column_types$numeric, drop = FALSE]

    # replace inf with NA ----
    n_inf <- sum(
      vapply(
        X = df_numeric,
        FUN = function(x) sum(!is.finite(x)),
        FUN.VALUE = integer(1)
      )
    )

    if (n_inf > 0) {
      if (quiet == FALSE) {
        #identify involved columns
        columns_inf <- colnames(df_numeric)[
          vapply(
            X = df_numeric,
            FUN = function(x) any(!is.finite(x)),
            FUN.VALUE = logical(1)
          )
        ]

        message(
          "\n",
          function_name,
          ": replaced ",
          n_inf,
          " Inf, -Inf, or NaN values with NA in these columns: \n - ",
          paste0(columns_inf, collapse = "\n - ")
        )
      }

      # replace Inf, -Inf and NaN with NA
      is.na(df_numeric) <- vapply(
        X = df_numeric,
        FUN = function(x) !is.finite(x),
        FUN.VALUE = logical(nrow(df_numeric))
      )
    }

    #recover non-numeric columns
    df <- cbind(
      df_numeric,
      df[, c(column_types$categorical), drop = FALSE]
    )

    #reorder columns
    column_order <- intersect(
      x = column_order,
      y = colnames(df)
    )

    df <- df[, column_order, drop = FALSE]
  }

  attr(
    x = df,
    which = "validated"
  ) <- TRUE

  df
}

#' @title Smart multicollinearity management
#'
#' @description
#'
#' Automates multicollinearity management in datasets with mixed variable types
#' (numeric, categorical, and logical) through an integrated system of five key
#' components:
#'
#' \describe{
#'   \item{Target Encoding Integration (opt-in)}{
#'     When \code{responses} is numeric, categorical predictors can be converted
#'     to numeric using response values as reference. This enables VIF and
#'     correlation analysis across mixed types. See \code{\link{target_encoding_lab}}.
#'   }
#'
#'   \item{Intelligent Predictor Ranking (active by default)}{
#'     Three prioritization strategies ensure the most relevant predictors are
#'     retained during filtering:
#'     \itemize{
#'       \item \strong{User-defined ranking} (argument \code{preference_order}):
#'         Accepts a character vector of predictor names or a dataframe from
#'         \code{\link{preference_order}}. Lower-ranked collinear predictors are removed.
#'       \item \strong{Response-based ranking} (\code{f}):
#'         Uses \code{\link{f_auto}}, \code{\link{f_numeric_glm}}, or
#'         \code{\link{f_binomial_rf}} to rank predictors by association with
#'         the response. Supports cross-validation via \code{\link{preference_order}}.
#'       \item \strong{Multicollinearity-based ranking} (default):
#'         When both \code{preference_order} and \code{f} are \code{NULL},
#'         predictors are ranked from lower to higher multicollinearity.
#'     }
#'   }
#'
#'   \item{Unified Correlation Framework (active by default)}{
#'     Computes pairwise correlations between variable types using Pearson
#'     (numeric–numeric), target encoding (numeric–categorical), and Cramer's V
#'     (categorical–categorical). See \code{\link{cor_df}}, \code{\link{cor_matrix}},
#'     and \code{\link{cor_cramer}}.
#'   }
#'
#'   \item{Adaptive Filtering Thresholds (active by default)}{
#'     When \code{max_cor} and \code{max_vif} are both \code{NULL}, thresholds
#'     are determined from the median correlation structure of the predictors.
#'   }
#'
#'   \item{Dual Filtering Strategy (active by default)}{
#'     Combines two complementary methods while respecting predictor rankings:
#'     \itemize{
#'       \item \strong{Pairwise Correlation Filtering}:
#'         Removes predictors with Pearson correlation or Cramer's V above
#'         \code{max_cor}. See \code{\link{cor_select}}.
#'       \item \strong{VIF-based Filtering}:
#'         Removes numeric predictors with VIF above \code{max_vif}. See
#'         \code{\link{vif_select}}, \code{\link{vif_df}}, and \code{\link{vif}}.
#'     }
#'   }
#' }
#'
#' This function accepts parallelization via \code{future::plan()} and progress
#' bars via \code{progressr::handlers()}. Parallelization benefits
#' \code{\link{target_encoding_lab}}, \code{\link{preference_order}}, and
#' \code{\link{cor_select}}.
#'
#' @section Adaptive Multicollinearity Thresholds:
#'
#' When both \code{max_cor} and \code{max_vif} are \code{NULL}, the function
#' determines thresholds as follows:
#' \enumerate{
#'   \item Compute the 75th percentile of pairwise correlations via
#'     \code{\link{cor_stats}}.
#'   \item Map that value through a sigmoid between 0.545 (VIF~2.5) and 0.785
#'     (VIF~7.5), centered at 0.665, to get \code{max_cor}.
#'   \item Compute \code{max_vif} from \code{max_cor} using
#'     \code{\link{gam_cor_to_vif}}.
#' }
#'
#' @section Variance Inflation Factors:
#'
#' VIF for predictor \eqn{a} is computed as \eqn{1/(1-R^2)}, where \eqn{R^2} is
#' the multiple R-squared from regressing \eqn{a} on the other predictors.
#' Recommended maximums commonly used are 2.5, 5, and 10.
#'
#' @section VIF-based Filtering:
#'
#' \code{\link{vif_select}} ranks numeric predictors (user \code{preference_order}
#' if provided, otherwise from lower to higher VIF) and sequentially adds
#' predictors whose VIF against the current selection is below \code{max_vif}.
#'
#' @section Pairwise Correlation Filtering:
#'
#' \code{\link{cor_select}} computes the global correlation matrix, orders
#' predictors by \code{preference_order} or by lower-to-higher summed
#' correlations, and sequentially selects predictors with pairwise correlations
#' below \code{max_cor}.
#'
#' @param df (required; dataframe, tibble, or sf) A dataframe with responses
#'   (optional) and predictors. Must have at least 10 rows for pairwise
#'   correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
#'   Default: NULL.
#'
#' @param responses (optional; character, character vector, or NULL) Name of
#'   one or several response variables in \code{df}. Default: NULL.
#'
#' @param predictors (optional; character vector or NULL) Names of the
#'   predictors in \code{df}. If NULL, all columns except \code{responses} and
#'   constant/near-zero-variance columns are used. Default: NULL.
#'
#' @param encoding_method (optional; character or NULL) One of "loo", "mean",
#'   or "rank". If NULL, target encoding is disabled. Default: NULL.
#'
#' @param preference_order (optional; character vector, dataframe from
#'   \code{\link{preference_order}}, or NULL) Prioritizes predictors to preserve.
#'
#' @param f (optional; unquoted function name or NULL) Function to rank
#'   predictors by relationship with \code{responses}. See \code{\link{f_functions}}.
#'   Default: \code{f_auto}.
#'
#' @param max_cor (optional; numeric or NULL) Maximum allowed pairwise
#'   correlation (0.01–0.99). Recommended between 0.5 and 0.9. If NULL and
#'   \code{max_vif} is NULL, it is selected automatically. Default: NULL.
#'
#' @param max_vif (optional; numeric or NULL) Maximum allowed VIF. Recommended
#'   between 2.5 and 10. If NULL and \code{max_cor} is NULL, configured
#'   automatically. Default: NULL.
#'
#' @param quiet (optional; logical) If FALSE, messages are printed. Default: FALSE.
#'
#' @param ... (optional) Internal args (e.g. \code{function_name} for
#'   \code{\link{validate_arg_function_name}}, a precomputed correlation matrix
#'   \code{m}, or cross-validation args for \code{\link{preference_order}}).
#'
#' @return A list of class \code{collinear_output} with sublists of class
#'   \code{collinear_selection}. If \code{responses = NULL} a single sublist
#'   named "result" is returned; otherwise a sublist per response is returned.
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#' x <- collinear(df = vi_smol[, vi_predictors_numeric])
#'
#' @autoglobal
#' @references
#' \itemize{
#'  \item David A. Belsley, D.A., Kuh, E., Welsch, R.E. (1980). Regression
#'    Diagnostics: Identifying Influential Data and Sources of Collinearity.
#'    John Wiley & Sons. DOI: 10.1002/0471725153.
#'  \item Micci-Barreca, D. (2001) A Preprocessing Scheme for High-Cardinality
#'    Categorical Attributes in Classification and Prediction Problems. SIGKDD
#'    Explor. Newsl. 3, 1, 27-32. DOI: 10.1145/507533.507538
#' }
#' @family multicollinearity_filtering
#' @export
collinear <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  encoding_method = NULL,
  preference_order = NULL,
  f = f_auto,
  max_cor = NULL,
  max_vif = NULL,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::collinear()",
    function_name = dots$function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  quiet <- validate_arg_quiet(
    function_name = function_name,
    quiet = quiet
  )

  responses <- validate_arg_responses(
    df = df,
    responses = responses,
    quiet = quiet,
    function_name = function_name
  )

  predictors <- validate_arg_predictors(
    df = df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df.ncol <- ncol(df)

  df <- validate_arg_df(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  #revalidate predictors if columns were removed
  if (ncol(df) < df.ncol) {
    attributes(responses)$validated <- NULL
    attributes(predictors)$validated <- NULL

    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )

    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #identify variable types
  var.types <- identify_valid_variables(
    df = df,
    predictors = c(responses, predictors),
    quiet = quiet,
    function_name = function_name
  )

  #check if target encoding will be triggered
  target_encoding_needed <- FALSE
  if (
    !is.null(encoding_method) &&
      !is.null(responses) &&
      any(responses %in% var.types$numeric) &&
      any(predictors %in% var.types$categorical)
  ) {
    target_encoding_needed <- TRUE

    if (length(responses) == 1) {
      df <- target_encoding_lab(
        df = df,
        response = responses,
        predictors = predictors,
        encoding_method = encoding_method,
        overwrite = TRUE,
        quiet = quiet,
        function_name = function_name
      )

      target_encoding_needed <- FALSE
    }
  }

  #setup max cor and max vif if NULL
  m <- dots$m
  if (
    is.null(max_cor) &&
      is.null(max_vif)
  ) {
    #computing cor df and m here to avoid recomputation in collinear()
    cor.df <- cor_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )

    m <- cor_matrix(
      df = cor.df,
      quiet = quiet,
      function_name = function_name
    )

    collinearity.stats <- collinear_stats(
      df = cor.df,
      predictors = predictors,
      quiet = TRUE,
      function_name = function_name
    )

    #check if filtering is needed at all
    vif.max <- collinearity.stats[
      collinearity.stats$method == "vif" &
        collinearity.stats$statistic == "maximum",
      "value"
    ]

    cor.max <- collinearity.stats[
      collinearity.stats$method == "correlation" &
        collinearity.stats$statistic == "maximum",
      "value"
    ]

    #data needs no filtering
    if (vif.max <= 2.5) {
      max_vif <- 2.5000
      max_cor <- round(x = cor.max, digits = 4)
    } else {
      #adaptive max_cor threshold using sigmoid soft clamping
      #smooth transition between conservative and permissive filtering

      #correlation structure of the input data
      cor.statistic <- collinearity.stats[
        collinearity.stats$method == "correlation" &
          collinearity.stats$statistic == "quantile_0.75",
        "value"
      ]

      #max_cor = 0.545 corresponds to VIF ~ 2.5 (conservative)
      prediction_df <- collinear::prediction_cor_to_vif

      sigmoid_floor <- prediction_df[
        which.min(abs(prediction_df$max_vif - 2.5)),
        "max_cor"
      ]

      #max_cor = 0.785 corresponds to VIF ~ 7.5 (permissive)
      sigmoid_ceiling <- prediction_df[
        which.min(abs(prediction_df$max_vif - 7.5)),
        "max_cor"
      ]

      #sigmoid params
      sigmoid_midpoint <- (sigmoid_floor + sigmoid_ceiling) / 2
      sigmoid_range <- sigmoid_ceiling - sigmoid_floor

      #smooth sigmoid transition between floor and ceiling
      #parameter (-15) controls transition sharpness
      max_cor <- sigmoid_floor +
        sigmoid_range /
          (1 + exp(-15 * (cor.statistic - sigmoid_midpoint)))

      max_cor <- round(x = max_cor, digits = 4)

      max_vif <- mgcv::predict.gam(
        object = collinear::gam_cor_to_vif,
        newdata = data.frame(
          max_cor = max_cor
        )
      ) |>
        round(digits = 4)
    }

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": setting 'max_cor' to ",
        max_cor,
        "."
      )

      message(
        "\n",
        function_name,
        ": setting 'max_vif' to ",
        max_vif,
        "."
      )
    }
  }

  #compute correlation matrix if needed
  if (
    target_encoding_needed == FALSE &&
      is.null(m)
  ) {
    m <- cor_matrix(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  #manage response for loop
  if (
    length(responses) == 0 ||
      is.null(responses)
  ) {
    responses <- list(NULL)
  }

  #if several responses, invalidate predictors
  if (
    length(responses) > 1 &&
      any(responses %in% predictors)
  ) {
    attributes(predictors)$validated <- NULL
  }

  #parse f's name
  f <- validate_arg_f(
    f = f,
    f_name = deparse(substitute(f)),
    function_name = function_name
  )

  #ITERATION ----
  out <- list()

  ## start ----
  for (response in responses) {
    if (
      quiet == FALSE &&
        length(responses) > 1
    ) {
      msg <- paste0(
        function_name,
        ": processing response '",
        response,
        "'"
      )

      message("\n", msg)
      message(rep(x = "-", times = nchar(msg)))
    }

    ## copy df ----
    df.response <- df

    ## validate predictors ----
    predictors.response <- validate_arg_predictors(
      df = df.response,
      responses = response,
      predictors = predictors,
      function_name = function_name
    )

    ## subset correlation matrix to the given predictors
    if (
      !is.null(m) &&
        length(predictors.response) < length(colnames(m))
    ) {
      m <- m[
        predictors.response,
        predictors.response
      ]

      class(m) <- unique(c("collinear_cor_matrix", class(m)))
    }

    ## TARGET ENCODING ----
    if (
      target_encoding_needed &&
        response %in% var.types$numeric &&
        any(predictors.response %in% var.types$categorical)
    ) {
      df.response <- target_encoding_lab(
        df = df.response,
        response = response,
        predictors = predictors.response,
        encoding_method = encoding_method,
        overwrite = TRUE,
        quiet = quiet,
        function_name = function_name
      )
    }

    ## PREFERENCE ORDER ----
    preference_order.response <- preference_order

    if (
      is.null(preference_order) &&
        !is.null(response) &&
        !is.null(f)
    ) {
      if (is.null(dots$cv_training_fraction)) {
        dots$cv_training_fraction <- 1
      }

      if (is.null(dots$cv_iterations)) {
        dots$cv_iterations <- 1
      }

      preference_order.response <- preference_order(
        df = df.response,
        responses = response,
        predictors = predictors.response,
        f = f,
        cv_training_fraction = dots$cv_training_fraction,
        cv_iterations = dots$cv_iterations,
        quiet = quiet,
        function_name = function_name,
        m = m
      )
    }

    #validated here so we can put it in the output list
    preference_order.response <- validate_arg_preference_order(
      df = df.response,
      response = response,
      predictors = predictors.response,
      preference_order = preference_order.response,
      quiet = quiet,
      function_name = function_name,
      m = m
    )

    ##MULTICOLLINEARITY ANALYSIS ----
    selection.response <- collinear_select(
      df = df.response,
      response = response,
      predictors = predictors.response,
      preference_order = preference_order.response,
      max_cor = max_cor,
      max_vif = max_vif,
      quiet = quiet,
      function_name = function_name,
      m = m
    )

    ##PREPARE OUTPUT ----

    df.response <- df.response[,
      c(response, selection.response),
      drop = FALSE
    ]

    attr(
      x = df.response,
      which = "validated"
    ) <- TRUE

    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": selected predictors: \n - ",
        paste(selection.response, collapse = "\n - ")
      )
    }

    #prepare output list
    out.response <- list()

    if (!is.null(response)) {
      out.response$response <- response
    }

    out.response$df <- df.response
    out.response$preference_order <- preference_order.response
    out.response$selection <- selection.response

    if (
      all(
        c(
          !is.null(response),
          !is.null(selection.response)
        )
      )
    ) {
      response_type <- identify_valid_variables(
        df = df.response,
        predictors = response,
        function_name = function_name
      ) |>
        unlist() |>
        names()

      selection_type <- identify_valid_variables(
        df = df.response,
        predictors = selection.response,
        function_name = function_name
      )

      out.response$formulas <- list()

      general_formula <- model_formula(
        df = df.response,
        response = response,
        predictors = selection.response,
        quiet = quiet,
        function_name = function_name
      )

      general_formula_name <- c(
        numeric = "linear",
        logical = "binomial",
        categorical = "classification"
      )[[response_type]]

      out.response$formulas[[general_formula_name]] <- general_formula

      #add smooth formula if relevant
      if (
        response_type == "numeric" &&
          length(selection_type$numeric) > 0
      ) {
        out.response$formulas[["smooth"]] <- model_formula(
          df = df.response,
          response = response,
          predictors = selection.response,
          term_f = "s",
          quiet = quiet,
          function_name = function_name
        )
      }
    }

    class(out.response) <- c(
      "collinear_selection",
      class(out.response)
    )

    slot_name <- response
    if (is.null(response)) {
      slot_name <- "result"
    }

    out[[slot_name]] <- out.response
  } #end of loop

  class(out) <- c("collinear_output", class(out))

  out
}

#' Compute variance inflation factors from a correlation matrix
#'
#' @description Computes the Variance Inflation Factors from a correlation matrix in two steps:
#' \itemize{
#'   \item Applies \code{base::solve()} to transform the correlation matrix into a precision matrix, which is the inverse of the covariance matrix between all variables in \code{predictors}.
#'   \item Applies \code{base::diag()} to extract the diagonal of the precision matrix, which contains the variance of the regression of each predictor against all other predictors, also known as Variance Inflation Factor
#'}
#'
#' @inheritSection collinear Variance Inflation Factors
#'
#' @param m (required, matrix) Correlation matrix generated via \code{stats::cor()} or [cor_matrix()]. Must have named dimensions. Default: NULL
#' @inheritParams collinear
#'
#' @return named numeric vector
#' @family multicollinearity_assessment
#' @inherit vif_select references
#' @autoglobal
#' @export
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' m <- cor_matrix(
#'   df = vi_smol,
#'   predictors = vi_predictors_numeric[1:5]
#' )
#'
#' vif(m)
vif <- function(
  m = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::vif()",
    ... = ...
  )

  if (!"matrix" %in% class(m)) {
    stop(
      "\n",
      function_name,
      ": argument 'm' must be a correlation matrix generated with 'stats::cor()' or 'collinear::cor_matrix()'.",
      call. = FALSE
    )
  }

  if (length(unique(dim(m))) > 1) {
    stop(
      "\n",
      function_name,
      ": argument 'm' must be a square matrix.",
      call. = FALSE
    )
  }

  if (length(dimnames(m)[[1]]) != nrow(m)) {
    stop(
      "\n",
      function_name,
      ": argument 'm' must have dimension names.",
      call. = FALSE
    )
  }

  #condition number
  # kappa_m <- kappa(m, exact = TRUE)
  # if(kappa_m > 1e10 && quiet == FALSE){
  #   message(
  #     "\n",
  #     function_name,
  #     ": VIF values may be numerically unstable due to severe multicollinearity."
  #   )
  # }

  #tol arg of solve()
  #prevents both false singularity detection and numerical instability
  tol <- min(
    .Machine$double.eps * nrow(m) * max(abs(m)),
    sqrt(.Machine$double.eps)
  )

  out <- tryCatch(
    #first attempt with original matrix
    {
      diag(solve(m, tol = tol))
    },

    #try with adjusted matrix
    error = function(e) {
      #compute minimum eigenvalue
      min_eigen <- min(eigen(m, only.values = TRUE)$values)

      #apply ridge regularization if needed
      if (min_eigen < 0.001) {
        #shift eigenvalues to ensure positive definiteness
        ridge <- abs(min_eigen) + 0.001
        m_adj <- m + diag(ridge, nrow(m))
      } else {
        m_adj <- m
      }

      #try again with adjusted matrix
      tryCatch(
        {
          diag(solve(m_adj, tol = tol))
        },
        error = function(e2) {
          warning(
            "\n",
            function_name,
            ": the correlation matrix is singular and cannot be solved, returning infinite VIF scores.",
            call. = FALSE
          )
          out <- rep(Inf, times = ncol(m))
          names(out) <- colnames(m)
          return(out)
        }
      )
    }
  )

  # cap
  vif_cap <- 1e6
  if (any(out > vif_cap, na.rm = TRUE)) {
    if (quiet == FALSE) {
      message(
        "\n",
        function_name,
        ": some VIF values exceeded 1M and were set to Inf."
      )
    }
    out[out > vif_cap] <- Inf
  }

  #add names
  names(out) <- colnames(m)

  #sort
  out <- sort(
    x = out,
    decreasing = TRUE
  ) |>
    round(digits = 4)

  out
}

#' @rdname target_encoding_methods
#' @autoglobal
#' @export
target_encoding_rank <- function(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = NULL,
  ...
) {
  if (is.null(encoded_name)) {
    encoded_name <- paste0(
      predictor,
      "__encoded"
    )
  }

  #aggregate by groups
  df.map <- tapply(
    X = df[[response]],
    INDEX = df[[predictor]],
    FUN = mean,
    na.rm = TRUE
  ) |>
    sort()

  #add rank column
  df.map <- data.frame(
    names(df.map),
    seq_along(df.map) #rank column
  )
  names(df.map) <- c(predictor, encoded_name)

  #column order for merged df
  df.cols <- unique(
    c(
      colnames(df),
      colnames(df.map)
    )
  )

  #merge
  df <- merge(
    x = df,
    y = df.map,
    by = predictor,
    sort = FALSE
  )

  #reorder columns
  df <- df[, df.cols]

  df
}

testthat::test_that("`target_encoding_rank()` works", {
  testthat::skip_on_cran()

  data(vi_smol, package = "spatialData")

  #rank
  df <- target_encoding_rank(
    df = vi_smol,
    response = "vi_numeric",
    predictor = "soil_type",
    encoded_name = "encoded_column"
  )

  testthat::expect_true(
    class(df[["encoded_column"]]) == "integer"
  )

  df <- target_encoding_rank(
    df = vi_smol,
    response = "vi_numeric",
    predictor = "soil_type",
    encoded_name = NULL
  )

  testthat::expect_true(
    "soil_type__encoded" %in% colnames(df)
  )
})

#' @title Multicollinearity filtering by pairwise correlation threshold
#'
#' @description
#'
#' Wraps [collinear_select()] to automatize multicollinearity filtering via pairwise correlation in dataframes with numeric and categorical predictors.
#'
#' The argument \code{max_cor} determines the maximum variance inflation factor allowed in the resulting selection of predictors.
#'
#' The argument \code{preference_order} accepts a character vector of predictor names ranked from first to last index, or a dataframe resulting from [preference_order()]. When two predictors in this vector or dataframe are highly collinear, the one with a lower ranking is removed. This option helps protect predictors of interest. If not provided, predictors are ranked from lower to higher multicollinearity.
#'
#' Please check the section **Pairwise Correlation Filtering** at the end of this help file for further details.
#'
#' @inheritSection collinear Pairwise Correlation Filtering
#'
#' @inheritParams collinear_select
#' @return character vector of selected predictors
#' @examples
#' data(vi_smol, package = "spatialData")
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' #predictors
#' predictors = c(
#'   "koppen_zone", #character
#'   "soil_type", #factor
#'   "topo_elevation", #numeric
#'   "soil_temperature_mean" #numeric
#' )
#'
#' #predictors ordered from lower to higher multicollinearity
#' x <- cor_select(
#'   df = vi_smol,
#'   predictors = predictors,
#'   max_cor = 0.7
#' )
#'
#' x
#'
#'
#' #with custom preference order
#' x <- cor_select(
#'   df = vi_smol,
#'   predictors = predictors,
#'   preference_order = c(
#'     "koppen_zone",
#'     "soil_type"
#'   ),
#'   max_cor = 0.7
#' )
#'
#' x
#'
#' #with automated preference order
#' df_preference <- preference_order(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = predictors
#' )
#'
#' df_preference
#'
#' x <- cor_select(
#'   df = vi_smol,
#'   predictors = predictors,
#'   preference_order = df_preference,
#'   max_cor = 0.7
#' )
#'
#' x
#'
#' #OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_filtering
#' @author Blas M. Benito, PhD
#' @export
cor_select <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  preference_order = NULL,
  max_cor = 0.7,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::cor_select()",
    function_name = dots$function_name
  )

  if (is.null(max_cor)) {
    stop(
      "\n",
      function_name,
      ": argument 'max_cor' cannot be NULL.",
      call. = FALSE
    )
  }

  out <- collinear_select(
    df = df,
    response = response,
    predictors = predictors,
    preference_order = preference_order,
    max_cor = max_cor,
    max_vif = NULL,
    quiet = quiet,
    function_name = function_name,
    m = dots$m
  )

  out
}

#' Find valid numeric, categorical, and logical variables in a dataframe
#'
#' @description
#' Returns a list with the names of the valid numeric, categorical, and logical variables in a modelling dataframe.
#'
#' @inheritParams collinear
#' @inheritParams identify_zero_variance_variables
#' @param predictors (required, character vector) Names of the predictors to identify. Default: NULL
#' @return list
#' \itemize{
#'   \item \code{numeric}: character vector of numeric predictors.
#'   \item \code{categorical}: character vector of categorical (character and factor) predictors.
#'   \item \code{logical}: character vector of logical predictors.
#' }
#' @examples
#'
#' data(vi_smol, vi_predictors, package = "spatialData")
#'
#' x <- identify_valid_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )
#'
#' x
#'
#' @autoglobal
#' @family data_types
#' @author Blas M. Benito, PhD
#' @export
identify_valid_variables <- function(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  decimals = 4,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::identify_valid_variables()",
    ... = ...
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (!is.null(responses)) {
    responses <- validate_arg_responses(
      df = df,
      responses = responses,
      quiet = quiet,
      function_name = function_name
    )
  }

  if (!is.null(predictors)) {
    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  }

  variables_numeric <- identify_numeric_variables(
    df = df,
    responses = responses,
    predictors = predictors,
    decimals = decimals,
    quiet = quiet,
    function_name = function_name
  )$valid

  variables_categorical <- identify_categorical_variables(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )$valid

  variables_logical <- identify_logical_variables(
    df = df,
    responses = responses,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )$valid

  out_list <- list()

  if (!is.null(variables_numeric)) {
    out_list$numeric <- variables_numeric
  }

  if (!is.null(variables_categorical)) {
    out_list$categorical <- variables_categorical
  }

  if (!is.null(variables_logical)) {
    out_list$logical <- variables_logical
  }

  out_list
}

#' Automatic selection of predictor scoring method
#'
#' @description
#' Internal function to select a proper \code{f_...()} function to compute preference order depending on the types of the response variable and the predictors. The selection criteria is available as a dataframe generated by [f_auto_rules()].
#'
#'
#' @inheritParams collinear
#'
#' @param response (optional, character string) Name of a response variable in \code{df}. Default: NULL.
#'
#' @return function name
#' @examples
#' data(vi_smol, vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#' vi_predictors_categorical <- identify_categorical_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' f_auto(
#'   df = vi_smol,
#'   response = "vi_numeric",
#'   predictors = vi_predictors_numeric
#'   )
#'
#' f_auto(
#'   df = vi_smol,
#'   response = "vi_binomial",
#'   predictors = vi_predictors_numeric
#'   )
#'
#' f_auto(
#'   df = vi_smol,
#'   response = "vi_categorical",
#'   predictors = vi_predictors_categorical
#'   )
#'
#'
#' @family preference_order_tools
#' @export
#' @autoglobal
f_auto <- function(
  df = NULL,
  response = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  function_name <- validate_arg_function_name(
    default_name = "collinear::f_auto()",
    ... = ...
  )

  quiet <- validate_arg_quiet(
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  response <- validate_arg_responses(
    df = df,
    responses = response,
    max_responses = 1,
    quiet = quiet,
    function_name = function_name
  )

  predictors <- validate_arg_predictors(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  df <- validate_arg_df(
    df = df,
    responses = response,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  if (is.null(response)) {
    stop(
      "\n",
      function_name,
      ": argument 'response' must not be NULL."
    )
  }

  response_type <- identify_response_type(
    df = df,
    response = response,
    quiet = quiet,
    function_name = function_name
  )

  if (response_type == "unknown") {
    stop(
      function_name,
      ": response type is 'unknown', please select an f_...() function suitable for your response data.",
      call. = FALSE
    )
  }

  #identify types of the predictors
  predictors_type <- lapply(
    X = df[, predictors, drop = FALSE],
    FUN = class
  ) |>
    unlist() |>
    gsub(
      pattern = "integer",
      replacement = "numeric"
    ) |>
    gsub(
      pattern = "character|factor",
      replacement = "categorical"
    ) |>
    unique()

  if (length(predictors_type) > 1) {
    predictors_type <- "mixed"
  }

  if (!(predictors_type %in% c("numeric", "categorical", "mixed"))) {
    stop(
      "\n",
      function_name,
      ": predictors type is 'unknown', please select an f_...() function suitable for your predictors.",
      call. = FALSE
    )
  }

  #select function
  rules <- f_auto_rules()

  f_name <- rules[
    rules$response_type == response_type &
      rules$predictors_type == predictors_type,
    "name"
  ]

  if (quiet == FALSE) {
    message(
      "\n",
      function_name,
      ": selected function '",
      f_name,
      "()' to compute preference order."
    )
  }

  f_name
}

#' Compute summary statistics for correlation and VIF
#'
#' Computes the the minimum, mean, maximum, and quantiles 0.05, 0.25, median (0.5), 0.75, and 0.95 of the correlations and variance inflation factors in a given dataframe. Wraps the functions [cor_stats()] and [vif_stats()]
#'
#' @inheritParams cor_matrix
#' @returns dataframe with columns \code{method} (with values "correlation" and "vif"), \code{statistic} and \code{value}
#'
#' @examples
#' data(vi_smol, package = "spatialData")
#' data(vi_predictors, package = "spatialData")
#' vi_predictors_numeric <- identify_numeric_variables(
#'   df = vi_smol,
#'   predictors = vi_predictors
#' )$valid
#'
#' ## OPTIONAL: parallelization setup
#' ## irrelevant when all predictors are numeric
#' ## only worth it for large data with many categoricals
#' # future::plan(
#' #   future::multisession,
#' #   workers = future::availableCores() - 1
#' # )
#'
#' ## OPTIONAL: progress bar
#' # progressr::handlers(global = TRUE)
#'
#' x <- collinear_stats(
#'   df = vi_smol,
#'   predictors = vi_predictors_numeric
#' )
#'
#' x
#'
#' ## OPTIONAL: disable parallelization
#' #future::plan(future::sequential)
#' @autoglobal
#' @family multicollinearity_assessment
#' @export
collinear_stats <- function(
  df = NULL,
  predictors = NULL,
  quiet = FALSE,
  ...
) {
  dots <- list(...)

  function_name <- validate_arg_function_name(
    default_name = "collinear::collinear_stats()",
    function_name = dots$function_name
  )

  df <- validate_arg_df_not_null(
    df = df,
    function_name = function_name
  )

  if (!inherits(x = df, what = "collinear_cor_df")) {
    predictors <- validate_arg_predictors(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )

    ncol.df <- ncol(df)

    df <- validate_arg_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )

    if (ncol.df > ncol(df)) {
      attributes(predictors)$validated <- NULL

      predictors <- validate_arg_predictors(
        df = df,
        predictors = predictors,
        quiet = quiet,
        function_name = function_name
      )
    }

    cor.df <- cor_df(
      df = df,
      predictors = predictors,
      quiet = quiet,
      function_name = function_name
    )
  } else {
    cor.df <- df
    predictors <- unique(
      c(cor.df$x, cor.df$y)
    )
  }

  cor.m <- cor_matrix(
    df = cor.df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name,
    m = dots$m
  )

  cor.stats <- cor_stats(
    df = cor.df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name
  )

  cor.vif <- vif_stats(
    df = cor.df,
    predictors = predictors,
    quiet = quiet,
    function_name = function_name,
    m = cor.m
  )

  out <- rbind(
    cor.stats,
    cor.vif
  )

  out
}

#' List predictor scoring functions
#'
#' @return dataframe
#' @export
#' @autoglobal
#' @family preference_order_tools
#' @examples
#' f_functions()
f_functions <- function() {
  f_list <- list(
    c(
      "f_numeric_glm",
      "numeric",
      "mixed",
      "stats::glm(y ~ x, family = gaussian(link = 'identity'))",
      "R-squared"
    ),
    c(
      "f_numeric_gam",
      "numeric",
      "numeric",
      "mgcv::gam(y ~ s(x), family = gaussian(link = 'identity'))",
      "R-squared"
    ),
    c("f_numeric_rf", "numeric", "mixed", "ranger::ranger(y ~ x)", "R-squared"),
    c("f_count_rf", "integer", "mixed", "ranger::ranger(y ~ x)", "R-squared"),
    c(
      "f_count_glm",
      "integer",
      "mixed",
      "stats::glm(y ~ x, family = poisson(link = 'log'))",
      "R-squared"
    ),
    c(
      "f_count_gam",
      "integer",
      "mixed",
      "mgcv::gam(y ~ s(x), family = poisson(link = 'log'))",
      "R-squared"
    ),
    c(
      "f_binomial_glm",
      "binomial",
      "mixed",
      "stats::glm(y ~ x, family = quasibinomial(link = 'logit'), weights = case_weights(y))",
      "AUC"
    ),
    c(
      "f_binomial_gam",
      "binomial",
      "numeric",
      "mgcv::gam(y ~ s(x), family = quasibinomial(link = 'logit'), weights = collinear::case_weights(y))",
      "AUC"
    ),
    c(
      "f_binomial_rf",
      "binomial",
      "mixed",
      "ranger::ranger(y ~ x, case.weights = collinear::case_weights(y))",
      "AUC"
    ),
    c(
      "f_categorical_rf",
      "categorical",
      "mixed",
      "ranger::ranger(y ~ x, case.weights = collinear::case_weights(y))",
      "Cramer's V"
    )
  )

  f_df <- f_list |>
    as.data.frame() |>
    t() |>
    as.data.frame()

  rownames(f_df) <- NULL

  colnames(f_df) <- c(
    "name",
    "response_type",
    "predictors_types",
    "expression",
    "metric"
  )

  f_df
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target_encoding_lab.R
\name{target_encoding_lab}
\alias{target_encoding_lab}
\title{Convert categorical predictors to numeric via target encoding}
\usage{
target_encoding_lab(
  df = NULL,
  response = NULL,
  predictors = NULL,
  encoding_method = "loo",
  smoothing = 0,
  overwrite = FALSE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a numeric response variable in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{encoding_method}{(optional; character vector or NULL). Name of the target encoding methods. One or several of: "mean", "rank", "loo". If NULL, target encoding is ignored, and \code{df} is returned with no modification. Default: "loo"}

\item{smoothing}{(optional; integer vector) Argument of the method "mean". Groups smaller than this number have their means pulled towards the mean of the response across all cases. Default: 0}

\item{overwrite}{(optional; logical) If TRUE, the original predictors in \code{df} are overwritten with their encoded versions, but only one encoding method, smoothing, white noise, and seed are allowed. Otherwise, encoded predictors with their descriptive names are added to \code{df}. Default: FALSE}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe
}
\description{
Target encoding maps the values of categorical variables (of class \code{character} or \code{factor}) to numeric using another numeric variable as reference. The encoding methods implemented here are:

\itemize{
\item "mean" (implemented in \code{\link[=target_encoding_mean]{target_encoding_mean()}}): Maps each category to the average of reference numeric variable across the category cases. Variables encoded with this method are identified with the suffix "__encoded_mean". It has a method to control overfitting implemented via the argument \code{smoothing}. The integer value of this argument indicates a threshold in number of rows. Categories sized above this threshold are encoded with the group mean, while groups below it are encoded with a weighted mean of the group's mean and the global mean. This method is named "mean smoothing" in the relevant literature.
\item "rank" (implemented in \code{\link[=target_encoding_rank]{target_encoding_rank()}}): Returns the rank of the group as a integer, being 1 he group with the lower mean of the reference variable. Variables encoded with this method are identified with the suffix "__encoded_rank".
\item "loo" (implemented in \code{\link[=target_encoding_loo]{target_encoding_loo()}}): Known as the "leave-one-out method" in the literature, it encodes each categorical value with the mean of the response variable across all other group cases. This method controls overfitting better than "mean". Variables encoded with this method are identified with the suffix "__encoded_loo".
}

Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).
}
\examples{

data(vi_smol, package = "spatialData")

#applying all methods for a continuous response
df <- target_encoding_lab(
  df = vi_smol,
  response = "vi_numeric",
  predictors = "koppen_zone",
  encoding_method = c(
    "mean",
    "loo",
    "rank"
  )
)

#identify encoded predictors
predictors.encoded <- grep(
  pattern = "*__encoded*",
  x = colnames(df),
  value = TRUE
)

head(df[, predictors.encoded])


}
\references{
\itemize{
\item Micci-Barreca, D. (2001) A Preprocessing Scheme for High-Cardinality Categorical Attributes in Classification and Prediction Problems. SIGKDD Explor. Newsl. 3, 1, 27-32. doi: 10.1145/507533.507538
}
}
\seealso{
Other target_encoding:
\code{\link[=target_encoding_loo]{target_encoding_loo()}}
}
\author{
Blas M. Benito, PhD
}
\concept{target_encoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_auc.R
\name{score_auc}
\alias{score_auc}
\title{Compute area under the ROC curve between binomial observations and probabilistic predictions}
\usage{
score_auc(o = NULL, p = NULL, ...)
}
\arguments{
\item{o}{(required, numeric vector) Binomial observations (values 0 and 1). Default: NULL}

\item{p}{(required, numeric vector) Prediction of binomial model in the range 0-1. Default: NULL}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
numeric: Area Under the ROC Curve
}
\description{
Internal function to compute the AUC of binomial models within \code{\link[=preference_order]{preference_order()}}. Used within \code{\link[=f_binomial_glm]{f_binomial_glm()}}, \code{\link[=f_binomial_gam]{f_binomial_gam()}}, and \code{\link[=f_binomial_rf]{f_binomial_rf()}}. This function is build for speed and it does not check the inputs.
}
\examples{
  score_auc(
    o = c(1, 1, 1, 1, 0, 0, 0),
    p = c(1, 0.8, 0.7, 0.6, 0.5, 0.6, 0.7)
  )

}
\seealso{
Other modelling_tools:
\code{\link[=case_weights]{case_weights()}},
\code{\link[=model_formula]{model_formula()}},
\code{\link[=score_cramer]{score_cramer()}},
\code{\link[=score_r2]{score_r2()}}
}
\concept{modelling_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collinear_select.R
\name{collinear_select}
\alias{collinear_select}
\title{Dual multicollinearity filtering algorithm}
\usage{
collinear_select(
  df = NULL,
  response = NULL,
  predictors = NULL,
  preference_order = NULL,
  max_cor = 0.61,
  max_vif = 5,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional; character or NULL) Name of one response variable in \code{df}. Used to exclude columns when \code{predictors} is NULL, and to filter \code{preference_order} when it is a dataframe and contains several responses. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{preference_order}{(optional; character vector, dataframe from
\code{\link{preference_order}}, or NULL) Prioritizes predictors to preserve.}

\item{max_cor}{(optional; numeric or NULL) Maximum correlation allowed between pairs of \code{predictors}. Valid values are between 0.01 and 0.99, and recommended values are between 0.5 (strict) and 0.9 (permissive). Default: 0.7}

\item{max_vif}{(optional, numeric or NULL) Maximum Variance Inflation Factor allowed for \code{predictors} during multicollinearity filtering. Recommended values are between 2.5 (strict) and 10 (permissive). Default: 5}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
character vector: names of selected predictors
}
\description{
Automatizes multicollinearity filtering via pairwise correlation and/or variance inflation factors in dataframes with numeric and categorical predictors.

The argument \code{max_cor} determines the maximum pairwise correlation allowed in the resulting selection of predictors, while \code{max_vif} does the same for variance inflation factors.

The argument \code{preference_order} accepts a character vector of predictor names ranked from first to last index, or a dataframe resulting from \code{\link[=preference_order]{preference_order()}}. When two predictors in this vector or dataframe are highly collinear, the one with a lower ranking is removed. This option helps protect predictors of interest. If not provided, predictors are ranked from lower to higher multicollinearity.
Please check the sections \strong{Variance Inflation Factors}, \strong{VIF-based Filtering}, and \strong{Pairwise Correlation Filtering} at the end of this help file for further details.
}
\section{Pairwise Correlation Filtering}{


\code{\link{cor_select}} computes the global correlation matrix, orders
predictors by \code{preference_order} or by lower-to-higher summed
correlations, and sequentially selects predictors with pairwise correlations
below \code{max_cor}.
}

\section{Variance Inflation Factors}{


VIF for predictor \eqn{a} is computed as \eqn{1/(1-R^2)}, where \eqn{R^2} is
the multiple R-squared from regressing \eqn{a} on the other predictors.
Recommended maximums commonly used are 2.5, 5, and 10.
}

\section{VIF-based Filtering}{


\code{\link{vif_select}} ranks numeric predictors (user \code{preference_order}
if provided, otherwise from lower to higher VIF) and sequentially adds
predictors whose VIF against the current selection is below \code{max_vif}.
}

\examples{
data(vi_smol, package = "spatialData")

## OPTIONAL: parallelization setup
## irrelevant when all predictors are numeric
## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

x <- collinear_select(
  df = vi_smol,
  predictors = c(
  "koppen_zone",           #character
  "soil_type",             #factor
  "topo_elevation",        #numeric
  "soil_temperature_mean"  #numeric
    ),
  max_cor = 0.7,
  max_vif = 5
)

x

## OPTIONAL: disable parallelization
#future::plan(future::sequential)
}
\seealso{
Other multicollinearity_filtering:
\code{\link[=collinear]{collinear()}},
\code{\link[=cor_select]{cor_select()}},
\code{\link[=step_collinear]{step_collinear()}},
\code{\link[=vif_select]{vif_select()}}
}
\author{
Blas M. Benito, PhD
}
\concept{multicollinearity_filtering}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_function_name.R
\name{validate_arg_function_name}
\alias{validate_arg_function_name}
\title{Build hierarchical function names for messages}
\usage{
validate_arg_function_name(default_name = NULL, function_name = NULL, ...)
}
\arguments{
\item{default_name}{(optional, character) Name of the calling function. Default: NULL}

\item{function_name}{(optional, character) Name of the parent function. Default: NULL}

\item{...}{(optional) Used to pass \code{function_name} within these functions that don't have this argument.}
}
\value{
character
}
\description{
Concatenates parent and child function names for a better message, warning, and error tracing.
}
\examples{
x <- validate_arg_function_name(
  default_name = "child_function",
  function_name = "parent_function"
)

message(x)
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_cramer.R
\name{score_cramer}
\alias{score_cramer}
\title{Compute Cramer's V between categorical observations and predictions}
\usage{
score_cramer(o = NULL, p = NULL, ...)
}
\arguments{
\item{o}{(required; character vector) categorical observations. Default: NULL}

\item{p}{(required; character vector) categorical predictions. Default: NULL}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
numeric: Cramer's V
}
\description{
Internal function to compute the Cramer's V of categorical observations versus categorical model predictions. Please read the help file of \code{\link[=cor_cramer]{cor_cramer()}} for further details.
}
\examples{
score_cramer(
 o = c("a", "a", "b", "c", "c"),
 p = c("a", "b", "b", "c", "c")
 )

}
\seealso{
Other modelling_tools:
\code{\link[=case_weights]{case_weights()}},
\code{\link[=model_formula]{model_formula()}},
\code{\link[=score_auc]{score_auc()}},
\code{\link[=score_r2]{score_r2()}}
}
\concept{modelling_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_encoding_method.R
\name{validate_arg_encoding_method}
\alias{validate_arg_encoding_method}
\title{Check and validate argument \code{encoding_method}}
\usage{
validate_arg_encoding_method(
  encoding_method = "loo",
  overwrite = NULL,
  quiet = FALSE,
  function_name = NULL
)
}
\arguments{
\item{encoding_method}{(optional; character vector or NULL). Name of the target encoding methods. One or several of: "mean", "rank", "loo". If NULL, target encoding is ignored, and \code{df} is returned with no modification. Default: "loo"}

\item{overwrite}{(optional; logical) If TRUE, the original predictors in \code{df} are overwritten with their encoded versions, but only one encoding method, smoothing, white noise, and seed are allowed. Otherwise, encoded predictors with their descriptive names are added to \code{df}. Default: FALSE}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
character
}
\description{
Internal function to validate the argument \code{encoding_method} of \code{\link[=target_encoding_lab]{target_encoding_lab()}}.
}
\examples{
x <- validate_arg_encoding_method(
  encoding_method = "wrong_method"
  )
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif_stats.R
\name{vif_stats}
\alias{vif_stats}
\title{VIF Statistics}
\usage{
vif_stats(df = NULL, predictors = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe with columns \code{method} with the value "vif", \code{statistic} with the statistic name, and \code{value}.
}
\description{
Computes the the minimum, mean, maximum, and quantiles 0.05, 0.25, median (0.5), 0.75, and 0.95 of the column "vif" in the output of \code{\link[=vif_df]{vif_df()}}.
}
\examples{
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

# ## OPTIONAL: parallelization setup
# ## irrelevant when all predictors are numeric
# ## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

# ## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

x <- vif_stats(
  df = vi_smol,
  predictors = vi_predictors_numeric
)

x

## OPTIONAL: disable parallelization
#future::plan(future::sequential)
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_df]{vif_df()}}
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{experiment_cor_vs_vif}
\alias{experiment_cor_vs_vif}
\title{Dataframe with results of experiment comparing correlation and VIF thresholds}
\format{
A dataframe with 10,000 rows and 6 variables:
\describe{
\item{input_rows}{Number of rows in the input data subset.}
\item{input_predictors}{Number of predictors in the input data subset.}
\item{output_predictors}{Number of predictors selected by \code{\link[=vif_select]{vif_select()}} at the best-matching \code{max_vif}.}
\item{max_cor}{Maximum allowed pairwise correlation supplied to \code{\link[=cor_select]{cor_select()}}.}
\item{max_vif}{VIF threshold at which \code{\link[=vif_select]{vif_select()}} produced the highest Jaccard similarity with \code{\link[=cor_select]{cor_select()}} for the given \code{max_cor}.}
\item{out_selection_jaccard}{Jaccard similarity between the predictors selected by \code{\link[=cor_select]{cor_select()}} and \code{\link[=vif_select]{vif_select()}}.}
}
}
\usage{
data(experiment_cor_vs_vif)
}
\description{
A dataframe summarizing 10,000 experiments comparing the output of \code{\link[=cor_select]{cor_select()}} and \code{\link[=vif_select]{vif_select()}}. Each row records the input sampling parameters and the resulting feature-selection metrics.
}
\details{
The source data is a synthetic dataframe with 500 columns and 10,000 rows generated using \code{distantia::zoo_simulate()} with correlated time series (\code{independent = FALSE}).

Each iteration randomly subsets 10-50 predictors and 30-100 rows per predictor, applies \code{\link[=cor_select]{cor_select()}} with a random \code{max_cor} threshold, then finds the \code{max_vif} value that maximizes Jaccard similarity between the two selections.
}
\examples{
data(experiment_cor_vs_vif)
str(experiment_cor_vs_vif)
}
\seealso{
Other experiments:
\code{\link{experiment_adaptive_thresholds}},
\code{\link{gam_cor_to_vif}},
\code{\link{prediction_cor_to_vif}}
}
\concept{experiments}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{declare_spatialData}
\alias{declare_spatialData}
\title{Declare spatialData}
\usage{
declare_spatialData()
}
\value{
dataset spatialData::vi_smol
}
\description{
Declare spatialData
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/case_weights.R
\name{case_weights}
\alias{case_weights}
\title{Generate sample weights for imbalanced responses}
\usage{
case_weights(x = NULL, ...)
}
\arguments{
\item{x}{(required, integer, character, or factor vector) Values of a binomial, categorical, or factor variable. Default: NULL}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
numeric vector: case weights
}
\description{
Computes case weights adding to one for response variables of these types:

\itemize{
\item numeric binomial (1 and 0).
\item logical (TRUE and FALSE): converted to numeric internally.
\item categorical (character or factor)
}

Values NA, Inf, -Inf, and NaN are invalid for numeric and logical variables and will result in errors. For categorical variables, these are converted to their respective categories ("NA", "Inf", "-Inf", and "NaN") with their assigned case weights.

All returned weights sum to one.
}
\examples{
 #numeric vector
 y <- case_weights(
   x = c(0, 0, 1, 1)
   )

 #logical vector
 y <- case_weights(
   x = c(TRUE, TRUE, FALSE, FALSE)
   )

 #character vector
 y <- case_weights(
   x = c("a", "a", "b", "c")
   )
}
\seealso{
Other modelling_tools:
\code{\link[=model_formula]{model_formula()}},
\code{\link[=score_auc]{score_auc()}},
\code{\link[=score_cramer]{score_cramer()}},
\code{\link[=score_r2]{score_r2()}}
}
\concept{modelling_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collinear.R
\name{collinear}
\alias{collinear}
\title{Smart multicollinearity management}
\usage{
collinear(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  encoding_method = NULL,
  preference_order = NULL,
  f = f_auto,
  max_cor = NULL,
  max_vif = NULL,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{responses}{(optional; character, character vector, or NULL) Name of
one or several response variables in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{encoding_method}{(optional; character or NULL) One of "loo", "mean",
or "rank". If NULL, target encoding is disabled. Default: NULL.}

\item{preference_order}{(optional; character vector, dataframe from
\code{\link{preference_order}}, or NULL) Prioritizes predictors to preserve.}

\item{f}{(optional; unquoted function name or NULL) Function to rank
predictors by relationship with \code{responses}. See \code{\link{f_functions}}.
Default: \code{f_auto}.}

\item{max_cor}{(optional; numeric or NULL) Maximum allowed pairwise
correlation (0.01–0.99). Recommended between 0.5 and 0.9. If NULL and
\code{max_vif} is NULL, it is selected automatically. Default: NULL.}

\item{max_vif}{(optional; numeric or NULL) Maximum allowed VIF. Recommended
between 2.5 and 10. If NULL and \code{max_cor} is NULL, configured
automatically. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
A list of class \code{collinear_output} with sublists of class
\code{collinear_selection}. If \code{responses = NULL} a single sublist
named "result" is returned; otherwise a sublist per response is returned.
}
\description{
Automates multicollinearity management in datasets with mixed variable types
(numeric, categorical, and logical) through an integrated system of five key
components:

\describe{
\item{Target Encoding Integration (opt-in)}{
When \code{responses} is numeric, categorical predictors can be converted
to numeric using response values as reference. This enables VIF and
correlation analysis across mixed types. See \code{\link{target_encoding_lab}}.
}

\item{Intelligent Predictor Ranking (active by default)}{
Three prioritization strategies ensure the most relevant predictors are
retained during filtering:
\itemize{
\item \strong{User-defined ranking} (argument \code{preference_order}):
Accepts a character vector of predictor names or a dataframe from
\code{\link{preference_order}}. Lower-ranked collinear predictors are removed.
\item \strong{Response-based ranking} (\code{f}):
Uses \code{\link{f_auto}}, \code{\link{f_numeric_glm}}, or
\code{\link{f_binomial_rf}} to rank predictors by association with
the response. Supports cross-validation via \code{\link{preference_order}}.
\item \strong{Multicollinearity-based ranking} (default):
When both \code{preference_order} and \code{f} are \code{NULL},
predictors are ranked from lower to higher multicollinearity.
}
}

\item{Unified Correlation Framework (active by default)}{
Computes pairwise correlations between variable types using Pearson
(numeric–numeric), target encoding (numeric–categorical), and Cramer's V
(categorical–categorical). See \code{\link{cor_df}}, \code{\link{cor_matrix}},
and \code{\link{cor_cramer}}.
}

\item{Adaptive Filtering Thresholds (active by default)}{
When \code{max_cor} and \code{max_vif} are both \code{NULL}, thresholds
are determined from the median correlation structure of the predictors.
}

\item{Dual Filtering Strategy (active by default)}{
Combines two complementary methods while respecting predictor rankings:
\itemize{
\item \strong{Pairwise Correlation Filtering}:
Removes predictors with Pearson correlation or Cramer's V above
\code{max_cor}. See \code{\link{cor_select}}.
\item \strong{VIF-based Filtering}:
Removes numeric predictors with VIF above \code{max_vif}. See
\code{\link{vif_select}}, \code{\link{vif_df}}, and \code{\link{vif}}.
}
}
}

This function accepts parallelization via \code{future::plan()} and progress
bars via \code{progressr::handlers()}. Parallelization benefits
\code{\link{target_encoding_lab}}, \code{\link{preference_order}}, and
\code{\link{cor_select}}.
}
\section{Adaptive Multicollinearity Thresholds}{


When both \code{max_cor} and \code{max_vif} are \code{NULL}, the function
determines thresholds as follows:
\enumerate{
\item Compute the 75th percentile of pairwise correlations via
\code{\link{cor_stats}}.
\item Map that value through a sigmoid between 0.545 (VIF~2.5) and 0.785
(VIF~7.5), centered at 0.665, to get \code{max_cor}.
\item Compute \code{max_vif} from \code{max_cor} using
\code{\link{gam_cor_to_vif}}.
}
}

\section{Variance Inflation Factors}{


VIF for predictor \eqn{a} is computed as \eqn{1/(1-R^2)}, where \eqn{R^2} is
the multiple R-squared from regressing \eqn{a} on the other predictors.
Recommended maximums commonly used are 2.5, 5, and 10.
}

\section{VIF-based Filtering}{


\code{\link{vif_select}} ranks numeric predictors (user \code{preference_order}
if provided, otherwise from lower to higher VIF) and sequentially adds
predictors whose VIF against the current selection is below \code{max_vif}.
}

\section{Pairwise Correlation Filtering}{


\code{\link{cor_select}} computes the global correlation matrix, orders
predictors by \code{preference_order} or by lower-to-higher summed
correlations, and sequentially selects predictors with pairwise correlations
below \code{max_cor}.
}

\examples{
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid
x <- collinear(df = vi_smol[, vi_predictors_numeric])

}
\references{
\itemize{
\item David A. Belsley, D.A., Kuh, E., Welsch, R.E. (1980). Regression
Diagnostics: Identifying Influential Data and Sources of Collinearity.
John Wiley & Sons. DOI: 10.1002/0471725153.
\item Micci-Barreca, D. (2001) A Preprocessing Scheme for High-Cardinality
Categorical Attributes in Classification and Prediction Problems. SIGKDD
Explor. Newsl. 3, 1, 27-32. DOI: 10.1145/507533.507538
}
}
\seealso{
Other multicollinearity_filtering:
\code{\link[=collinear_select]{collinear_select()}},
\code{\link[=cor_select]{cor_select()}},
\code{\link[=step_collinear]{step_collinear()}},
\code{\link[=vif_select]{vif_select()}}
}
\concept{multicollinearity_filtering}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_f.R
\name{validate_arg_f}
\alias{validate_arg_f}
\title{Check and validate argument \code{f}}
\usage{
validate_arg_f(f = NULL, f_name = NULL, function_name = NULL)
}
\arguments{
\item{f}{(optional: function name) Unquoted function name without parenthesis (see \link{f_functions}). By default calls to \code{\link[=f_auto]{f_auto()}}, which selects a suitable function depending on the nature of the response and predictors. Set to NULL if \code{responses = NULL}. If NULL, predictors are ranked from lower to higher multicollinearity. Default: \code{f_auto}}

\item{f_name}{(optional, string) Name of the function \code{f}, as returned by \code{deparse(substitute(f))}. Default: NULL}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
function
}
\description{
Check and validate argument \code{f}
}
\examples{
x <- validate_arg_f(f = f_auto)
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{experiment_adaptive_thresholds}
\alias{experiment_adaptive_thresholds}
\title{Dataframe resulting from experiment to test the automatic selection of multicollinearity thresholds}
\format{
A dataframe with 10,000 rows and 9 variables:
\describe{
\item{input_rows}{Number of rows in the input data subset.}
\item{input_predictors}{Number of predictors in the input data subset.}
\item{output_predictors}{Number of predictors retained after filtering.}
\item{input_cor_q75}{75th percentile of pairwise correlations in the input data.}
\item{output_cor_q75}{75th percentile of pairwise correlations in the selected predictors.}
\item{input_cor_max}{Maximum pairwise correlation in the input data.}
\item{output_cor_max}{Maximum pairwise correlation in the selected predictors.}
\item{input_vif_max}{Maximum VIF in the input data.}
\item{output_vif_max}{Maximum VIF in the selected predictors.}
}
}
\usage{
data(experiment_adaptive_thresholds)
}
\description{
A dataframe summarizing 10,000 experiments validating the adaptive multicollinearity threshold system in \code{\link[=collinear]{collinear()}}. Each row records input data characteristics and the resulting multicollinearity metrics after filtering.
}
\details{
The source data is a synthetic dataframe with 500 columns and 10,000 rows generated using \code{distantia::zoo_simulate()} with correlated time series (\code{independent = FALSE}, \code{seasons = 0}).

Each iteration randomly subsets 10-100 predictors and 30-100 rows per predictor, then applies \code{\link[=collinear]{collinear()}} with automatic threshold configuration to assess:
\itemize{
\item Whether output VIF stays bounded between ~2.5 and ~7.5
\item How the system adapts to different correlation structures
\item How predictor retention scales with input size
}
}
\examples{
data(experiment_adaptive_thresholds)
str(experiment_adaptive_thresholds)
}
\seealso{
Other experiments:
\code{\link{experiment_cor_vs_vif}},
\code{\link{gam_cor_to_vif}},
\code{\link{prediction_cor_to_vif}}
}
\concept{experiments}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_quiet.R
\name{validate_arg_quiet}
\alias{validate_arg_quiet}
\title{Check and validate argument \code{quiet}}
\usage{
validate_arg_quiet(quiet = FALSE, function_name = NULL)
}
\arguments{
\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
logical
}
\description{
Internal function to validate the logical argument \code{quiet}, which triggers messaging when \code{FALSE}
}
\examples{

x <- validate_arg_quiet(
  quiet = TRUE
  )

attributes(x)$validated

}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.collinear_selection.R
\name{summary.collinear_selection}
\alias{summary.collinear_selection}
\title{Summarize single response selection results of \code{collinear}}
\usage{
\method{summary}{collinear_selection}(object = NULL, ...)
}
\arguments{
\item{object}{(sub-list in output of \code{\link[=collinear]{collinear()}} ) Object to summarize. Default: NULL}

\item{...}{Ignored, kept for consistency with generic.}
}
\value{
list: response name and character vector of selected predictors.
}
\description{
Summarize single response selection results of \code{collinear}
}
\seealso{
Other S3_methods:
\code{\link[=print.collinear_output]{print.collinear_output()}},
\code{\link[=print.collinear_selection]{print.collinear_selection()}},
\code{\link[=summary.collinear_output]{summary.collinear_output()}}
}
\concept{S3_methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif.R
\name{vif}
\alias{vif}
\title{Compute variance inflation factors from a correlation matrix}
\usage{
vif(m = NULL, quiet = FALSE, ...)
}
\arguments{
\item{m}{(required, matrix) Correlation matrix generated via \code{stats::cor()} or \code{\link[=cor_matrix]{cor_matrix()}}. Must have named dimensions. Default: NULL}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
named numeric vector
}
\description{
Computes the Variance Inflation Factors from a correlation matrix in two steps:
\itemize{
\item Applies \code{base::solve()} to transform the correlation matrix into a precision matrix, which is the inverse of the covariance matrix between all variables in \code{predictors}.
\item Applies \code{base::diag()} to extract the diagonal of the precision matrix, which contains the variance of the regression of each predictor against all other predictors, also known as Variance Inflation Factor
}
}
\section{Variance Inflation Factors}{


VIF for predictor \eqn{a} is computed as \eqn{1/(1-R^2)}, where \eqn{R^2} is
the multiple R-squared from regressing \eqn{a} on the other predictors.
Recommended maximums commonly used are 2.5, 5, and 10.
}

\examples{
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

m <- cor_matrix(
  df = vi_smol,
  predictors = vi_predictors_numeric[1:5]
)

vif(m)
}
\references{
\itemize{
\item David A. Belsley, D.A., Kuh, E., Welsch, R.E. (1980). Regression Diagnostics: Identifying Influential Data and Sources of Collinearity. John Wiley & Sons. DOI: 10.1002/0471725153.
}
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif_df]{vif_df()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f_binomial_glm.R
\name{f_binomial_glm}
\alias{f_binomial_glm}
\title{Area Under the Curve of Binomial GLM predictions vs. observations}
\usage{
f_binomial_glm(df, ...)
}
\arguments{
\item{df}{(required, dataframe) with columns:
\itemize{
\item "x": (numeric, character, factor) predictor.
\item "y" (integer) binomial response with unique values 0 and 1.
}}

\item{...}{(optional) Accepts the arguments \code{cv_training_fraction} (numeric between 0 and 1) and \code{cv_iterations} (integer between 1 and Inf) for cross validation.}
}
\value{
numeric or numeric vector: AUC
}
\description{
Fits a Quasibinomial GLM model \code{y ~ x} with the binomial  response \code{y} (values 0 and 1) and the numeric, character, or factor predictor \code{x} using \code{stats::glm()} and returns the area under the ROC curve of the observations against the predictions (see \code{\link[=score_auc]{score_auc()}}).

Cases are weighted with \code{\link[=case_weights]{case_weights()}} to prevent issues arising from class imbalance.

Supports cross-validation via the arguments arguments \code{cv_training_fraction} (numeric between 0 and 1) and \code{cv_iterations} (integer between 1 and \code{n}) introduced via ellipsis (\code{...}). See \code{\link[=preference_order]{preference_order()}} for further details.
}
\examples{
data(vi_smol, package = "spatialData")

df <- data.frame(
  y = vi_smol[["vi_binomial"]],
  x = vi_smol[["swi_max"]]
)

#no cross-validation
f_binomial_glm(df = df)

#cross-validation
f_binomial_glm(
  df = df,
  cv_training_fraction = 0.5,
  cv_iterations = 10
  )

#categorical predictor
df <- data.frame(
  y = vi_smol[["vi_binomial"]],
  x = vi_smol[["koppen_zone"]]
)

f_binomial_glm(df = df)
}
\seealso{
Other preference_order_functions:
\code{\link[=f_binomial_gam]{f_binomial_gam()}},
\code{\link[=f_binomial_rf]{f_binomial_rf()}},
\code{\link[=f_categorical_rf]{f_categorical_rf()}},
\code{\link[=f_count_gam]{f_count_gam()}},
\code{\link[=f_count_glm]{f_count_glm()}},
\code{\link[=f_count_rf]{f_count_rf()}},
\code{\link[=f_numeric_gam]{f_numeric_gam()}},
\code{\link[=f_numeric_glm]{f_numeric_glm()}},
\code{\link[=f_numeric_rf]{f_numeric_rf()}},
\code{\link[=preference_order]{preference_order()}}
}
\concept{preference_order_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_preference_order.R
\name{validate_arg_preference_order}
\alias{validate_arg_preference_order}
\title{Check and complete argument \code{preference_order}}
\usage{
validate_arg_preference_order(
  df = NULL,
  response = NULL,
  predictors = NULL,
  preference_order = NULL,
  quiet = FALSE,
  function_name = NULL,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a numeric response variable in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{preference_order}{(optional; character vector, dataframe from
\code{\link{preference_order}}, or NULL) Prioritizes predictors to preserve.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
character vector: ranked variable names
}
\description{
Internal function to validate the argument \code{preference_order} in \code{\link[=cor_select]{cor_select()}}, \code{\link[=vif_select]{vif_select()}}, \code{\link[=collinear_select]{collinear_select()}}, \code{\link[=collinear]{collinear()}}, and \code{\link[=collinear]{collinear()}}. Predictors not in \code{preference_order} are ranked from lower to higher sum of Pearson correlations with all other predictors.
}
\examples{
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

#input arguments must be validated first
df <- validate_arg_df(
  df = vi_smol,
  response = "vi_numeric",
  predictors = vi_predictors_numeric,
  quiet = TRUE
)

response <- validate_arg_responses(
  df = df,
  responses = "vi_numeric"
)

predictors <- validate_arg_predictors(
  df = df,
  response = response,
  predictors = vi_predictors_numeric[1:10]
)



#no preference order
#no response
#ranks predictor from lower to higher multicollinearity
y <- validate_arg_preference_order(
  df = df,
  predictors = predictors,
  preference_order = NULL
)

y
attributes(y)$validated


#validate character vector
y <- validate_arg_preference_order(
  df = df,
  predictors = predictors,
  preference_order = c(
    "swi_max",
    "swi_min",
    "swi_deviance" #does not exist
  )
)

y
attributes(y)$validated

#validate output of preference order
x <- preference_order(
  df = df,
  responses = response,
  predictors = predictors
)

x

y <- validate_arg_preference_order(
  df = df,
  response = response,
  predictors = predictors,
  preference_order = x
)

y
attributes(y)$validated
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f_auto.R
\name{f_auto}
\alias{f_auto}
\title{Automatic selection of predictor scoring method}
\usage{
f_auto(df = NULL, response = NULL, predictors = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a response variable in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
function name
}
\description{
Internal function to select a proper \code{f_...()} function to compute preference order depending on the types of the response variable and the predictors. The selection criteria is available as a dataframe generated by \code{\link[=f_auto_rules]{f_auto_rules()}}.
}
\examples{
data(vi_smol, vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid
vi_predictors_categorical <- identify_categorical_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

f_auto(
  df = vi_smol,
  response = "vi_numeric",
  predictors = vi_predictors_numeric
  )

f_auto(
  df = vi_smol,
  response = "vi_binomial",
  predictors = vi_predictors_numeric
  )

f_auto(
  df = vi_smol,
  response = "vi_categorical",
  predictors = vi_predictors_categorical
  )


}
\seealso{
Other preference_order_tools:
\code{\link[=f_auto_rules]{f_auto_rules()}},
\code{\link[=f_functions]{f_functions()}}
}
\concept{preference_order_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_df_not_null.R
\name{validate_arg_df_not_null}
\alias{validate_arg_df_not_null}
\title{Ensure that argument \code{df} is not \code{NULL}}
\usage{
validate_arg_df_not_null(df = NULL, function_name = NULL)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
dataframe
}
\description{
Internal function to validate the default value of the argument \code{df}.
}
\examples{
data(vi_smol, package = "spatialData")
df <- validate_arg_df_not_null(
  df = vi_smol
  )
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f_auto_rules.R
\name{f_auto_rules}
\alias{f_auto_rules}
\title{Decision rules for \code{f_auto()}}
\usage{
f_auto_rules()
}
\value{
dataframe
}
\description{
Dataframe with rules used by \code{\link[=f_auto]{f_auto()}} to select the function \code{f} in \code{\link[=f_functions]{f_functions()}} to compute preference order in \code{\link[=preference_order]{preference_order()}}. In most cases, random forest is selected as base model to provide homogeneous results across case types.
}
\examples{
f_auto_rules()
}
\seealso{
Other preference_order_tools:
\code{\link[=f_auto]{f_auto()}},
\code{\link[=f_functions]{f_functions()}}
}
\concept{preference_order_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_responses.R
\name{validate_arg_responses}
\alias{validate_arg_responses}
\title{Check and validate arguments \code{response} and \code{responses}}
\usage{
validate_arg_responses(
  df = NULL,
  responses = NULL,
  max_responses = NULL,
  quiet = FALSE,
  function_name = NULL
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{responses}{(optional; character, character vector, or NULL) Name of
one or several response variables in \code{df}. Default: NULL.}

\item{max_responses}{(required, integer or NULL) Maximum number of responses to consider. Default: NULL}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
character string: response name
}
\description{
Internal function validate the arguments \code{response} and \code{responses}. It checks that its value exists as a column name of \code{df},
}
\examples{

data(vi_smol, package = "spatialData")

x <- validate_arg_responses(
  df = vi_smol,
  responses = "vi_numeric"
)

attributes(x)$validated

}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_max_vif.R
\name{validate_arg_max_vif}
\alias{validate_arg_max_vif}
\title{Check and constrain argument \code{max_vif}}
\usage{
validate_arg_max_vif(max_vif = NULL, quiet = FALSE, function_name = NULL)
}
\arguments{
\item{max_vif}{(optional, numeric or NULL) Maximum Variance Inflation Factor allowed for \code{predictors} during multicollinearity filtering. Recommended values are between 2.5 (strict) and 10 (permissive). Default: 5}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
numeric or NULL
}
\description{
Check and constrain argument \code{max_vif}
}
\examples{
max_vif <- validate_arg_max_vif(
  max_vif = 11, #wrong value
  quiet = FALSE
)

max_vif
attributes(max_vif)$validated
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.collinear_selection.R
\name{print.collinear_selection}
\alias{print.collinear_selection}
\title{Print single selection results from \code{collinear}}
\usage{
\method{print}{collinear_selection}(x = NULL, n = 5, ...)
}
\arguments{
\item{x}{(required, sub-list in output of \code{\link[=collinear]{collinear()}}) Object to print. Default: NULL}

\item{n}{(optional, integer) Maximum printed vector length. Default: 5.}

\item{...}{Ignored, kept for consistency with generic.}
}
\description{
Print single selection results from \code{collinear}
}
\seealso{
Other S3_methods:
\code{\link[=print.collinear_output]{print.collinear_output()}},
\code{\link[=summary.collinear_output]{summary.collinear_output()}},
\code{\link[=summary.collinear_selection]{summary.collinear_selection()}}
}
\concept{S3_methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cor_cramer.R
\name{cor_cramer}
\alias{cor_cramer}
\title{Quantify association between categorical variables}
\usage{
cor_cramer(x = NULL, y = NULL, check_input = TRUE, ...)
}
\arguments{
\item{x}{(required; vector) Values of a categorical variable (character or vector). Converted to character if numeric or logical. Default: NULL}

\item{y}{(required; vector) Values of a categorical variable (character or vector). Converted to character if numeric or logical. Default: NULL}

\item{check_input}{(required; logical) If FALSE, disables data checking for a slightly faster execution. Default: TRUE}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
numeric: Cramer's V
}
\description{
Cramer's V extends the chi-squared test to quantify how strongly the categories of two variables co-occur. The value ranges from 0 to 1, where 0 indicates no  association and 1 indicates perfect association.

This function implements a bias-corrected version of Cramer's V, which adjusts for sample size and is more accurate for small samples. However, this bias correction means that even for binary variables, Cramer's V will no equal the Pearson correlation (the standard, uncorrected Cramer's V does match Pearson for binary data).

As the number of categories increases, Cramer's V and Pearson correlation measure increasingly different aspects of association and should not be directly compared.

If you intend to combine these measures in a multicollinearity analysis, interpret them with care. It is often preferable to convert non-numeric variables to numeric form (for example, via target encoding) before assessing multicollinearity.
}
\examples{

# perfect one-to-one association
cor_cramer(
  x = c("a", "a", "b", "c"),
  y = c("a", "a", "b", "c")
)

# still perfect: labels differ but mapping is unique
cor_cramer(
  x = c("a", "a", "b", "c"),
  y = c("a", "a", "b", "d")
)

# high but < 1: mostly aligned, one category of y repeats
cor_cramer(
  x = c("a", "a", "b", "c"),
  y = c("a", "a", "b", "b")
)

# appears similar by position, but no association by distribution
# (x = "a" mixes with y = "a" and "b")
cor_cramer(
  x = c("a", "a", "a", "c"),
  y = c("a", "a", "b", "b")
)

# numeric inputs are coerced to character internally
cor_cramer(
  x = c(1, 1, 2, 3),
  y = c(1, 1, 2, 2)
)

# logical inputs are also coerced to character
cor_cramer(
  x = c(TRUE, TRUE, FALSE, FALSE),
  y = c(TRUE, TRUE, FALSE, FALSE)
)

}
\references{
\itemize{
\item Cramér, H. (1946). Mathematical Methods of Statistics. Princeton: Princeton University Press, page 282 (Chapter 21. The two-dimensional case). ISBN 0-691-08004-6
}
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_df]{vif_df()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\author{
Blas M. Benito, PhD
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.collinear_output.R
\name{summary.collinear_output}
\alias{summary.collinear_output}
\title{Summarize all results of \code{collinear()}}
\usage{
\method{summary}{collinear_output}(object = NULL, ...)
}
\arguments{
\item{object}{(required, list of class \code{collinear_output}) Object to summarize. Default: NULL}

\item{...}{Ignored, kept for consistency with generic.}
}
\value{
list:
If \code{object} was created with \code{responses = NULL}, a sublist named "result" containing a vector with the selected predictors. Otherwise, a list named after each response containing the corresponding variable selection.
}
\description{
Summarize all results of \code{collinear()}
}
\seealso{
Other S3_methods:
\code{\link[=print.collinear_output]{print.collinear_output()}},
\code{\link[=print.collinear_selection]{print.collinear_selection()}},
\code{\link[=summary.collinear_selection]{summary.collinear_selection()}}
}
\concept{S3_methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cor_clusters.R
\name{cor_clusters}
\alias{cor_clusters}
\title{Group predictors by hierarchical correlation clustering}
\usage{
cor_clusters(
  df = NULL,
  predictors = NULL,
  max_cor = 0.7,
  method = "complete",
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with predictors or the output of \code{\link[=cor_df]{cor_df()}}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{max_cor}{(optional; numeric or NULL) Correlation value used to separate clustering groups. Valid values are between 0.01 and 0.99. Default: 0.7}

\item{method}{(optional, character string) Argument of \code{\link[stats:hclust]{stats::hclust()}} defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC) or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well. Default: "complete".}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
list:
\itemize{
\item df: dataframe with predictor names and their cluster IDs.
\item hclust: clustering object
}
}
\description{
Hierarchical clustering of predictors from their correlation matrix. Computes the correlation matrix with \code{\link[=cor_df]{cor_df()}} and \code{\link[=cor_matrix]{cor_matrix()}}, transforms it to a distance matrix using \code{stats::dist()}, computes a clustering solution with \code{stats::hclust()}, and applies \code{stats::cutree()} to separate groups based on the value of the argument \code{max_cor}.

Returns a dataframe with predictor names and their clusters, and optionally, prints a dendrogram of the clustering solution.

Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).
}
\examples{
data(vi_smol, package = "spatialData")

## OPTIONAL: parallelization setup
## irrelevant when all predictors are numeric
## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

#group predictors using max_cor as clustering threshold
clusters <- cor_clusters(
  df = vi_smol,
  predictors = c(
    "koppen_zone", #character
    "soil_type", #factor
    "topo_elevation", #numeric
    "soil_temperature_mean" #numeric
  ),
  max_cor = 0.75
)

#clusters dataframe
clusters$df

##plot hclust object
# graphics::plot(clusters$hclust)

##plot max_cor threshold
# graphics::abline(
#   h = 1 - 0.75,
#   col = "red4",
#   lty = 3,
#   lwd = 2
# )

## OPTIONAL: disable parallelization
#future::plan(future::sequential)
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_df]{vif_df()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/identify_valid_variables.R
\name{identify_valid_variables}
\alias{identify_valid_variables}
\title{Find valid numeric, categorical, and logical variables in a dataframe}
\usage{
identify_valid_variables(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  decimals = 4,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{responses}{(optional; character, character vector, or NULL) Name of
one or several response variables in \code{df}. Default: NULL.}

\item{predictors}{(required, character vector) Names of the predictors to identify. Default: NULL}

\item{decimals}{(required, integer) Number of decimal places for the zero variance test. Smaller numbers will increase the number of variables detected as near-zero variance. Recommended values will depend on the range of the numeric variables in 'df'. Default: 4}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
list
\itemize{
\item \code{numeric}: character vector of numeric predictors.
\item \code{categorical}: character vector of categorical (character and factor) predictors.
\item \code{logical}: character vector of logical predictors.
}
}
\description{
Returns a list with the names of the valid numeric, categorical, and logical variables in a modelling dataframe.
}
\examples{

data(vi_smol, vi_predictors, package = "spatialData")

x <- identify_valid_variables(
  df = vi_smol,
  predictors = vi_predictors
)

x

}
\seealso{
Other data_types:
\code{\link[=identify_categorical_variables]{identify_categorical_variables()}},
\code{\link[=identify_logical_variables]{identify_logical_variables()}},
\code{\link[=identify_numeric_variables]{identify_numeric_variables()}},
\code{\link[=identify_response_type]{identify_response_type()}},
\code{\link[=identify_zero_variance_variables]{identify_zero_variance_variables()}}
}
\author{
Blas M. Benito, PhD
}
\concept{data_types}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_formula.R
\name{model_formula}
\alias{model_formula}
\title{Build model formulas from response and predictors}
\usage{
model_formula(
  df = NULL,
  response = NULL,
  predictors = NULL,
  term_f = NULL,
  term_args = NULL,
  random_effects = NULL,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a response variable in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{term_f}{(optional; string). Name of function to apply to each term in the formula, such as "s" for \code{mgcv::s()} or any other smoothing function, "poly" for \code{stats::poly()}. Default: NULL}

\item{term_args}{(optional; string). Arguments of the function applied to each term. For example, for "poly" it can be "degree = 2, raw = TRUE". Default: NULL}

\item{random_effects}{(optional, string or character vector). Names of variables to be used as random effects. Each element is added to the final formula as \code{+(1 | random_effect_name)}. Default: NULL}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
list if \code{predictors} is a list or length of \code{response} is higher than one, and character vector otherwise.
}
\description{
Generates model formulas from a dataframe, a response name, and a vector of predictors that can be the output of a multicollinearity management function such as \code{\link[=collinear_select]{collinear_select()}} and the likes. Intended to help fit exploratory models from the result of a multicollinearity analysis.

The types of formulas it can generate are:
\itemize{
\item additive: \code{y ~ x + z}
\item polynomial: \code{y ~ poly(x, ...) + poly(z, ...)}
\item GAM: \code{y ~ s(x) + s(z)}
\item random effect: \code{y ~ x + (1 \ z)}
}
}
\examples{
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

#reduce collinearity
x <- collinear_select(
  df = vi_smol,
  predictors = vi_predictors_numeric
)

#additive formula
y <- model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = x
)

y

#using a formula in a model
m <- stats::lm(
 formula = y,
 data = vi_smol
 )

summary(m)

#classification formula (character response)
y <- model_formula(
  df = vi_smol,
  response = "vi_categorical",
  predictors = x
)

y


#polynomial formula (3rd degree)
y <- model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = x,
  term_f = "poly",
  term_args = "degree = 3, raw = TRUE"
)

y

#gam formula
y <- model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = x,
  term_f = "s"
)

y

#random effect
y <- model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = x,
  random_effects = "country_name" #from vi_smol$country_name
)

y
}
\seealso{
Other modelling_tools:
\code{\link[=case_weights]{case_weights()}},
\code{\link[=score_auc]{score_auc()}},
\code{\link[=score_cramer]{score_cramer()}},
\code{\link[=score_r2]{score_r2()}}
}
\concept{modelling_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{prediction_cor_to_vif}
\alias{prediction_cor_to_vif}
\title{Prediction of the model \code{gam_cor_to_vif} across correlation values}
\format{
A dataframe with 901 rows and 2 numeric columns:
\describe{
\item{max_cor}{Maximum allowed pairwise correlation, from 0.10 to 1.00 in steps of 0.001.}
\item{max_vif}{Predicted VIF threshold corresponding to each \code{max_cor}.}
}
}
\usage{
data(prediction_cor_to_vif)
}
\description{
Dataframe with predicted VIF threshold corresponding to a given correlation threshold..
}
\details{
Values were generated by applying \code{mgcv::predict.gam()} to the fitted model \link{gam_cor_to_vif} and rounding to three decimal places.
}
\examples{
data(prediction_cor_to_vif)
head(prediction_cor_to_vif)
plot(max_vif ~ max_cor, data = prediction_cor_to_vif, type = "l")
}
\seealso{
Other experiments:
\code{\link{experiment_adaptive_thresholds}},
\code{\link{experiment_cor_vs_vif}},
\code{\link{gam_cor_to_vif}}
}
\concept{experiments}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{toy}
\alias{toy}
\title{Toy dataframe with varying levels of multicollinearity}
\format{
dataframe with 2000 rows and 5 columns.
}
\usage{
data(toy)
}
\description{
Dataframe with known relationship between responses and predictors useful to illustrate multicollinearity concepts.
}
\details{
Columns:
\itemize{
\item \code{y}: response variable generated from \code{a * 0.75 + b * 0.25 + noise}.
\item \code{a}: most important predictor of \code{y}, uncorrelated with \code{b}.
\item \code{b}: second most important predictor of \code{y}, uncorrelated with \code{a}.
\item \code{c}: generated from \code{a + noise}.
\item \code{d}: generated from \code{(a + b)/2 + noise}.
}

These are variance inflation factors of the predictors in \code{toy}.
variable vif
b       4.062
d       6.804
c       13.263
a       16.161
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{gam_cor_to_vif}
\alias{gam_cor_to_vif}
\title{GAM describing the relationship between correlation and VIF thresholds}
\format{
A \code{\link[mgcv]{gam}} object.
}
\source{
Generated internally from \link{experiment_cor_vs_vif}.
}
\usage{
data(gam_cor_to_vif)
}
\description{
A fitted generalized additive model describing \code{max_vif} as a function of \code{max_cor} in \link{experiment_cor_vs_vif}.
}
\details{
The model parameters (basis dimension \code{k} and weight exponent) were selected via optimization, filtering for models in the top 90\\% of R-squared and bottom 10\\% of effective degrees of freedom to balance fit quality and parsimony.

The final model uses squared Jaccard similarity as weights to emphasize cases with high agreement between \code{\link[=cor_select]{cor_select()}} and \code{\link[=vif_select]{vif_select()}}.

Model performance:
\itemize{
\item Adjusted R-squared: 0.834
\item Deviance explained: 83.4\\%
\item Effective degrees of freedom for smooth: ~6
}
}
\examples{
data(gam_cor_to_vif)
plot(gam_cor_to_vif, shade = TRUE)
}
\seealso{
Other experiments:
\code{\link{experiment_adaptive_thresholds}},
\code{\link{experiment_cor_vs_vif}},
\code{\link{prediction_cor_to_vif}}
}
\concept{experiments}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f_functions.R
\name{f_functions}
\alias{f_functions}
\title{List predictor scoring functions}
\usage{
f_functions()
}
\value{
dataframe
}
\description{
List predictor scoring functions
}
\examples{
f_functions()
}
\seealso{
Other preference_order_tools:
\code{\link[=f_auto]{f_auto()}},
\code{\link[=f_auto_rules]{f_auto_rules()}}
}
\concept{preference_order_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_r2.R
\name{score_r2}
\alias{score_r2}
\title{Compute R-squared between numeric observations and predictions}
\usage{
score_r2(o = NULL, p = NULL, ...)
}
\arguments{
\item{o}{(required, numeric vector) Observations. Default: NULL}

\item{p}{(required, numeric vector) Predictions. Default: NULL}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
numeric: Pearson R-squared
}
\description{
Internal function to compute the R-squared of observations versus predictions via \code{stats::cor()}. Used within \code{\link[=f_numeric_glm]{f_numeric_glm()}}, \code{\link[=f_numeric_gam]{f_numeric_gam()}}, \code{\link[=f_numeric_rf]{f_numeric_rf()}}, \code{\link[=f_count_glm]{f_count_glm()}}, and \code{\link[=f_count_gam]{f_count_gam()}}.
}
\examples{
  score_r2(
    o = c(1, 1, 1, 0.5, 0.5, 0, 0),
    p = c(1, 0.8, 0.7, 0.6, 0.5, 0.1, 0)
  )
}
\seealso{
Other modelling_tools:
\code{\link[=case_weights]{case_weights()}},
\code{\link[=model_formula]{model_formula()}},
\code{\link[=score_auc]{score_auc()}},
\code{\link[=score_cramer]{score_cramer()}}
}
\concept{modelling_tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif_df.R
\name{vif_df}
\alias{vif_df}
\title{Compute variance inflation factors dataframe}
\usage{
vif_df(df = NULL, predictors = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe with columns:
\itemize{
\item \code{predictor}: Character, predictor name.
\item \code{vif}: Numeric, variance inflation factor
}
}
\description{
Computes the pairwise correlation matrix between all pairs of predictors via \code{\link[=cor_df]{cor_df()}} and \code{\link[=cor_matrix]{cor_matrix()}}, applies \code{\link[=vif]{vif()}} to the resulting matrix to compute Variance Inflation Factors, and returns the result as a dataframe.
}
\section{Variance Inflation Factors}{


VIF for predictor \eqn{a} is computed as \eqn{1/(1-R^2)}, where \eqn{R^2} is
the multiple R-squared from regressing \eqn{a} on the other predictors.
Recommended maximums commonly used are 2.5, 5, and 10.
}

\examples{

data(vi_smol, package = "spatialData")

# ## OPTIONAL: parallelization setup
# ## irrelevant when all predictors are numeric
# ## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

# ## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

#predictors
predictors = c(
  "koppen_zone", #character
  "soil_type", #factor
  "topo_elevation", #numeric
  "soil_temperature_mean" #numeric
)

x <- vif_df(
  df = vi_smol,
  predictors = predictors
)

x

## OPTIONAL: disable parallelization
#future::plan(future::sequential)

}
\references{
\itemize{
\item David A. Belsley, D.A., Kuh, E., Welsch, R.E. (1980). Regression Diagnostics: Identifying Influential Data and Sources of Collinearity. John Wiley & Sons. DOI: 10.1002/0471725153.
}
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cor_df.R
\name{cor_df}
\alias{cor_df}
\title{Compute signed pairwise correlations dataframe}
\usage{
cor_df(df = NULL, predictors = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe with columns:
\itemize{
\item \code{x}: character, first predictor name.
\item \code{y}: character, second predictor name.
\item \code{correlation}: numeric, Pearson correlation (numeric vs. numeric and numeric vs. categorical) or Cramer's V (categorical vs. categorical).
}
}
\description{
Computes pairwise correlations between predictors using appropriate methods for different variable types:
\itemize{
\item \strong{Numeric vs. Numeric}: Pearson correlation via \code{stats::cor()}.
\item \strong{Numeric vs. Categorical}: Target-encodes the categorical variable  using the numeric variable as reference via \code{\link[=target_encoding_lab]{target_encoding_lab()}} with leave-one-out method, then computes Pearson correlation.
\item \strong{Categorical vs. Categorical}: Cramer's V via \code{\link[=cor_cramer]{cor_cramer()}} as a measure of association. See \code{\link[=cor_cramer]{cor_cramer()}} for important notes on mixing Pearson correlation and Cramer's V in multicollinearity analysis.
}

Parallelization via \code{future::plan()} and progress bars via \code{progressr::handlers()} are supported but only beneficial for large datasets with categorical predictors. Numeric-only correlations do not use parallelization or progress bars. Example: With 16 workers, 30k rows (dataframe \link[spatialData:vi]{spatialData::vi}), 49 numeric and 12 categorical predictors (see \link[spatialData:vi_predictors]{spatialData::vi_predictors}), parallelization achieves a 5.4x speedup (147s → 27s).
}
\examples{
data(vi_smol, package = "spatialData")

## OPTIONAL: parallelization setup
## irrelevant when all predictors are numeric
## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

#predictors
predictors = c(
  "koppen_zone", #character
  "soil_type", #factor
  "topo_elevation", #numeric
  "soil_temperature_mean" #numeric
)

x <- cor_df(
  df = vi_smol,
  predictors = predictors
)

x

## OPTIONAL: disable parallelization
#future::plan(future::sequential)
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_matrix]{cor_matrix()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_df]{vif_df()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/identify_response_type.R
\name{identify_response_type}
\alias{identify_response_type}
\title{Detect response variable type for model selection}
\usage{
identify_response_type(df = NULL, response = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a response variable in \code{df}. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
character string: response type
}
\description{
Used by \code{\link[=f_auto]{f_auto()}} to identify the type of a response variable and select a proper modelling method to compute preference order.
Supported types are:
\itemize{
\item "continuous-binary": decimal numbers and two unique values; results in a warning, as this type is difficult to model.
\item "continuous-low": decimal numbers and 3 to 5 unique values; results in a message, as this type is difficult to model.
\item "continuous-high": decimal numbers and more than 5 unique values.
\item "integer-binomial": integer with 0s and 1s, suitable for binomial models.
\item "integer-binary": integer with 2 unique values other than 0 and 1; returns a warning, as this type is difficult to model.
\item "integer-low": integer with 3 to 5 unique values or meets specified thresholds.
\item "integer-high": integer with more than 5 unique values suitable for count modelling.
\item "categorical": character or factor with 2 or more levels.
\item "unknown": when the response type cannot be determined.
}
}
\examples{
data(vi_smol, package = "spatialData")

identify_response_type(
  df = vi_smol,
  response = "vi_numeric"
)

identify_response_type(
  df = vi_smol,
  response = "vi_counts"
)

identify_response_type(
  df = vi_smol,
  response = "vi_binomial"
)

identify_response_type(
  df = vi_smol,
  response = "vi_categorical"
)

identify_response_type(
  df = vi_smol,
  response = "vi_factor"
)

}
\seealso{
Other data_types:
\code{\link[=identify_categorical_variables]{identify_categorical_variables()}},
\code{\link[=identify_logical_variables]{identify_logical_variables()}},
\code{\link[=identify_numeric_variables]{identify_numeric_variables()}},
\code{\link[=identify_valid_variables]{identify_valid_variables()}},
\code{\link[=identify_zero_variance_variables]{identify_zero_variance_variables()}}
}
\concept{data_types}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preference_order.R
\name{preference_order}
\alias{preference_order}
\title{Rank predictors by importance or multicollinearity}
\usage{
preference_order(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  f = f_auto,
  cv_training_fraction = 1,
  cv_iterations = 1,
  seed = 1,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{responses}{(optional; character, character vector, or NULL) Name of
one or several response variables in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{f}{(optional: function name) Unquoted function name without parenthesis (see \link{f_functions}). By default calls to \code{\link[=f_auto]{f_auto()}}, which selects a suitable function depending on the nature of the response and predictors. Set to NULL if \code{responses = NULL}. If NULL, predictors are ranked from lower to higher multicollinearity. Default: \code{f_auto}}

\item{cv_training_fraction}{(optional, numeric) Value between 0.1 and 1 defining the training faction used in cross-validation. If 1 (default), no cross-validation is performed, and the resulting metric is computed from all observations and predictions. Automatically set to 1 when \code{cv_iterations = 1}. Default: 1}

\item{cv_iterations}{(optional, integer) Number of cross-validation iterations to perform. The recommended range lies between 30 and 100. In general, smaller datasets and large values of \code{cv_training_fraction} require more iterations to achieve stability. Automatically set to 1 when \code{cv_training_fraction = 1}. Default: 1}

\item{seed}{(optional, integer) Random seed, required for reproducibility when using cross-validation or random forest models. Default: 1}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe:
\itemize{
\item \code{response}: character, response name, if any, or \code{"none"} otherwise.

\item \code{predictor}: character, name of the predictor.

\item \code{f}: name of the function used to compute the preference order. If argument \code{f} is NULL, the value "stats::cor()" is added to this column.

\item \code{metric}: name of the metric used to assess strength of association. Usually one of "R-squared", "AUC" (Area Under the ROC Curve), or "Cramer's V". If \code{f} is a custom function not in \code{\link[=f_functions]{f_functions()}}, then \code{metric} is set to "custom". If \code{f} is NULL, then "1 - R-squared" is returned in this column.

\item \code{score}: value of the metric returned by \code{f} to assess the association between the \code{response} and each given \code{predictor}.

\item \code{rank}: integer value indicating the rank of the predictor.
}
}
\description{
Generates a valid input for the argument \code{preference_order} of the functions \code{\link[=vif_select]{vif_select()}}, \code{\link[=cor_select]{cor_select()}}, \code{\link[=collinear_select]{collinear_select()}}, and \code{\link[=collinear]{collinear()}}. This argument helps preserve important predictors during multicollinearity filtering.

The function works in two different ways:
\itemize{
\item When \code{f} is NULL, it ranks the predictors from lower to higher multicollinearity, computed as one minus the average Pearson correlation between the given predictor against all others. This option is useful when the goal is to limit redundancy in a large dataset and there is not an specific model to train in mind.
\item When \code{responses} and \code{f} are not NULL, it ranks the predictors by the strength of their association with a response based on the evaluation of univariate models. This is the best possible option when the end-goal is training a model.
}

The argument \code{f} (requires a valid \code{resopnses} argument) defines how the strength of association between the response and each predictor is computed. By default it calls \code{\link[=f_auto]{f_auto()}}, which uses \code{\link[=f_auto_rules]{f_auto_rules()}} to select a suitable function depending on the types of the response and the predictors. This option is designed to provide sensible, general-purpose defaults optimized for speed and stability rather than any specific modeling approach.

For more fine-tuned control, the package offers the following \code{f} functions (see \code{\link[=f_functions]{f_functions()}}):

\itemize{
\item \strong{Numeric response}:
\itemize{
\item \code{\link[=f_numeric_glm]{f_numeric_glm()}}: Pearson's R-squared of response versus the predictions of a Gaussian GLM.
\item \code{\link[=f_numeric_gam]{f_numeric_gam()}}: GAM model fitted with \code{mgcv::gam()}.
\item \code{\link[=f_numeric_rf]{f_numeric_rf()}}: Random Forest model fitted with \code{ranger::ranger()}.
}
\item \strong{Integer counts response}:
\itemize{
\item \code{\link[=f_count_glm]{f_count_glm()}}: Pearson's R-squared of a Poisson GLM.
\item \code{\link[=f_count_gam]{f_count_gam()}}: Poisson GAM.
\item \code{\link[=f_count_rf]{f_count_rf()}}: Random Forest model fitted with \code{ranger::ranger()}.
}
\item \strong{Binomial response (1 and 0)}:
\itemize{
\item \code{\link[=f_binomial_glm]{f_binomial_glm()}}: AUC of Quasibinomial GLM with weighted cases.
\item \code{\link[=f_binomial_gam]{f_binomial_gam()}}: AUC of Quasibinomial GAM with weighted cases.
\item \code{\link[=f_binomial_rf]{f_binomial_rf()}}: AUC of a Random Forest model with weighted cases.
}
\item \strong{Categorical response}:
\itemize{
\item \code{\link[=f_categorical_rf]{f_categorical_rf()}}: Cramer's V of the response against the predictions of a classification Random Forest model.
}
}

These functions accept a cross-validation setup via the arguments \code{cv_iterations} and \code{cv_training_fraction}.

Additionally, the argument \code{f} accepts any custom function taking a dataframe with the columns "x" (predictor) and "y" (response) and returning a numeric indicator of association.

Accepts a parallelization setup via \code{future::plan()} and a progress bar via \code{progressr::handlers()} (see examples).

Accepts a character vector of response variables as input for the argument \code{responses}. When more than one response is provided, the output is a named list of preference data frames.
}
\examples{
#load example data
data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

##OPTIONAL: parallelization setup
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

##OPTIONAL: progress bar
##does not work in R examples
# progressr::handlers(global = TRUE)

#ranking predictors from lower to higher multicollinearity
#------------------------------------------------
x <- preference_order(
  df = vi_smol,
  responses = NULL, #default value
  predictors = vi_predictors_numeric[1:10],
  f = NULL #must be explicit
)

x

#automatic selection of ranking function
#------------------------------------------------
x <- preference_order(
  df = vi_smol,
  responses = c("vi_numeric", "vi_categorical"),
  predictors = vi_predictors_numeric[1:10],
  f = f_auto
  )

x

#user selection of ranking function
#------------------------------------------------
#Poisson GLM for a integer counts response
x <- preference_order(
  df = vi_smol,
  responses = "vi_binomial",
  predictors = vi_predictors_numeric[1:10],
  f = f_binomial_glm
)

x

#cross-validation
#------------------------------------------------
x <- preference_order(
  df = vi_smol,
  responses = "vi_binomial",
  predictors = vi_predictors_numeric[1:10],
  f = f_binomial_glm,
  cv_training_fraction = 0.5,
  cv_iterations = 10
)

x

#custom pairwise correlation function
#------------------------------------------------
#custom functions need the ellipsis argument
f_rsquared <- function(df, ...){
    stats::cor(
      x = df$x,
      y = df$y,
      use = "complete.obs"
    )^2
}

x <- preference_order(
  df = vi_smol,
  responses = "vi_numeric",
  predictors = vi_predictors_numeric[1:10],
  f = f_rsquared
)

x

#resetting to sequential processing
#future::plan(future::sequential)
}
\seealso{
Other preference_order_functions:
\code{\link[=f_binomial_gam]{f_binomial_gam()}},
\code{\link[=f_binomial_glm]{f_binomial_glm()}},
\code{\link[=f_binomial_rf]{f_binomial_rf()}},
\code{\link[=f_categorical_rf]{f_categorical_rf()}},
\code{\link[=f_count_gam]{f_count_gam()}},
\code{\link[=f_count_glm]{f_count_glm()}},
\code{\link[=f_count_rf]{f_count_rf()}},
\code{\link[=f_numeric_gam]{f_numeric_gam()}},
\code{\link[=f_numeric_glm]{f_numeric_glm()}},
\code{\link[=f_numeric_rf]{f_numeric_rf()}}
}
\author{
Blas M. Benito, PhD
}
\concept{preference_order_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cor_matrix.R
\name{cor_matrix}
\alias{cor_matrix}
\title{Signed pairwise correlation matrix}
\usage{
cor_matrix(df = NULL, predictors = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with predictors or the output of \code{\link[=cor_df]{cor_df()}}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
correlation matrix
}
\description{
Computes a square matrix of pairwise correlations for a set of numeric and/or categorical predictors.

If \code{df} is already a correlation dataframe generated by \code{\link[=cor_df]{cor_df()}}), the function transforms it into a correlation matrix. Otherwise, \code{\link[=cor_df]{cor_df()}} is used internally to compute pairwise correlations before generating the matrix.

Supports parallel computation via \code{future::plan()} and optional progress reporting via \code{progressr::handlers()}.
}
\examples{
data(vi_smol, package = "spatialData")

## OPTIONAL: parallelization setup
## irrelevant when all predictors are numeric
## only worth it for large data with many categoricals
# future::plan(
#   future::multisession,
#   workers = future::availableCores() - 1
# )

## OPTIONAL: progress bar
# progressr::handlers(global = TRUE)

predictors <- c(
  "koppen_zone", #character
  "soil_type", #factor
  "topo_elevation", #numeric
  "soil_temperature_mean" #numeric
)

#from dataframe with predictors
x <- cor_matrix(
  df = vi_smol,
  predictors = predictors
)

x

#from correlation dataframe
x <- cor_df(
  df = vi_smol,
  predictors = predictors
) |>
  cor_matrix()

x

## OPTIONAL: disable parallelization
#future::plan(future::sequential)
}
\seealso{
Other multicollinearity_assessment:
\code{\link[=collinear_stats]{collinear_stats()}},
\code{\link[=cor_clusters]{cor_clusters()}},
\code{\link[=cor_cramer]{cor_cramer()}},
\code{\link[=cor_df]{cor_df()}},
\code{\link[=cor_stats]{cor_stats()}},
\code{\link[=vif]{vif()}},
\code{\link[=vif_df]{vif_df()}},
\code{\link[=vif_stats]{vif_stats()}}
}
\author{
Blas M. Benito, PhD
}
\concept{multicollinearity_assessment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_arg_df.R
\name{validate_arg_df}
\alias{validate_arg_df}
\title{Check and prepare argument \code{df}}
\usage{
validate_arg_df(
  df = NULL,
  responses = NULL,
  predictors = NULL,
  quiet = FALSE,
  function_name = NULL
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{responses}{(optional; character, character vector, or NULL) Name of
one or several response variables in \code{df}. Default: NULL.}

\item{predictors}{(optional; character vector or NULL) Names of the
predictors in \code{df}. If NULL, all columns except \code{responses} and
constant/near-zero-variance columns are used. Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{function_name}{(optional, character string) Name of the function performing the argument check. Default: NULL}
}
\value{
dataframe
}
\description{
Internal function to validate the integrity of the argument \code{df}. It ensures that the dataframe has suitable dimensions for a multicollinearity analysis, transforms logical columns to numeric, character columns to factors, and converts \code{NaN}, \code{Inf} and \code{-Inf} to NA. Additionally, it checks the values of \code{responses} and \code{predictors} if these arguments are provided.
}
\examples{

data(vi_smol, vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

df <- validate_arg_df(
  df = vi_smol,
  responses = "vi_numeric",
  predictors = vi_predictors_numeric,
  quiet = FALSE
)

attributes(vi)$validated
}
\seealso{
Other argument_validation:
\code{\link[=drop_geometry_column]{drop_geometry_column()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target_encoding_loo.R,
%   R/target_encoding_mean.R, R/target_encoding_rank.R
\name{target_encoding_loo}
\alias{target_encoding_loo}
\alias{target_encoding_mean}
\alias{target_encoding_rank}
\title{Encode categories as response means}
\usage{
target_encoding_loo(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = NULL,
  ...
)

target_encoding_mean(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = 0,
  ...
)

target_encoding_rank(
  df = NULL,
  response = NULL,
  predictor = NULL,
  encoded_name = NULL,
  smoothing = NULL,
  ...
)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{response}{(optional, character string) Name of a numeric response variable in \code{df}. Default: NULL.}

\item{predictor}{(required; string) Name of the categorical predictor to encode. Default: NULL}

\item{encoded_name}{(optional, string) Name of the encoded predictor. Default: NULL}

\item{smoothing}{(optional; integer) Groups smaller than this number have their means pulled towards the mean of the response across all cases. Ignored by \code{\link[=target_encoding_rank]{target_encoding_rank()}} and \code{\link[=target_encoding_loo]{target_encoding_loo()}}. Default: 0}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe
}
\description{
Encode categories as response means
}
\examples{

#  loading example data
data(vi_smol, package = "spatialData")

#mean encoding
#-------------

df <- target_encoding_mean(
  df = vi_smol,
  response = "vi_numeric",
  predictor = "soil_type", #categorical
  encoded_name = "soil_type_encoded"
)

if(interactive()){

  plot(
    x = df$soil_type_encoded,
    y = df$vi_numeric,
    xlab = "encoded variable",
    ylab = "response"
  )

}


#rank encoding
#----------

df <- target_encoding_rank(
  df = vi_smol,
  response = "vi_numeric",
  predictor = "soil_type",
  encoded_name = "soil_type_encoded"
)

if(interactive()){

  plot(
    x = df$soil_type_encoded,
    y = df$vi_numeric,
    xlab = "encoded variable",
    ylab = "response"
  )

}


#leave-one-out
#-------------

df <- target_encoding_loo(
  df = vi_smol,
  response = "vi_numeric",
  predictor = "soil_type",
  encoded_name = "soil_type_encoded"
)

if(interactive()){

  plot(
    x = df$soil_type_encoded,
    y = df$vi_numeric,
    xlab = "encoded variable",
    ylab = "response"
  )

}

}
\seealso{
Other target_encoding:
\code{\link[=target_encoding_lab]{target_encoding_lab()}}
}
\concept{target_encoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_geometry_column.R
\name{drop_geometry_column}
\alias{drop_geometry_column}
\title{Removes \code{geometry} Column From \code{sf} Dataframes}
\usage{
drop_geometry_column(df = NULL, quiet = FALSE, ...)
}
\arguments{
\item{df}{(required; dataframe, tibble, or sf) A dataframe with responses
(optional) and predictors. Must have at least 10 rows for pairwise
correlation analysis, and \code{10 * (length(predictors) - 1)} for VIF.
Default: NULL.}

\item{quiet}{(optional; logical) If FALSE, messages are printed. Default: FALSE.}

\item{...}{(optional) Internal args (e.g. \code{function_name} for
\code{\link{validate_arg_function_name}}, a precomputed correlation matrix
\code{m}, or cross-validation args for \code{\link{preference_order}}).}
}
\value{
dataframe
}
\description{
Remove geometry column from sf objects
}
\examples{
data(vi_smol, package = "spatialData")

#creating fake geometry column without sf loaded
vi_smol$geometry <- NA
attr(
  x = vi_smol,
  which = "sf_column"
  ) <- "geometry"

#check new attribute
attributes(vi_smol)$sf_column

#drop geometry column
df <- drop_geometry_column(
  df = vi_smol
  )

#checking that the geometry was droppped
"geometry" \%in\% colnames(df)
attributes(df)$sf_column

}
\seealso{
Other argument_validation:
\code{\link[=validate_arg_df]{validate_arg_df()}},
\code{\link[=validate_arg_df_not_null]{validate_arg_df_not_null()}},
\code{\link[=validate_arg_encoding_method]{validate_arg_encoding_method()}},
\code{\link[=validate_arg_f]{validate_arg_f()}},
\code{\link[=validate_arg_function_name]{validate_arg_function_name()}},
\code{\link[=validate_arg_max_cor]{validate_arg_max_cor()}},
\code{\link[=validate_arg_max_vif]{validate_arg_max_vif()}},
\code{\link[=validate_arg_predictors]{validate_arg_predictors()}},
\code{\link[=validate_arg_preference_order]{validate_arg_preference_order()}},
\code{\link[=validate_arg_quiet]{validate_arg_quiet()}},
\code{\link[=validate_arg_responses]{validate_arg_responses()}}
}
\author{
Blas M. Benito, PhD
}
\concept{argument_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_collinear.R
\name{step_collinear}
\alias{step_collinear}
\alias{prep.step_collinear}
\alias{bake.step_collinear}
\title{Tidymodels recipe step for multicollinearity filtering}
\usage{
step_collinear(
  recipe,
  ...,
  role = NA,
  trained = FALSE,
  options = list(),
  selected = NULL,
  skip = FALSE,
  keep_original_cols = FALSE,
  id = recipes::rand_id("collinear")
)

\method{prep}{step_collinear}(x = NULL, training = NULL, info = NULL, ...)

\method{bake}{step_collinear}(object = NULL, new_data = NULL, ...)
}
\arguments{
\item{recipe}{(required, recipe) A recipe object to which this step will be added.}

\item{...}{(optional) Additional arguments (currently ignored).}

\item{role}{(optional, character) Not used by this step since no new variables are created. Default: NA}

\item{trained}{(optional, logical) Indicates if the step has been trained. Default: FALSE}

\item{options}{(optional, list) Named list of arguments passed to \code{\link[=collinear]{collinear()}}. Common options include:
\itemize{
\item \code{max_cor}: Maximum correlation threshold. If NULL (default), automatically set based on median correlation of predictors.
\item \code{max_vif}: Maximum VIF threshold. If NULL (default), automatically set to match the auto-computed max_cor.
\item \code{preference_order}: Vector of predictor names in priority order.
\item \code{f}: Function to compute preference order (default: \code{f_auto})
\item \code{quiet}: Suppress messages (default: TRUE)
}
Note: \code{encoding_method} is not supported in this step. The automatic threshold selection adapts to each dataset's correlation structure.}

\item{selected}{(character vector) Predictors retained after filtering. Populated during training and used during baking. Default: NULL}

\item{skip}{(optional, logical) Trigger to skip this step when baking. Default: FALSE.}

\item{keep_original_cols}{(optional, logical) Whether to keep original columns. Default: FALSE.}

\item{id}{(optional, character) Unique identifier for this step.}

\item{x}{(required, step_collinear object) The step to be trained. Default: NULL}

\item{training}{(required, data.frame) The training dataset used to estimate quantities.}

\item{info}{(optional, data.frame) Preprocessed information about variables in \code{training}. Default: NULL}

\item{object}{(required, \code{step_collinear} object) The trained step. Default: NULL}

\item{new_data}{(required, data.frame) New data to apply the step to. Default: NULL}
}
\value{
Updated recipe with new step.
}
\description{
Adds a step to a recipe created by \code{recipes::recipe()]+} to apply multicollinearity filtering via \code{\link[=collinear]{collinear()}}.

This function requires the \code{recipes} package to be installed.

Unlike \code{collinear()}, this wrapper does not perform target encoding, and the default value for the argument \code{quiet} is \code{TRUE}.
}
\examples{
\dontrun{
if(requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("workflows", quietly = TRUE)
    ){

data(vi_smol, package = "spatialData")
data(vi_predictors, package = "spatialData")
vi_predictors_numeric <- identify_numeric_variables(
  df = vi_smol,
  predictors = vi_predictors
)$valid

# model formula
vi_formula <- collinear::model_formula(
  df = vi_smol,
  response = "vi_numeric",
  predictors = vi_predictors_numeric
)

# recipe
vi_recipe <- recipes::recipe(
  formula = vi_formula,
  data = vi_smol
  ) |>
  #multicollinearity filtering
  collinear::step_collinear(
    recipes::all_predictors(),
    options = list(
      max_cor = 0.7,
      max_vif = 5,
      f = collinear::f_numeric_glm
      )
  ) |>
  #normalization
  recipes::step_normalize(
    recipes::all_predictors()
    )

# define linear regression model
vi_model <- parsnip::linear_reg() |>
  parsnip::set_engine("lm")

# create and fit workflow
vi_workflow <- workflows::workflow() |>
  workflows::add_recipe(vi_recipe) |>
  workflows::add_model(vi_model) |>
  workflows::fit(data = vi_smol)

vi_workflow

}
}
}
\seealso{
Other multicollinearity_filtering:
\code{\link[=collinear]{collinear()}},
\code{\link[=collinear_select]{collinear_select()}},
\code{\link[=cor_select]{cor_select()}},
\code{\link[=vif_select]{vif_select()}}
}
\concept{multicollinearity_filtering}
